package org.eso.oca.dal;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * Data Access Object class for MetaTable objects.
 * This class is used to retrieve MetaTable objects
 * from the sybase tables meta_tables and meta_columns.
 * 
 * @author szampier
 *
 */
public class MetaTableDAO {

	/** The Apache Log4j logger. */
	static Logger logger = Logger.getLogger(MetaTableDAO.class);
	
	/** The database fields to select. */
	private static final String ALL_FIELDS = "meta_tables.instrume"
			+ ",meta_tables.name"
            + ",meta_columns.name"
            + ",meta_columns.type"
            + ",meta_columns.length"
            + ",meta_columns.keyword";

	/**
	 * Loads all records from meta_tables and meta_columns and
	 * returns the corresponding array of MetaTable objects.
	 * 
	 * @param conn the JDBC connection.
	 * @return array of MetaTable objects corresponding to the contents
	 *         of meta_tables and meta_columns. 
	 * @throws SQLException
	 */
	public MetaTable[] loadAll(Connection conn) throws SQLException {
		final String thisMethod = "MetaTableDAO::loadAll()";
		logger.trace(thisMethod);

		// Check parameter
		if (conn == null) {
			throw new IllegalArgumentException(thisMethod
					+ " - connection must not be null");
		}
		logger.debug(thisMethod
				+ " - loading meta-data from database " + conn.getCatalog());

		// Build query
		StringBuffer query = new StringBuffer();
		query.append(" select " + ALL_FIELDS);
		query.append(" from meta_tables, meta_columns");
		query.append(" where meta_tables.name = meta_columns.tname");
		logger.debug(query);

		// Execute query
		PreparedStatement stmt = conn.prepareStatement(query.toString());
		ResultSet rs = stmt.executeQuery();

		Vector<MetaTable> metaTables = new Vector<MetaTable>();
		Vector<MetaColumn> metaColumns = new Vector<MetaColumn>();
		MetaTable metaTable = new MetaTable();
		MetaColumn metaColumn = new MetaColumn();
		String oldUniqueKey = null;
		String uniqueKey = null;

		// Parse results.
		while (rs.next()) {
			String instrume = rs.getString(1);
			String tableName = rs.getString(2);
			String columnName = rs.getString(3);
			String columnType = rs.getString(4);
			int columnLength = rs.getInt(5);
			String keyword = rs.getString(6);
			uniqueKey = instrume + tableName; 

			if (oldUniqueKey == null) {
				oldUniqueKey = uniqueKey;
			}
			
			if (! uniqueKey.equals(oldUniqueKey)) {
				metaTable.setMetaColumns(metaColumns.toArray(new MetaColumn[metaColumns.size()]));
				metaTables.add(metaTable);
				metaTable = new MetaTable();
				metaColumns = new Vector<MetaColumn>();
				oldUniqueKey = uniqueKey;
			}
			
			metaTable.setName(tableName);
			metaTable.setInstrume(instrume);
			
			metaColumn = new MetaColumn();
			metaColumn.setName(columnName);
			metaColumn.setType(columnType);
			metaColumn.setLength(columnLength);
			metaColumn.setKeyword(keyword);
			metaColumns.add(metaColumn);
		}
		
		return metaTables.toArray(new MetaTable[metaTables.size()]);
	}
}
