//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//COPYRIGHT (c) 2003 European Southern Observatory 
//LICENSE 
//
//PROJECT: VLT Data Flow System 
//AUTHOR: Olivier Chuzel - ESO/DMD/DFS 
//SUBSYSTEM: OCA 
//CI NUMBER: DFS-XXX-NNN 
//
//Created on May 30, 2005
//Name:  TableDAO.java
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

package org.eso.oca.data;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eso.oca.fitsmapper.DbFitsMapper;
import org.eso.oca.fitsmapper.RawDbFitsMapper;

/**
* Class TableDAO
* 
* This class handles the creation of a table in a database and its manipulation
* It contains a definition of the table and allow creating it, inserting,
* updating and selecting
* 
*/

public class TableDAO
{

  /**
   * Database connection
   */
  private DbConnManager connManager;

  private Connection conn;

  /*
   * Table definition
   */
  private Hashtable columns;

  private String tableName;

  /*
   * Logger
   */
  private static Logger logger = Logger.getLogger(TableDAO.class);

  /*
   * DBFitsMapper
   */
  private DbFitsMapper dbFitsMapper;

  /**
   * Constructor
   * 
   * @param tableName
   * @throws SQLException
   * @throws JdbcConnectionException
   */
  public TableDAO(String tableName) throws SQLException,
          JdbcConnectionException
  {
      this.tableName = tableName;
      connManager = DbConnManager.getInstance();
      conn = connManager.getObsConnection();
      this.columns = new Hashtable();
      this.dbFitsMapper = RawDbFitsMapper.getInstance();
  }

  public TableDAO(String tableName,
                  DbFitsMapper mapper,
                  Connection connexion) throws SQLException, JdbcConnectionException
	{
		this.tableName = tableName;
		//connManager = DbConnManager.getInstance();
		conn = connexion; //connManager.getObsConnection();
		this.columns = new Hashtable();
		this.dbFitsMapper = mapper;
	}
  
  /**
   * Adds a column to the table definition Adds it to the DBFitsMapper
   * 
   * @param keyword
   * @param type
   */
  public void addTableColumn(String keyword,
                             String type)
  {
      String colName = keywordToColumn(keyword);
      columns.put(colName, type);
      dbFitsMapper.addColumn(keyword, tableName, colName);
  }

  /**
   * Converts a keyword to a column name in the database Takes a keyword, and
   * replaces dots "." and hifens "-" by underscore characters
   * 
   * @param keyword
   * @return
   */
  public static String keywordToColumn(String keyword)
  {
      String tmp = keyword.replace('.', '_');
      String colName = tmp.replace('-', '_');
      return (colName);
  }

  /**
   * Clears table columns from the definition
   */
  public void clearTableColumns()
  {
      columns.clear();
  }

  /**
   * Performs and update to the temporary table, given a set of pairs
   * keyword/value and a where clause
   * 
   * Each pair keyword/value is part of the "set" clause, and each keyword is
   * converted to a column format For example, if one provides the following
   * pairs: (DO.CLASS,"BIAS"),(RAW.TYPE,"BIAS") will be converted to " set
   * DO_CLASS='BIAS' and RAW_TYPE='BIAS' "
   * 
   * @param metaKeywords
   * @param whereClause
   * @throws SQLException
   * @throws JdbcConnectionException
   */
  public void update(Hashtable metaKeywords, String fromClause,
          String whereClause) throws SQLException, JdbcConnectionException
  {
      PreparedStatement ps = null;
      StringBuffer _sql = null;
      String keyword;
      String value;
      String colName;

      // build the update
      _sql = new StringBuffer();
      _sql.append(" update " + getTableName() + " set  ");

      // set clause
      Enumeration keys = metaKeywords.keys();
      while (keys.hasMoreElements())
      {
          keyword = (String) keys.nextElement();
          value = (String) metaKeywords.get(keyword);

          // converts the keyword to a column name
          colName = TableDAO.keywordToColumn(keyword);
          _sql.append(colName + "='" + value + "'");

          if (keys.hasMoreElements())
              _sql.append(",");
      }

      // from and where clauses
      _sql.append(" from " + fromClause);
      _sql.append(" where " + whereClause);

      ps = conn.prepareStatement(_sql.toString());
      ps.executeUpdate();
      logger.debug(_sql.toString());
  }

  /**
   * Drops the temporary table from the database
   * 
   * @throws JdbcConnectionException
   * @throws SQLException
   *  
   */
  public void dropTable() throws SQLException, JdbcConnectionException
  {
      PreparedStatement ps = null;
      StringBuffer _sql = new StringBuffer();

      _sql.append("drop table " + getTableName());
      ps = conn.prepareStatement(_sql.toString());
      ps.executeUpdate();
  }

  /**
   * @return Returns the classification tableName.
   */
  public String getTableName()
  {
      return tableName;
  }

  /**
   * It returns a row from the table corresponding to the dpId
   * 
   * @param dpId
   * @param metaKeywords
   * @return
   * @throws SQLException
   */
  public Map getRow(String dpId, String[] metaKeywords) throws SQLException
  {
      Map res = new Hashtable();
      String sql = "select * from " + getTableName() + " where dp_id='"
              + dpId + "'";

      //execute the query
      PreparedStatement stmt = conn.prepareStatement(sql);
      ResultSet rs = stmt.executeQuery();

      if (rs.next())
      {
          for (int i = 0; i < metaKeywords.length; i++)
          {
              String metaKeyword = metaKeywords[i];
              String value = rs.getString(keywordToColumn(metaKeyword));
              if (value != null)
                  res.put(metaKeyword, value);
              else
                  res.put(metaKeyword, "");
          }
      }
      return (res);
  }

  /**
   * @param metaKeywords
   * @return
   * @throws SQLException
   */
  public String[][] getAllRows(String[] metaKeywords) throws SQLException
  {
      String[][] res;
      LinkedList rows = new LinkedList();

      String sql = "select * from " + getTableName();

      //execute the query
      PreparedStatement stmt = conn.prepareStatement(sql);
      ResultSet rs = stmt.executeQuery();

      String[] values = new String[metaKeywords.length + 1];

      // iterate through the rows
      while (rs.next())
      {
          values = new String[metaKeywords.length + 1];
          values[0] = rs.getString("dp_id");

          for (int i = 0; i < metaKeywords.length; i++)
          {
              String metaKeyword = metaKeywords[i];
              String value = rs.getString(keywordToColumn(metaKeyword));

              if (value != null)
                  values[i + 1] = value;
              else
                  values[i + 1] = "";
          }
          rows.add(values);
      }

      // copy the list to an array
      res = new String[rows.size()][metaKeywords.length + 1];

      int i = 0;
      for (Iterator iter = rows.iterator(); iter.hasNext();)
      {
          String[] row = (String[]) iter.next();

          res[i] = row;
          i++;
      }
      return (res);
  }
  
  //public DbConnManager getConnManager() { return connManager;};

  public Connection getConnection() { return conn;};

  public Hashtable getColumns() { return columns;};
  
  public DbFitsMapper getDbFitsMapper() {return dbFitsMapper;};
}


