package org.eso.oca.fits;

import nom.tam.fits.BasicHDU;
import nom.tam.fits.Fits;
import nom.tam.fits.FitsException;
import nom.tam.fits.FitsFactory;
import nom.tam.fits.Header;
import nom.tam.fits.HeaderCard;
import nom.tam.util.Cursor;

import org.apache.log4j.Logger;

/** This class encapsulates a FITS file and provides convenience methods to 
 * applications using OCA, allowing them to be written with minimal knowledge 
 * of the FITS library chosen by OCA's implementors. 
 *
 * @author Neil Ferguson ESO/DMD/DFS <nferguso@eso.org> 
 */
public class FITSHandler implements DataTransportFormatHandler
{
    static Logger logger = Logger.getLogger(FITSHandler.class);

    /** The index of the primary HDU in a FITS file. 
     *
     * @see #getFITSCards(int)
     * @see #getFITSCards(int,String[])
     */
    public static final int PRIMARY_HDU_INDEX = 0; 

    /** The FITS file held by this FITSHandler object. */ 
    private Fits fitsFile; 

    /** The filename of the FITS file held by this FITSHandler object. */ 
    private String fitsFilename; 
    
    /** whether or not we have iterated through all HDUs in the file */
    private boolean readAllHDUs = false;

    /** Constructs a FITSHandler to operate on the FITS file with the 
     * specified filename. 
     *
     * @throws FileHandlerException if the FITS file could not be opened. 
     */
    public FITSHandler(String filename) throws FileHandlerException 
    {
        /* Enable hierarchical keywords. */
        FitsFactory.setUseHierarch(true); 

        /* Open the FITS file. */ 
        try
        {
            fitsFile = new Fits(filename); 
        }
        catch(FitsException fe)
        {
            String message = "Could not open FITS file '" + filename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message, fe); 
        }

        fitsFilename = filename; 
    }

    /** Returns an array of TypedHeaderCards containing all of the header 
     * cards from the specified HDU. 
     * 
     * @param hdu the hdu whose header cards should be returned. 
     * 
     * @throws FileHandlerException if the HDU could not be read. 
     * @throws TypedHeaderCardException if a FITS header card could not be 
     *                                  stored as a TypedHeaderCard. 
     */ 
    public TypedHeaderCard[] getFITSCards(int hdu) 
        throws FileHandlerException, TypedHeaderCardException 
    {
        BasicHDU selectedHDU; 

        /* Get the specified Header + Data Unit. */
        try
        {
            selectedHDU = fitsFile.getHDU(hdu); 
        }
        catch(Exception e)
        {
            /* An error occurred whilst trying to retrieve the HDU - the 
             * HEASARC Java FITS documentation does not specify what can 
             * cause an exception at this point. Problems such as files 
             * being deleted whilst open are likely to show up here. 
             */
            String message = 
                "Could not retrieve HDU " + hdu + " from FITS file '" 
                + fitsFilename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message, e); 
        }

        if(selectedHDU == null)
        {
            /* The specified HDU could not be found. */ 
            String message = 
                "HDU " + hdu + " could not be found in FITS file '" 
                + fitsFilename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message); 
        }

        /* Extract the FITS header from the HDU. */
        Header header = selectedHDU.getHeader(); 
        
        /* Get a cursor/iterator for the header cards in the header. */ 
        Cursor iterator = header.iterator(); 

        TypedHeaderCard[] results = 
            new TypedHeaderCard[header.getNumberOfCards()]; 
        HeaderCard currentCard; 
        int i = 0;              /* The index of the current header card. */ 

        while(iterator.hasNext()) 
        {
            /* Get the current header card. */ 
            currentCard = (HeaderCard) iterator.next(); 

            /* Convert it into a TypedHeaderCard and store it for return. */ 
            try
            {
                results[i++] = new TypedHeaderCard(currentCard); 
            }
            catch(Exception e)
            {
                String message = 
                    "Could not use FITS header card '" + currentCard 
                    + "' to construct a TypedHeaderCard object."; 

                logger.error(message); 
                throw new TypedHeaderCardException(message, e); 
            }
        }

        return(results); 
    }

    /** Returns an array of TypedHeaderCards containing the requested header 
     * cards from the specified HDU. 
     * 
     * @param hdu the HDU to read the header cards from. 
     * @param requestedCards an array of Strings containing the keys of the 
     *                       header cards to be extracted from the HDU. 
     *
     * @throws FileHandlerException if the HDU could not be read. 
     * @throws TypedHeaderCardException if a FITS header card could not be 
     *                                  stored as a TypedHeaderCard. 
     */ 
    public TypedHeaderCard[] getFITSCards(int hdu, String[] requestedCards) 
        throws FileHandlerException, TypedHeaderCardException 
    {
        BasicHDU selectedHDU; 

        /* Get the specified Header + Data Unit. */
        try
        {
            selectedHDU = fitsFile.getHDU(hdu); 
        }
        catch(Exception e)
        {
            /* An error occurred whilst trying to retrieve the HDU - the 
             * HEASARC Java FITS documentation does not specify what can 
             * cause an exception at this point. Problems such as files 
             * being deleted whilst open are likely to show up here. 
             */
            String message = 
                "Could not retrieve HDU " + hdu + " from FITS file '" 
                + fitsFilename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message, e); 
        }

        if(selectedHDU == null)
        {
            /* The specified HDU could not be found. */ 
            String message = 
                "HDU " + hdu + " could not be found in FITS file '" 
                + fitsFilename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message); 
        }

        /* Extract the FITS header from the HDU. */
        Header header = selectedHDU.getHeader(); 

        TypedHeaderCard[] results = 
            new TypedHeaderCard[requestedCards.length]; 
        String currentKey;
        HeaderCard currentCard; 
        
        for(int i = 0; i < requestedCards.length; ++i)
        {
            /* Ensure that the current key is in long format with dot 
             * separators, as required by the HEASARC nom.tam.fits 
             * library classes. 
             */ 
            currentKey = 
                TypedHeaderCard.formatKey(requestedCards[i],
                                          TypedHeaderCard.LONG_DOTTED); 

            /* Find the header card corresponding to the current key. */
            currentCard = header.findCard(currentKey); 
            
            if(currentCard == null)
            {
                /* No header card could be found for that key - store an 
                 * undefined header card. Value and comment are left null. 
                 */
                try
                {
                    results[i] = new TypedHeaderCard(currentKey, false); 
                }
                catch(Exception e)
                {
                    String message = 
                        "Could not use undefined keyword '" + currentKey 
                        + "' to construct a TypedHeaderCard object " 
                        + "with type set to UNDEFINED."; 

                    logger.error(message); 
                    throw new TypedHeaderCardException(message, e); 
                }
            }
            else
            {
                /* A header card was found for that key - convert it to 
                 * a TypedHeaderCard & store it for return. 
                 */
                try
                {
                    results[i] = new TypedHeaderCard(currentCard); 
                }
                catch(Exception e)
                {
                    String message = 
                        "Could not use FITS header card '" + currentCard 
                        + "' to construct a TypedHeaderCard object."; 

                    logger.error(message); 
                    throw new TypedHeaderCardException(message, e); 
                }
            }
        }

        return(results); 
    }
    
	public int getNumberOfHDUs() throws FileHandlerException, TypedHeaderCardException {
		/* if the HDUs have never been read getNumberOFHDUs return 1, so the first 
		 * time we have to read all of them 
		 */
		if (!readAllHDUs) {
			try {
				int i = 0;
				while(fitsFile.getHDU(i++) != null);
			} catch (Exception e){
				readAllHDUs = true;
			}
		}
		return fitsFile.getNumberOfHDUs();
	}
}
