package org.eso.oca.fits;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import nom.tam.fits.FitsFactory;

import org.apache.log4j.Logger;

/** This class encapsulates an HDR file and provides convenience methods to 
 * applications using OCA, allowing them to be written with minimal knowledge 
 * of any libraries adopted by OCA's implementors. 
 *
 * @author Neil Ferguson ESO/DMD/DFS <nferguso@eso.org> 
 */
public class HDRHandler implements DataTransportFormatHandler
{
    static Logger logger = Logger.getLogger(HDRHandler.class);

    /** The index of the primary HDU in an HDR file. 
     *
     * @see #getFITSCards(int)
     * @see #getFITSCards(int,String[])
     */
    public static final int PRIMARY_HDU_INDEX = 0; 

    /** The HDR file held by this HDRHandler object. */ 
    private BufferedReader hdrFile; 

    /** The filename of the HDR file held by this HDRHandler object. */ 
    private String hdrFilename; 

    /** True if the contents of this HDR file have already been read into 
     * memory. 
     */ 
    private boolean headersInMemory = false; 

    /** An array of arrays of TypedHeaderCards, containing all of the 
     * TypedHeaderCards from this HDR file. The left index is the header 
     * index, and the right index is the index of a given TypedHeaderCard 
     * within that header. 
     */ 
    private TypedHeaderCard[][] headers; 

    /** An array of keyword string-TypedHeaderCard maps, where the index 
     * is the header index. These maps are generated as required by the 
     * getFITSCardMap(int) method. 
     */ 
    private Map[] headerMaps; 

    /** Constructs a HDRHandler to operate on the HDR file with the 
     * specified filename. 
     *
     * @throws FileHandlerException if the HDR file could not be opened. 
     */
    public HDRHandler(String filename) throws FileHandlerException 
    {
        /* Enable hierarchical keywords. */
        FitsFactory.setUseHierarch(true); 

        /* Open the HDR file. */ 
        try
        {
            hdrFile = new BufferedReader(new FileReader(filename)); 
        }
        catch(FileNotFoundException fnfe)
        {
            String message = "Could not open HDR file '" + filename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message, fnfe); 
        }

        hdrFilename = filename; 
    }

    /** Returns an array of arrays of TypedHeaderCards containing all of the 
     * header cards from this HDR file. 
     *  
     * @throws FileHandlerException if the HDR file could not be read. 
     * @throws TypedHeaderCardException if a FITS header card could not be 
     *                                  stored as a TypedHeaderCard. 
     */
    private TypedHeaderCard[][] getFITSCards() 
        throws FileHandlerException, TypedHeaderCardException 
    {
        if(!headersInMemory)
        {
            /* The contents of this HDR file need to be read into memory. */ 

            /* The list of headers read from the HDR file. */
            LinkedList headerList = new LinkedList(); 
            /* The list of header cards in the current header, */ 
            LinkedList currentHeader = new LinkedList(); 

            /* The current line in the HDR file. */ 
            String currentLine;
            /* The current header card from the HDR file. */ 
            TypedHeaderCard currentCard; 

            try
            {
                /* Read and process lines from the HDR file. */ 
                while((currentLine = hdrFile.readLine()) != null)
                {
                    if(currentLine.trim().equals(""))
                    {
                        /* Blank line read - skip to the next line. */ 
                        continue; 
                    }

                    /* Convert the current line into a TypedHeaderCard. */
                    try
                    {
                        currentCard = new TypedHeaderCard(currentLine); 
                    }
                    catch(Exception e)
                    {
                        String message = 
                            "Could not use FITS header card '" + currentLine 
                            + "' to construct a TypedHeaderCard object."; 

                        logger.error(message); 
                        throw new TypedHeaderCardException(message, e); 
                    }
                    
                    /* Add the TypedHeaderCard to the current header. */ 
                    currentHeader.add(currentCard); 

                    if(currentCard.getKey().equals("END"))
                    {
                        /* The end of this header has been detected. Store it 
                         * in the list of headers, then start a new header.     
                         */ 
                        headerList.add(currentHeader);
                        currentHeader = new LinkedList(); 
                    }
                }
            }
            catch(IOException ioe) 
            {
                /* An IO error occurred whilst reading the HDR file. */ 
                String message = "An IO error occurred whilst reading from " 
                    + "HDR file '" + hdrFilename + "'."; 

                logger.error(message); 
                throw new FileHandlerException(message, ioe); 
            }

            /* Store the newly-read data as arrays of TypedHeaderCards. */

            /* First, create an array to hold an array of TypedHeaderCards 
             * for each header. 
             */
            int numHeaders = headerList.size(); 
            headers = new TypedHeaderCard[numHeaders][];

            /* Second, store the array of TypedHeaderCards for each header. */
            int i = 0;
            Iterator iterator = headerList.iterator(); 
            TypedHeaderCard[] modelArray = new TypedHeaderCard[0]; 

            while(iterator.hasNext())
            {
                /* Convert the next list of TypedHeaderCards to an array of 
                 * TypedHeaderCards and store it. 
                 */ 
                headers[i++] = 
                    (TypedHeaderCard[]) 
                    ((LinkedList) iterator.next()).toArray(modelArray); 
            }

            /* Make array space for storing keyword string-TypedHeaderCard 
             * maps, should these be created later by getFITSCardMap(int). 
             */
            headerMaps = new Map[numHeaders]; 

            /* The HDR file's contents have all been read into memory - flag 
             * that and dispose of the now-redundant BufferedReader. 
             */ 
            headersInMemory = true; 
            hdrFile = null; 
        }

        return headers; 
    }

    /** Returns an array of TypedHeaderCards containing all of the header 
     * cards from the specified header. 
     * 
     * @param header the header whose header cards should be returned. 
     * 
     * @throws FileHandlerException if the header could not be read. 
     * @throws TypedHeaderCardException if a FITS header card could not be 
     *                                  stored as a TypedHeaderCard. 
     */ 
    public TypedHeaderCard[] getFITSCards(int header) 
        throws FileHandlerException, TypedHeaderCardException 
    {
        if(!(header < getFITSCards().length)) 
        {
            /* The specified header could not be found. */ 
            String message = 
                "Header " + header + " could not be found in HDR file '" 
                + hdrFilename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message); 
        }

        /* Return the TypedHeaderCards from the specified header. */ 
        return(getFITSCards()[header]); 
    }

    /** Returns an array of TypedHeaderCards containing the requested header 
     * cards from the specified header. 
     * 
     * @param header the header to read the header cards from. 
     * @param requestedCards an array of Strings containing the keys of the 
     *                       header cards to be extracted from the header. 
     *
     * @throws FileHandlerException if the header could not be read. 
     * @throws TypedHeaderCardException if a FITS header card could not be 
     *                                  stored as a TypedHeaderCard. 
     */ 
    public TypedHeaderCard[] getFITSCards(int header, String[] requestedCards)
        throws FileHandlerException, TypedHeaderCardException 
    {
        if(!(header < getFITSCards().length)) 
        {
            /* The specified header could not be found. */ 
            String message = 
                "Header " + header + " could not be found in HDR file '" 
                + hdrFilename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message); 
        }

        /* Get the header cards for the specified header in an efficiently 
         * searchable container. 
         */ 
        Map headerCards = getFITSCardMap(header); 

        TypedHeaderCard[] results = 
            new TypedHeaderCard[requestedCards.length]; 
        String currentKey;
        TypedHeaderCard currentCard; 
        
        for(int i = 0; i < requestedCards.length; ++i)
        {
            /* Ensure that the current key is in long format with dot 
             * separators, as required by the HEASARC nom.tam.fits 
             * library classes. 
             */ 
            currentKey = 
                TypedHeaderCard.formatKey(requestedCards[i],
                                          TypedHeaderCard.LONG_DOTTED); 

            /* Find the header card corresponding to the current key. */
            currentCard = (TypedHeaderCard) headerCards.get(currentKey); 
            
            if(currentCard == null)
            {
                /* No header card could be found for that key - store an 
                 * undefined header card. Value and comment are left null. 
                 */
                try
                {
                    results[i] = new TypedHeaderCard(currentKey, false); 
                }
                catch(Exception e)
                {
                    String message = 
                        "Could not use undefined keyword '" + currentKey 
                        + "' to construct a TypedHeaderCard object " 
                        + "with type set to UNDEFINED."; 

                    logger.error(message); 
                    throw new TypedHeaderCardException(message, e); 
                }
            }
            else
            {
                /* A matching header card was found for that key - store it 
                 * for return. 
                 */
                results[i] = currentCard; 
            }
        }

        return(results); 
    }

    /** Returns a Map of keyword strings to TypedHeaderCards, containing 
     * an entry for each of the header cards in the specified header. 
     *
     * @param header the header whose header cards should be returned. 
     * 
     * @throws FileHandlerException if the header could not be read. 
     * @throws TypedHeaderCardException if a FITS header card could not be 
     *                                  stored as a TypedHeaderCard. 
     */
    private Map getFITSCardMap(int header) 
        throws FileHandlerException, TypedHeaderCardException 
    {
        if(!(header < getFITSCards().length)) 
        {
            /* The specified header could not be found. */ 
            String message = 
                "Header " + header + " could not be found in HDR file '" 
                + hdrFilename + "'."; 

            logger.error(message); 
            throw new FileHandlerException(message); 
        }

        if(headerMaps[header] == null) 
        {
            /* No keyword string-TypedHeaderCard map has been generated for 
             * this header yet - generate one now. 
             */

            /* Make a new map (using Hashtable) with a suitable capacity. */ 
            headerMaps[header] = 
                new Hashtable((int)(getFITSCards().length / 0.75) + 1); 

            /* Get the array of TypedHeaderCards for this header. */ 
            TypedHeaderCard[] headerCards = getFITSCards(header); 

            /* Insert each header card into the map. */ 
            for(int i = 0; i < headerCards.length; ++i)
            {
                /* The keyword name (in long dotted format) is used as the 
                 * key for the corresponding TypedHeaderCard object. 
                 */
                headerMaps[header].put(headerCards[i].getKey(), 
                                       headerCards[i]); 
            }
        }

        return(headerMaps[header]); 
    }

	public int getNumberOfHDUs() throws FileHandlerException, TypedHeaderCardException {
		return getFITSCards().length;
	}
}
