package org.eso.util.dal;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Hashtable;
import java.util.List;

import org.apache.commons.dbcp.BasicDataSource;
import org.apache.log4j.Logger;

/**
 * This class allows to execute SQL statements by means of connection pools.
 * 
 * Note: if you need to specify properties such as the character set, you can do
 * it on the database server URL.
 * 
 * E.g.: //wu0ola.hq.eso.org:5000?CHARSET=utf8 <- use UTF-8 as char set
 * 
 * @author mpadovan
 * 
 */
public class SybaseConnectionService {

    /** Apache Log4J logger. */
    static final Logger logger =
            Logger.getLogger(SybaseConnectionService.class);

    /** Used to format times for Sybase up to the milliseconds. */
    public static final SimpleDateFormat dateFormat =
            new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");

    /**
     * Holds the available data sources.
     * 
     * When requesting a connection from a data source, that connection is
     * returned from a pool of connections.
     * 
     * The key is a String that uniquely identify the data source.
     */
    private static Hashtable<String, BasicDataSource> dataSources =
            new Hashtable<String, BasicDataSource>();

    /**
     * Hides the constructor. This class cannot be instantiated.
     */
    private SybaseConnectionService() {
        super();
    }

    /**
     * Returns a connection to the specified data base.
     * 
     * @param serverUrl
     *            The data base server URL.
     * @param userName
     *            The user name.
     * @param userPassword
     *            The user password.
     * @return The connection.
     * @throws IllegalArgumentException
     *             should any of the arguments be null or not valid.
     * @throws ConnectionServiceException
     *             in case of data base errors.
     */
    public static Connection getConnection(String serverUrl, String userName,
            String userPassword) throws IllegalArgumentException,
            ConnectionServiceException {

        testStringArg(serverUrl,
                "The data base server URL cannot be null or empty.");
        testStringArg(userName, "The user name cannot be null or empty.");
        testStringArg(userPassword, "The password cannot be null or empty.");

        final String logHead = "SybaseConnectionService::getInstance - ";

        // create a key to identify the connection pool
        final String connectionPoolName = serverUrl.concat(userName);

        BasicDataSource dataSource = null;

        String debugMessage;

        synchronized (dataSources) {

            if (connectionPoolExists(connectionPoolName)) {

                // the connection pool already exists, get it from the list

                dataSource =
                        SybaseConnectionService.getConnectionPool(connectionPoolName);

                debugMessage =
                        "Using existing connection pool for data base server ["
                                + serverUrl + "], user [" + userName + "].";

            } else {

                // the connection pool does not exist, create it and add it to
                // the list

                dataSource = new BasicDataSource();

                /* new driver: disabled for testing */
                //dataSource.setDriverClassName("net.sourceforge.jtds.jdbc.Driver");
                //dataSource.setUrl("jdbc:jtds:sybase:" + serverUrl);
                
                /* old driver */
                dataSource.setDriverClassName("com.sybase.jdbc2.jdbc.SybDriver");
                dataSource.setUrl("jdbc:sybase:Tds:" + serverUrl);
                
                dataSource.setUsername(userName);
                dataSource.setPassword(userPassword);
                
                /* a connection can be removed after being idle for at least an hour */
                dataSource.setMinEvictableIdleTimeMillis(3600*1000);
                /* look for unused connections every 10 minutes */
                dataSource.setTimeBetweenEvictionRunsMillis(60*10);
                /* enforce object validation */
                dataSource.setTestOnBorrow(true);
                dataSource.setTestWhileIdle(true);
                
                dataSource.setMaxActive(4);
                dataSource.setMaxIdle(4);
                
                SybaseConnectionService.addConnectionPool(connectionPoolName,
                        dataSource);

                debugMessage =
                        "New connection pool created for data base server ["
                                + serverUrl + "], user [" + userName + "].";
            }
        }

        logger.debug(logHead + debugMessage);

        Connection connection = null;
        try {
            connection = dataSource.getConnection();
            logger.debug(logHead + "Connected to server [" + serverUrl
                    + "], user [" + userName + "].");
        } catch (SQLException e) {
            throw new ConnectionServiceException(logHead
                    + "Cannot get connection to data base server [" + serverUrl
                    + "], user [" + userName + "]. Reason: " + e.getMessage());
        }

        return connection;
    }

    /**
     * Executes an SQL statement.
     * 
     * @param connection
     *            The connection to be used to execute the SQL statement.
     *            It is highly recommended to close the connection when
     *            finished with the ResultSet object.
     * @param sqlStatement
     *            The SQL statement to executed.
     * @return A result set containing the output of the SQL statement. The
     *         result set is null if the statement does not return anything.
     * @throws IllegalArgumentException
     *             should any of the arguments be null or not valid.
     * @throws ConnectionServiceException
     *             in case of data base errors.
     */
    public static ResultSet executeStatement(final Connection connection,
            final String sqlStatement) throws IllegalArgumentException,
            ConnectionServiceException {

        testArg(connection, "The connection cannot be null.");
        testStringArg(sqlStatement,
                "The sql statement cannot be null or empty.");

        final String logHead = "SybaseConnectionService::executeStatement - ";

        ResultSet resultSet = null;

        try {

            Statement statement = connection.createStatement();

            statement.execute(sqlStatement);

            resultSet = statement.getResultSet();
            //resultSet.setFetchSize(statement.getMaxRows());

            logger.debug(logHead + "SQL statement [" + sqlStatement
                    + "] successfully executed.");

        } catch (SQLException e) {

            final String errorMessage =
                    "Error while executing SQL statement [" + sqlStatement
                            + "]. Reason: " + e.getMessage();
            logger.error(logHead + errorMessage);
            throw new ConnectionServiceException(errorMessage);

        } catch (Exception e) {

            final String errorMessage =
                    "Unexpected error while executing SQL statement ["
                            + sqlStatement + "]. Reason: " + e.getMessage();
            logger.error(logHead + errorMessage);
            throw new ConnectionServiceException(errorMessage);

        } finally {

            // The connection should be closed here, but if you do that,
            // the result set will become invalid !
            // For this reason, the user must close the connection manually
        	// when finished with the result set.
            //
            // closeConnection(connection);
        }

        return resultSet;
    }

    /**
     * Executes an SQL statement that modifies the database content, like update
     * or delete.
     * 
     * @param connection
     *            The connection to be used to execute the SQL statement.
     *            It is highly recommended to close the connection when
     *            finished with the ResultSet object.
     * @param sqlStatement
     *            The SQL statement to executed.
     * @return the number of updated rows
     * @throws IllegalArgumentException
     *             should any of the arguments be null or not valid.
     * @throws ConnectionServiceException
     *             in case of data base errors.
     */
    public static int executeUpdateStatement(final Connection connection,
            final String sqlStatement) throws IllegalArgumentException,
            ConnectionServiceException {

        testArg(connection, "The connection cannot be null.");
        testStringArg(sqlStatement,
                "The sql statement cannot be null or empty.");
        int updatedRowNb = 0;

        final String logHead = "SybaseConnectionService::executeStatement - ";

        try {

            Statement statement = connection.createStatement();

            statement.execute(sqlStatement);

            updatedRowNb = statement.getUpdateCount();

            logger.debug(logHead + "SQL statement [" + sqlStatement
                    + "] successfully executed " + updatedRowNb
                    + " rows updated ");

        } catch (SQLException e) {

            final String errorMessage =
                    "Error while executing SQL statement [" + sqlStatement
                            + "]. Reason: " + e.getMessage();
            logger.error(logHead + errorMessage);
            updatedRowNb = -1;
            throw new ConnectionServiceException(errorMessage);

        } catch (Exception e) {

            final String errorMessage =
                    "Unexpected error while executing SQL statement ["
                            + sqlStatement + "]. Reason: " + e.getMessage();
            logger.error(logHead + errorMessage);
            updatedRowNb = -1;
            throw new ConnectionServiceException(errorMessage);

        } finally {

            // Although the connection could be closed here, it is not: this
        	// way both "execute" methods work in the same way by leaving the
        	// user the freedom/task to close the connection. Moreover,
        	// keeping the connection open allows for transactions.
        }

        return updatedRowNb;
    }

    /**
     * Executes a batched SQL statement.
     * 
     * @param connection
     *            The connection to be used to execute the SQL statement.
     *            It is highly recommended to close the connection when
     *            finished with the ResultSet object.
     * @param sqlStatements
     *            A list of SQL statements which will be executed at one time.
     * @return A result set containing the output of the SQL statement. The
     *         result set is null if the statement does not return anything.
     * @throws IllegalArgumentException
     *             should any of the arguments be null or not valid.
     * @throws ConnectionServiceException
     *             in case of data base errors.
     */
    public static ResultSet executeBatchedStatement(final Connection connection,
            final List<String> sqlStatements) throws IllegalArgumentException,
            ConnectionServiceException {

        testArg(connection, "The connection cannot be null.");
        testStringListArg(sqlStatements,
                "The sql statements cannot be null or empty.");

        final String logHead = "SybaseConnectionService::executeStatement - ";

        ResultSet resultSet = null;

        try {

            Statement statement = connection.createStatement();

            for (String query: sqlStatements) {
            	statement.addBatch(query);
            }
            
            statement.executeBatch();

            resultSet = statement.getResultSet();
       
            logger.debug(logHead + "SQL statement [" + sqlStatements
                    + "] successfully executed.");

        } catch (SQLException e) {

            final String errorMessage =
                    "Error while executing SQL statement [" + sqlStatements
                            + "]. Reason: " + e.getMessage();
            logger.error(logHead + errorMessage);
            throw new ConnectionServiceException(errorMessage);

        } catch (Exception e) {

            final String errorMessage =
                    "Unexpected error while executing SQL statement ["
                            + sqlStatements + "]. Reason: " + e.getMessage();
            logger.error(logHead + errorMessage);
            throw new ConnectionServiceException(errorMessage);

        } finally {

            // The connection should be closed here, but if you do that,
            // the result set will become invalid !
            // For this reason, the user must close the connection manually
        	// when finished with the result set.
            //
            // closeConnection(connection);
        }

        return resultSet;
    }
    
    /**
     * Closes the given connection.
     * 
     * @param connection
     *            The connection to close.
     * @param logHead
     *            The header to be used for logging debug and error information.
     * @throws ConnectionServiceException
     *             in case of errors.
     */
    public static void closeConnection(Connection connection)
            throws ConnectionServiceException {

        final String logHead = "SybaseConnectionService::closeConnection() - ";

        try {
            if (connection != null) {
                connection.close();
                logger.debug(logHead + "Released connection to the database.");
            }
        } catch (Exception e) {
            final String errorMessage =
                    "Cannot release connection to the database. Reason: "
                            + e.getMessage();
            logger.error(logHead + errorMessage);
            throw new ConnectionServiceException(errorMessage);
        }

    }

    /**
     * Returns true if the data source with the specified name exists.
     */
    private static boolean connectionPoolExists(final String connectionPoolName) {
        return SybaseConnectionService.dataSources.containsKey(connectionPoolName.toLowerCase());
    }

    /**
     * Add a new data source with the specified name.
     */
    private static void addConnectionPool(final String connectionPoolName,
            final BasicDataSource dataSource) {
        SybaseConnectionService.dataSources.put(connectionPoolName.toLowerCase(),
                dataSource);
    }

    /**
     * Returns the data source with the specified name.
     */
    private static BasicDataSource getConnectionPool(
            final String connectionPoolName) {
        return SybaseConnectionService.dataSources.get(connectionPoolName.toLowerCase());
    }

    /**
     * Checks if the string is null or empty.
     * 
     * @throws IllegalArgumentException
     *             when something is wrong.
     */
    private static void testStringArg(final String arg,
            final String errorMessage) throws IllegalArgumentException {

        final String logHead = "SybaseConnectionService::testStringArg - ";

        if (!(arg != null && arg.length() > 0)) {
            logger.error(logHead + errorMessage);
            throw new IllegalArgumentException(errorMessage);
        }
    }
    
    /**
     * Checks if the string list is null or contains empty strings.
     * 
     * @throws IllegalArgumentException
     *             when something is wrong.
     */
    private static void testStringListArg(final List<String> arg,
            final String errorMessage) throws IllegalArgumentException {

        final String logHead = "SybaseConnectionService::testStringArg - ";

        if (arg == null) {
        	logger.error(logHead + errorMessage);
        	throw new IllegalArgumentException(errorMessage);
        } else {
        	for (String s: arg) {
        		if (!(s != null && s.length() > 0)) {
        			logger.error(logHead + errorMessage);
        			throw new IllegalArgumentException(errorMessage);
        		}
        	}
        }
    }

    /**
     * Checks if the argument is null.
     * 
     * @throws IllegalArgumentException
     *             when something is wrong.
     */
    private static void testArg(final Object arg, final String errorMessage)
            throws IllegalArgumentException {

        final String logHead = "SybaseConnectionService::testArg - ";

        if (arg == null) {
            logger.error(logHead + errorMessage);
            throw new IllegalArgumentException(errorMessage);
        }
    }

}
