package org.eso.util.dal;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eso.util.dal.TransferRequest.Status;
import org.eso.util.filesystem.FileSystemOperations;

/**
 * Implementation of a TransferRequestDAO when transfer requests are implemented
 * as files
 * 
 * @author cmoins
 * 
 */
public class TransferRequestDAOFile implements TransferRequestDAO {

    /** Apache Log4J logger. */
    static final Logger logger = Logger.getLogger(TransferRequestDAOFile.class);
    /** The name of this class for logging purposes. */
    private static final String classLogName = "TransferRequestDAOFile";

    /** Mandatory keyword indicating the beginning of a transfer request file */
    private static final String BEGIN = "TRANSFER.REQUEST.BEGIN";
    /** Mandatory keyword indicating the end of a transfer request file */
    private static final String END = "TRANSFER.REQUEST.END";
    /** Line separator */
    private static final String EOL = System.getProperty("line.separator");

    /** Status to directory map */
    Map<TransferRequest.Status, File> statusDirMap = null;

    /** Available keywords in a transfer request file */
    final String FILEID = "FILEID";
    final String FILENAME = "FILENAME";
    final String UNCOMPRESSED_FILENAME = "UNCOMPRESSED.FILENAME";
    final String TRANSFER_METHOD = "TRANSFER.METHOD";
    final String TRANSFER_PRIORITY = "TRANSFER.PRIORITY";
    final String COMPRESSION_METHOD = "COMPRESSION.METHOD";
    final String SIZE = "SIZE";
    final String UNCOMPRESSED_SIZE = "UNCOMPRESSED.SIZE";
    final String CHECKSUM = "CHECKSUM";
    final String TRANSFER_CATG = "TRANSFER.CATG";
    final String USER = "USER";
    final String DESTDIR = "DESTDIR";
    final String MULTI_FILE_COMPONENTS = "MULTI.FILE.COMPONENTS";
    final String CREATION_TIME = "CREATION.TIME";

    /** Directory where incoming transfer requests are stored */
    final File incomingDirectory;
    /** Directory where being processed transfer requests are stored */
    final File processingDirectory;
    /** Directory where processed transfer requests are stored */
    final File processedDirectory;
    /** Directory where rejected transfer requests are stored */
    final File rejectedDirectory;

    /** File extension of the transfer requests */
    String transferRequestExtension;

    /** Transfer Request Validator */
    TransferRequestValidator transferRequestValidator = null;

    /** Used to format polling times in log messages. */
    private final SimpleDateFormat dateFormat =
            new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS");

    /**
     * Constructor It will set the incoming, processing, processed and rejected
     * directories and the transfer request file extension.
     * 
     * @param incomingDirectory
     *            directory where Status.INCOMING transfer requests are stored
     * @param processingDirectory
     *            directory where Status.PROCESSING transfer requests are stored
     * @param processedDirectory
     *            directory where Status.PROCESSED transfer requests are stored
     * @param rejectedDirectory
     *            directory where Status.REJECTED transfer requests are stored
     * @param transferRequestExtension
     *            transfer request file extension
     * @param transferRequestValidator
     *            transfer request validator
     * @throws DAOException
     */
    public TransferRequestDAOFile(File incomingDirectory,
            File processingDirectory, File processedDirectory,
            File rejectedDirectory, String transferRequestExtension,
            TransferRequestValidator transferRequestValidator)
            throws DAOException {
        if ((incomingDirectory == null) || (processingDirectory == null)
                || (processedDirectory == null) || (rejectedDirectory == null)
                || (transferRequestExtension == null)
                || (transferRequestValidator == null)) {
            String message =
                    "Missing parameter to instantiate a TransferRequestDAOFile object";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }

        if ((incomingDirectory.exists() == false)
                || (incomingDirectory.isDirectory() == false)) {
            String message = incomingDirectory + " does not exist ";
            logger.error(message);
            throw new DAOException(message);
        }
        if ((processingDirectory.exists() == false)
                || (processingDirectory.isDirectory() == false)) {
            String message = processingDirectory + " does not exist ";
            logger.error(message);
            throw new DAOException(message);
        }
        if ((processedDirectory.exists() == false)
                || (processedDirectory.isDirectory() == false)) {
            String message = processedDirectory + " does not exist ";
            logger.error(message);
            throw new DAOException(message);
        }
        if ((rejectedDirectory.exists() == false)
                || (rejectedDirectory.isDirectory() == false)) {
            String message = rejectedDirectory + " does not exist ";
            logger.error(message);
            throw new DAOException(message);
        }
        this.incomingDirectory = incomingDirectory;
        this.processingDirectory = processingDirectory;
        this.processedDirectory = processedDirectory;
        this.rejectedDirectory = rejectedDirectory;
        this.transferRequestValidator = transferRequestValidator;

        statusDirMap = new HashMap<TransferRequest.Status, File>();
        statusDirMap.put(TransferRequest.Status.INCOMING, incomingDirectory);
        statusDirMap.put(TransferRequest.Status.PROCESSING, processingDirectory);
        statusDirMap.put(TransferRequest.Status.PROCESSED, processedDirectory);
        statusDirMap.put(TransferRequest.Status.REJECTED, rejectedDirectory);

        this.transferRequestExtension = transferRequestExtension;
    }

    /**
     * Constructor It will set the incoming directory and the transfer request
     * file extension.
     * 
     * @param incomingDirectory
     *            directory where incoming transfer requests are stored
     * @param transferRequestExtension
     *            transfer request file extension
     * @throws DAOException
     */
    public TransferRequestDAOFile(File incomingDirectory,
            String transferRequestExtension,
            TransferRequestValidator transferRequestValidator)
            throws DAOException {
        if ((incomingDirectory == null) || (transferRequestExtension == null)) {
            String message =
                    "Missing parameter to instantiate a TransferRequestDAOFile object";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }

        if ((incomingDirectory.exists() == false)
                || (incomingDirectory.isDirectory() == false)) {
            String message = incomingDirectory + " does not exist ";
            logger.error(message);
            throw new DAOException(message);
        }

        this.incomingDirectory = incomingDirectory;
        this.processingDirectory = null;
        this.processedDirectory = null;
        this.rejectedDirectory = null;

        statusDirMap = new HashMap<TransferRequest.Status, File>();
        statusDirMap.put(TransferRequest.Status.INCOMING, incomingDirectory);

        this.transferRequestExtension = transferRequestExtension;

        this.transferRequestValidator = transferRequestValidator;
    }

    /**
     * Finds all the transfer requests in a given status. Note : It will up to
     * the user to check the validity of the transfer requests
     * 
     * @param status
     *            status of the transfer requests to be retrieved
     * @return all the transfer requests in the given status
     */
    public List<TransferRequest> findByStatus(TransferRequest.Status status)
            throws DAOException {
        final String methodLogName = classLogName + "::findByStatus()";
        logger.trace(methodLogName);

        if (status == null) {
            String message = "Status is missing";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }

        // Only readable files with the transfer request file extension will be
        // selected
        FileFilter filter = new FileFilter() {
            public boolean accept(File file) {
                boolean accepted =
                        file.getName().endsWith(transferRequestExtension);
                if (accepted && !file.canRead()) {
                    logger.warn(methodLogName + " - Can not read file "
                            + file.getAbsolutePath());
                    accepted = false;
                }
                return accepted;
            }
        };
        File statusDir = statusDirMap.get(status);
        if (statusDir == null) {
            String message =
                    methodLogName + " - No directory associated to status "
                            + status;
            logger.error(message);
            throw new DAOException(message);
        }
        File fileArray[] = statusDir.listFiles(filter);
        List<TransferRequest> listFile = new ArrayList<TransferRequest>();
        TransferRequest curTransferRequest = null;
        for (int i = 0; i < fileArray.length; ++i) {
            try {
                curTransferRequest = load(fileArray[i]);

            } catch (DAOException daoe) {
                curTransferRequest = null;
            } finally {
                if (curTransferRequest != null) {
                    curTransferRequest.setStatus(status);
                    listFile.add(curTransferRequest);
                } else {
                    logger.error("Could not build a transfer request from "
                            + fileArray[i]);

                    // Moves the transfer request to the directory associated
                    // to the specified status
                    File destDir = statusDirMap.get(Status.REJECTED);
                    if (destDir == null) {
                        String message =
                                "No directory associated to " + Status.REJECTED;
                        logger.error(message);
                        throw new DAOException(message);
                    }
                    if (FileSystemOperations.moveToDir(fileArray[i], destDir)) {
                        logger.info(methodLogName + " - " + fileArray[i]
                                + " moved to "
                                + statusDirMap.get(Status.REJECTED));
                    } else {
                        logger.error(methodLogName + " - " + " Could not move "
                                + fileArray[i].getAbsolutePath() + "to "
                                + statusDirMap.get(Status.REJECTED));
                    }
                }
            }
        }
        return listFile;
    }

    /**
     * Updates the status of a specified transfer request It will move the file
     * implementing the transfer request in the directory associated to the
     * specified status
     * 
     * @param transferRequest
     *            transfer request whose status should be updated
     * @param status
     *            new status of the transfer request
     * @return the number of updated files
     * @throws DAOException
     */
    public int updateStatus(TransferRequest transferRequest,
            TransferRequest.Status status) throws DAOException {

        final String methodLogName = classLogName + "::updateStatus()";
        logger.trace(methodLogName);
        int nbUpdatedFiles = 0;
        if (transferRequest == null) {
            String message = methodLogName + " No transfer request available";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }
        if (status == null) {
            String message = "No status value available";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }
        if (transferRequest.getRequestId() == null) {
            String message = "Transfer request should get an identifier";
            logger.error(message);
            throw new DAOException(message);
        }
        String trFileName = transferRequest.getRequestId();
        TransferRequest.Status trStatus = transferRequest.getStatus();
        if (trStatus == null) {
            String message =
                    "No status associated to transfer request " + trFileName;
            logger.error(message);
            throw new DAOException(message);
        }

        /** gets the current directory where the transfer request is stored */
        File currentDir = statusDirMap.get(trStatus);
        if (currentDir == null) {
            String message =
                    "No directory associated to status " + trStatus
                            + " for transfer request " + trFileName;
            logger.error(message);
            throw new DAOException(message);
        }

        File transferRequestFile =
                new File(currentDir.getAbsolutePath() + File.separatorChar
                        + trFileName);

        if (!transferRequestFile.exists()) {
            String message = transferRequestFile + " does not exist";
            logger.error(message);
            throw new DAOException(message);
        }
        // Moves the transfer request to the directory associated
        // to the specified status
        File destDir = statusDirMap.get(status);
        if (destDir == null) {
            String message = "No directory associated to " + status;
            logger.error(message);
            throw new DAOException(message);
        }
        if (FileSystemOperations.moveToDir(transferRequestFile, destDir)) {
            transferRequest.setStatus(status);
            logger.debug(methodLogName + " - "
                    + transferRequestFile.getAbsolutePath() + " moved to "
                    + statusDirMap.get(status));
            nbUpdatedFiles = 1;
        } else {
            logger.error(methodLogName + " - " + " Could not move "
                    + transferRequestFile.getAbsolutePath() + " to "
                    + statusDirMap.get(status));
            nbUpdatedFiles = 0;
        }
        return nbUpdatedFiles;
    }

    /**
     * Saves a transfer request by creating a file whose name will be equal to
     * file name + transfer request file extension. Note : it first creates a
     * temporary file in the incoming directory and renames it afterwards with
     * the specified name to prevent any race condition. The specified transfer
     * request should also be valid
     * 
     * @param TransferRequest
     *            transfer request to be saved
     * @return the number of saved files
     * @throws DAOException
     */
    public int save(TransferRequest transferRequest) throws DAOException {
        final String TEMP_FILE_EXT = ".tmp";
        String methodLogName = classLogName + "::save()";
        int nbSavedTransferRequest = 0;
        // Check input parameter
        if (transferRequest == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - transferRequest must not be null");
        }
        // The transfer request should be a valid one
        if (!transferRequestValidator.isValid(transferRequest)) {
            String message =
                    methodLogName
                            + " - Cannot save a non valid TransferRequest.";
            logger.error(message);
            throw new DAOException(message);
        }

        // The transfer request file will be stored in the incoming directory
        // whose existence is checked in the constructor.
        // Its name will be made of the file id value of the transfer request +
        // the transfer request file extension
        File trDir = statusDirMap.get(TransferRequest.Status.INCOMING);
        String fileName = transferRequest.getFileName();

        if (fileName == null) {
            String message =
                    "Could not build a file name as file name field is null";
            logger.error(message);
            throw new DAOException(message);
        }
        File proviFile = new File(fileName);
        fileName =
                trDir.getAbsolutePath() + File.separatorChar
                        + proviFile.getName() + transferRequestExtension;

        // First a temporary file is created where all the metadata will be
        // stored
        String tempFileName = fileName + TEMP_FILE_EXT;
        FileWriter trFileWriter;
        try {
            trFileWriter = new FileWriter(new File(tempFileName));
            trFileWriter.write(BEGIN);
            trFileWriter.write(EOL);

            String key;
            String value;
            String currentCard;
            if (transferRequest.getFileId() != null) {
                key = FILEID;
                value = transferRequest.getFileId();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getFileName() != null) {
                key = FILENAME;
                value = transferRequest.getFileName();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getUncompressedFileName() != null) {
                key = UNCOMPRESSED_FILENAME;
                value = transferRequest.getUncompressedFileName();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getTransferMethod() != null) {
                key = TRANSFER_METHOD;
                value = transferRequest.getTransferMethod();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getTransferPriority() != null) {
                key = TRANSFER_PRIORITY;
                value = String.valueOf(transferRequest.getTransferPriority());
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getCompressionMethod() != null) {
                key = COMPRESSION_METHOD;
                value = transferRequest.getCompressionMethod();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getSize() != null) {
                key = SIZE;
                value = String.valueOf(transferRequest.getSize());
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getUncompressedSize() != null) {
                key = UNCOMPRESSED_SIZE;
                value = String.valueOf(transferRequest.getUncompressedSize());
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getChecksum() != null) {
                key = CHECKSUM;
                value = String.valueOf(transferRequest.getChecksum());
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getTransferCatg() != null) {
                key = TRANSFER_CATG;
                value = transferRequest.getTransferCatg();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getUser() != null) {
                key = USER;
                value = transferRequest.getUser();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getDestDir() != null) {
                key = DESTDIR;
                value = transferRequest.getDestDir();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }
            if (transferRequest.getMultiFileComponents() != null) {
                key = MULTI_FILE_COMPONENTS;
                value = transferRequest.getMultiFileComponents();
                currentCard = key + " = " + value;
                trFileWriter.write(currentCard);
                trFileWriter.write(EOL);
            }

            // Set the transfer request file creation
            key = CREATION_TIME;
            Long creationTime = System.currentTimeMillis();
            currentCard = key + " = " + dateFormat.format(creationTime);
            trFileWriter.write(currentCard);
            transferRequest.setCreationTime(creationTime);
            trFileWriter.write(EOL);

            trFileWriter.write(END);
            trFileWriter.close();
        } catch (IOException ioe) {
            String message =
                    "Could not open TRQ temporary file " + tempFileName
                            + " IOException " + ioe.getMessage();
            logger.error(message);
            throw new DAOException(message);
        }
        // The temporary file is renamed to the specified file name
        File transferRequestFile = new File(tempFileName);
        File destFile = new File(fileName);
        if (transferRequestFile.renameTo(destFile) == true) {
            transferRequest.setStatus(TransferRequest.Status.INCOMING);
            transferRequest.setRequestId(destFile.getName());
            logger.debug(destFile.getAbsolutePath() + " successfully created");
            nbSavedTransferRequest = 1;
        } else {
            String message =
                    "Could not rename " + tempFileName + " to " + fileName;
            logger.error(message);
            transferRequest.setStatus(TransferRequest.Status.REJECTED);
            throw new DAOException(message);
        }
        return nbSavedTransferRequest;
    }

    /**
     * Deletes a transfer requests from the repository. It assumes that the this
     * file is stored in the directory associated to the current status of the
     * transfer request.
     * 
     * @param TransferRequest
     *            transfer request to be saved
     * @returns the number of deleted files
     * @throws DAOException
     */
    public int delete(TransferRequest transferRequest) throws DAOException {
        String thisMethod = classLogName + "::delete()";
        int nbDeletedTransferRequests = 0;
        // Check input parameter
        if (transferRequest == null) {
            throw new IllegalArgumentException(thisMethod
                    + " - transferRequest must not be null");
        }

        // The transfer request file will be stored in the directory associated
        // to the transfer request status. Its name will be made of the file id
        // value of the transfer request + the transfer request file extension
        File trDir = statusDirMap.get(transferRequest.getStatus());
        if (trDir == null) {
            String message =
                    thisMethod + " - Status not valid for transfer request";
            logger.error(message);
            throw new DAOException(message);
        }
        String fileName = transferRequest.getRequestId();
        if (fileName == null) {
            String message = "No identifier for the transfer request";
            logger.error(message);
            throw new DAOException(message);
        }
        File trFile =
                new File(trDir.getAbsolutePath() + File.separatorChar
                        + fileName);
        if (trFile.exists() == false) {
            String message =
                    thisMethod + " - file " + trFile + " does not exist";
            logger.error(message);
            throw new DAOException(message);
        }
        if (FileSystemOperations.delete(trFile) == true) {
            logger.debug(trFile + " successfully deleted");
            nbDeletedTransferRequests = 1;
        } else {
            String message = "Could not delete " + trFile;
            logger.error(message);
            throw new DAOException(message);
        }
        return nbDeletedTransferRequests;
    }

    /**
     * checks if a transfer request is valid or not
     * 
     * @param transferRequest
     *            transfer request to be validated
     * @return true if the transfer request is valid, false otherwise
     */
    public Boolean isValid(TransferRequest transferRequest) {

        String methodLogName = classLogName + "::isValid";

        // Check input parameter
        if (transferRequest == null) {
            throw new IllegalArgumentException(methodLogName
                    + " - transferRequest must not be null");
        }
        return transferRequestValidator.isValid(transferRequest);
    }

    private TransferRequest load(File transferRequestFile) throws DAOException {
        String methodLogName = classLogName + "::load";
        logger.trace(methodLogName);

        // Transfer request object that shall be instantiated relying on the
        // transfer request file
        TransferRequest transferRequest = new TransferRequest();
        transferRequest.setRequestId(transferRequestFile.getName());

        BufferedReader trFileReader;
        try {
            logger.debug(methodLogName + " - Opening file "
                    + transferRequestFile);
            trFileReader =
                    new BufferedReader(new FileReader(transferRequestFile));
        } catch (FileNotFoundException fnfe) {
            String message =
                    "Could not open file " + transferRequestFile + " : "
                            + fnfe.getMessage();
            logger.error(methodLogName + " - " + message);
            throw new DAOException(message);
        }

        // The current line in the TRQ file.
        String currentLine;
        String[] currentCard;
        boolean endKeywordFound = false;

        // Read and process lines from the TransferRequest file.
        try {
            // Read the first line of the file.
            currentLine = trFileReader.readLine();
            // If the first line is null, the file is empty.
            if (currentLine == null) {
                String message = "File " + transferRequestFile + " is empty.";
                logger.error(methodLogName + " - " + message);
                throw new DAOException(message);
            }
            // The first line of a TransferRequest must contain the keyword
            // TransferRequest
            if (!currentLine.equals(BEGIN)) {
                String message =
                        "File " + transferRequestFile
                                + " is not a valid TransferRequest"
                                + " because it doesn't start with '" + BEGIN
                                + "'.";
                logger.error(methodLogName + " - " + message);
                throw new DAOException(message);
            }
            // Parse the TransferRequest contents.
            while ((currentLine = trFileReader.readLine()) != null) {
                logger.debug(methodLogName + " - Parsing line: " + currentLine);
                if (endKeywordFound == true) {
                    String message =
                            "File '" + transferRequestFile
                                    + "' is not a valid TransferRequest"
                                    + " because it doesn't end with '" + END
                                    + "'.";
                    logger.error(methodLogName + " - " + message);
                    throw new DAOException(message);
                }
                if (currentLine.equals(END)) {
                    endKeywordFound = true;
                    continue;
                }
                if (currentLine.trim().equals("")) {
                    /* Blank line read - skip to the next line. */
                    continue;
                }
                currentCard = currentLine.split("=");
                if (currentCard.length != 2) {
                    String message = "Could not parse line " + currentLine;
                    logger.error(methodLogName + " - " + message);
                    throw new DAOException(message);
                }
                // association between methods and fields
                // Possible leading and trailing whitespace are removed
                if (setTrField(transferRequest,
                        currentCard[0].trim(),
                        currentCard[1].trim()) != 0) {
                    String message =
                            "Failed to set value " + currentCard[1]
                                    + " associated to " + currentCard[0]
                                    + " from " + transferRequestFile;
                    logger.error(message);
                    throw new DAOException(message);
                }
            }
            if (endKeywordFound == false) {
                String message =
                        "File '" + transferRequestFile
                                + "' is not a valid TransferRequest"
                                + " because it doesn't end with '" + END + "'.";
                logger.error(methodLogName + " - " + message);
                throw new DAOException(message);
            }
        } catch (IOException ioe) {
            /* An IO error occurred whilst reading the TRQ file. */
            String message =
                    "An IO error occurred whilst reading from " + "TRQ file '"
                            + transferRequestFile + "' : " + ioe.getMessage();
            logger.error(methodLogName + " - " + message);
            throw new DAOException(message);
        } finally {
            try {
                logger.debug(methodLogName + " - Closing file "
                        + transferRequestFile);
                trFileReader.close();
            } catch (IOException ioe) {
                String message =
                        "Could not close file " + transferRequestFile + " : "
                                + ioe.getMessage();
                logger.error(methodLogName + " - " + message);
                throw new DAOException(message);
            }
        }
        return transferRequest;
    }

    /**
     * Calls the setter method of a transfer request matching a key and gives to
     * it a field value.
     * 
     * @param transferRequest
     *            transfer request whose a setter method should be called
     * @param key
     *            identifier of the setter method to call
     * @param value
     *            value to give to the setter method
     * 
     * @return 0 if a setter method has been called, -1 otherwise
     */
    private int setTrField(TransferRequest transferRequest, String key,
            String value) {
        int retStatus = 0;

        try {
            if (key.compareTo(FILEID) == 0) {
                transferRequest.setFileId(value);
            } else if (key.compareTo(FILENAME) == 0) {
                transferRequest.setFileName(value);
            } else if (key.compareTo(UNCOMPRESSED_FILENAME) == 0) {
                transferRequest.setUncompressedFileName(value);
            } else if (key.compareTo(TRANSFER_METHOD) == 0) {
                transferRequest.setTransferMethod(value);
            } else if (key.compareTo(TRANSFER_PRIORITY) == 0) {
                transferRequest.setTransferPriority(Integer.valueOf(value));
            } else if (key.compareTo(COMPRESSION_METHOD) == 0) {
                transferRequest.setCompressionMethod(value);
            } else if (key.compareTo(SIZE) == 0) {
                transferRequest.setSize(Long.valueOf(value));
            } else if (key.compareTo(UNCOMPRESSED_SIZE) == 0) {
                transferRequest.setUncompressedSize(Long.valueOf(value));
            } else if (key.compareTo(CHECKSUM) == 0) {
                transferRequest.setChecksum(value);
            } else if (key.compareTo(TRANSFER_CATG) == 0) {
                transferRequest.setTransferCatg(value);
            } else if (key.compareTo(USER) == 0) {
                transferRequest.setUser(value);
            } else if (key.compareTo(DESTDIR) == 0) {
                transferRequest.setDestDir(value);
            } else if (key.compareTo(MULTI_FILE_COMPONENTS) == 0) {
                transferRequest.setMultiFileComponents(value);
            } else if (key.compareTo(CREATION_TIME) == 0) {
                transferRequest.setCreationTime(new Long(dateFormat.parse(value)
                        .getTime()));
            } else {
                logger.error(key + " is not a valid keyword");
                retStatus = -1;
            }
        } catch (NumberFormatException ne) {
            logger.error("NumberFormatException raised for " + key
                    + " Bad value : " + value);
            retStatus = -1;
        } catch (ParseException e) {
            logger.error("ParseException raised for " + key + " Bad value : "
                    + value);
            retStatus = -1;
        }
        return retStatus;
    }

    /**
     * gets transfer requests associated to a list of instruments, in a given
     * status, starting from a given date ordered according to a given field in
     * a given order. If no instrument is given, all the instruments will be
     * selected
     * 
     * @param startDate
     *            date the research should start from
     * @param instrumentList
     *            List of instruments whose files shall be retrieved.
     *            Instruments will be separated by a space
     * @param status
     *            transfer request status
     * @param orderBy
     *            results sorting criteria
     * @param sortingOrder
     *            sorting order
     * @return a list of files fulfilling the user conditions
     */
    public List<TransferRequest> findByDateGreaterThan(Date startDate,
            String instrumentList, TransferRequest.Status status,
            TransferRequest.FIELDS orderBy, SortingOrder sortingOrder)
            throws DAOException {
        final String methodLogName = classLogName + "::findByDateGreaterThan()";
        List<TransferRequest> emptyList = new ArrayList<TransferRequest>();
        logger.warn(methodLogName + " - Method not implemented");
        return emptyList;
    }

    /**
     * gets transfer requests associated to a list of instruments of a given
     * observation date, in a given status, ordered according to some given
     * fields in a given order. If no instrument is given, all the instruments
     * will be selected
     * 
     * @param obsDate
     *            observation date of the files to be retrieved, starting from
     *            noon of the given day till noon of the following day
     * @param instrumentList
     *            List of instruments whose files shall be retrieved.
     *            Instruments will be separated by a space
     * @param statusList
     *            list of transfer request status
     * @param fileType
     *            type of the file to be transfered. Non mandatory
     * @param orderByList
     *            results sorting criterions
     * @param sortingOrder
     *            sorting order
     * @return a list of files fulfilling the user conditions
     */
    public List<TransferRequest> findByDate(Date obsDate,
            String instrumentList, List<TransferRequest.Status> statusList,
            String fileType, List<TransferRequest.FIELDS> orderByList,
            SortingOrder sortingOrder) throws DAOException {
        final String methodLogName = classLogName + "::findByDate()";
        List<TransferRequest> emptyList = new ArrayList<TransferRequest>();
        logger.warn(methodLogName + " - Method not implemented");
        return emptyList;
    }

    public void deleteByStatus(TransferRequest.Status status)
            throws DAOException {
        final String methodLogName = classLogName + "::deleteByStatus()";
        logger.trace(methodLogName);

        if (status == null) {
            String message = "Status is missing";
            logger.error(message);
            throw new IllegalArgumentException(message);
        }

        File statusDir = statusDirMap.get(status);

        List<TransferRequest> trqs = findByStatus(status);
        for (TransferRequest trq : trqs) {
            File trqFile =
                    new File(statusDir.getAbsolutePath()
                            + System.getProperty("file.separator")
                            + trq.getRequestId());
            if (trqFile.delete())
                logger.info(trqFile.getAbsolutePath() + " successfully deleted");
            else
                throw new DAOException("unable to delete "
                        + trqFile.getAbsolutePath());
        }
    }
}
