package org.eso.util.filesystem; 

import java.io.BufferedWriter; 
import java.io.File; 
import java.io.FilenameFilter; 
import java.io.FileWriter; 

import java.util.List; 
import java.util.ListIterator; 
import java.util.Set; 

import org.apache.log4j.Logger; 

/** <p>This class provides static methods for performing common filesystem 
 * operations with logging via Log4J. In addition, no checked exceptions 
 * are thrown.</p> 
 *
 * <p><b>WARNING:</b> the {@link #moveAllToDir()}, {@link #moveToDir()} 
 * and {@link #move()} methods are subject to the limitations of the Java 
 * {@link java.io.File#renameTo(java.io.File)} method. In particular, 
 * files cannot be moved from one filesystem to another. If you need this 
 * functionality, please contact the author (it would need to be implemented 
 * with an OS-specific command and a {@link java.lang.Runtime.exec(String[])} 
 * call).</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class FileSystemOperations
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(FileSystemOperations.class);

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "FileSystemOperations"; 

    /** For each Directory specified: if the directory is required, checks 
     * for its presence and creates it if necessary. This method and the 
     * accompanying {@link Directory} class are supplied as a convenience 
     * for applications which need to loop through a set of directories 
     * whose parameters may change between runs. The method delegates to 
     * the underlying {@link #requireDirectory()} method. 
     * 
     * @param directories an array of Directory objects to check. 
     * 
     * @throws NullPointerException if directories is null. 
     *
     * @return true if all of the directories which were actually required 
     *         were already present or have been created. False otherwise. 
     */
    public static boolean checkForDirectories(final Directory[] directories) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::checkForDirectories()";

        logger.trace(methodLogName); 

        if(directories == null)
        {
            final String message = 
                methodLogName + " - directories must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // True if all required directories are in place. 
        boolean allDirsInPlace = true; 

        // Text shared between log messages. 
        String commonLogText = null; 

        // Deal with each directory in turn. 
        for(int i = 0; i < directories.length; ++i)
        {
            if(directories[i].isRequired)
            {
                // Pre-assemble the common log text. 
                commonLogText = 
                    methodLogName + (directories[i].logText == null ? 
                                     " -" : " - " + directories[i].logText) 
                    + " directory [" + directories[i].path + "]"; 

                // Delegate to the requireDirectory() method. 
                if(FileSystemOperations.requireDirectory(
                       directories[i].path, 
                       directories[i].createDir)) 
                {
                    logger.debug(commonLogText + " in place."); 
                }
                else
                {
                    logger.error(commonLogText + " not available.");

                    allDirsInPlace = false; 
                }
            }
        }

        return(allDirsInPlace); 
    }

    /** Ensures that a directory exists at the specified path, creating it if 
     * necessary. 
     * 
     * @param path the path at which a directory is required. 
     * @param createDir true if the directory should be created when it does 
     *                  not already exist, false otherwise. 
     * 
     * @throws NullPointerException if path is null. 
     *
     * @return true if the directory was already present, or if it was 
     *         successfully created. Otherwise, false. 
     */
    public static boolean requireDirectory(final File path, 
                                           final boolean createDir) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::requireDirectory()"; 

        logger.trace(methodLogName); 

        if(path == null)
        {
            final String message = 
                methodLogName + " - path must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        try
        {
            if(path.exists()) 
            {
                if(path.isDirectory()) 
                {
                    logger.debug(methodLogName + " - existing directory [" 
                                 + path + "] found."); 

                    return(true); 
                }
                else 
                {
                    logger.error(methodLogName 
                                 + " - a non-directory is already present" 
                                 + " at path [" + path + "]."); 

                    return(false); 
                }
            }
            else
            {
                // The directory could not be found. 
                if(createDir)
                {
                    // Try to create the directory. 
                    if(path.mkdir()) 
                    {
                        logger.info(methodLogName 
                                    + " - directory [" + path + "] created.");

                        return(true); 
                    }
                    else
                    {
                        logger.error(methodLogName + " - directory [" + path 
                                     + "] could not be created."); 

                        return(false); 
                    }
                }
                else
                {
                    logger.error(methodLogName + " - directory [" + path 
                                 + "] is not present."); 

                    return(false); 
                }
            }
        }
        catch(Exception e)
        {
            logger.error(methodLogName + " - error [" + e.getMessage() 
                         + "] whilst attempting to require" 
                         + " directory [" + path + "]."); 
        }

        return(false); 
    }

    /** <p>Moves all files listed by sourcePaths into destDir. Successfully 
     * moved files are removed from sourcePaths; if a move fails, that file 
     * will be left in the list. An attempt will be made to move each file, 
     * even if some moves fail.</p> 
     *
     * <p><b>WARNING:</b> see the warning in the 
     * {@link FileSystemOperations class documentation}.</p> 
     *
     * @param sourcePaths the files to move. 
     * @param destDir the directory the files should be moved into. 
     *
     * @throws NullPointerException if sourcePaths or destDir is null. 
     * @throws IllegalArgumentException if destDir is not a directory, 
     *                                  or could not be accessed. 
     * 
     * @return true if all moves were successful, false otherwise. 
     */
    public static boolean moveAllToDir(final List<File> sourcePaths, 
                                       final File destDir) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::moveAllToDir()"; 

        logger.trace(methodLogName); 

        if((sourcePaths == null) || (destDir == null)) 
        {
            final String message = 
                methodLogName 
                + " - sourcePaths" + (sourcePaths == null ? " [null]" : "") 
                + " and destDir [" + destDir + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Move each file to the destination directory in turn. 
        boolean allFilesMoved = true; 
        final ListIterator<File> fileIterator = sourcePaths.listIterator(); 
        File currentFile = null; 

        while(fileIterator.hasNext()) 
        {
            currentFile = fileIterator.next(); 

            // Delegate to moveToDir(). 
            if(moveToDir(currentFile, destDir)) 
            {
                // File moved successfully; remove it from the list. 
                fileIterator.remove(); 
            }
            else
            {
                // File could not be moved; flag the problem. 
                allFilesMoved = false; 
            }
        }
        
        return(allFilesMoved); 
    }

    /** <p>Moves the file at sourcePath into destDir.</p> 
     *
     * <p><b>WARNING:</b> see the warning in the 
     * {@link FileSystemOperations class documentation}.</p> 
     *
     * @param sourcePath the file to move. 
     * @param destDir the directory the file should be moved into. 
     *
     * @throws NullPointerException if sourcePath or destDir is null. 
     * @throws IllegalArgumentException if destDir is not a directory, 
     *                                  or could not be accessed. 
     *
     * @return true if the move was successful, false otherwise. 
     */
    public static boolean moveToDir(final File sourcePath, final File destDir)
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = classLogName + "::moveToDir()"; 

        logger.trace(methodLogName); 

        if((sourcePath == null) || (destDir == null)) 
        {
            final String message = 
                methodLogName
                + " - sourcePath [" + sourcePath 
                + "] and destDir [" + destDir + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        } 

        try
        {
            // The destination must be a directory. 
            if(!destDir.isDirectory()) 
            {
                final String message = 
                    methodLogName 
                    + " - destDir [" + destDir + "] is not a directory."; 

                logger.fatal(message); 
                throw new IllegalArgumentException(message); 
            }
        }
        catch(Exception e) 
        {
            final String message = 
                methodLogName + " - error [" + e.getMessage() 
                + "] whilst attempting to access destination directory [" 
                + destDir + "]."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Assemble the destination path. 
        final File destPath = new File(destDir, sourcePath.getName()); 

        // Delegate to move(). 
        return(move(sourcePath, destPath)); 
    }

    /** <p>Moves the file at sourcePath to destPath.</p> 
     *
     * <p><b>WARNING:</b> see the warning in the 
     * {@link FileSystemOperations class documentation}.</p> 
     *
     * @param sourcePath the file to move. 
     * @param destPath the path to move the file to. 
     *
     * @throws NullPointerException if sourcePath or destPath is null. 
     * 
     * @return true if the move was successful, false otherwise. 
     */
    public static boolean move(final File sourcePath, final File destPath) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::move()"; 

        logger.trace(methodLogName); 

        if((sourcePath == null) || (destPath == null)) 
        {
            final String message = 
                methodLogName + " - sourcePath [" + sourcePath 
                + "] and destPath [" + destPath + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        // Attempt to move the file by renaming it. 
        try
        {
            if(sourcePath.renameTo(destPath)) 
            {
                // The move was successful. 
                logger.debug(methodLogName + " - moved [" + sourcePath 
                             + "] to [" + destPath + "]."); 

                return(true); 
            }
            else
            {
                // The move failed. 
                logger.error(methodLogName + " - unable to move [" 
                             + sourcePath + "] to [" + destPath + "]."); 

                return(false); 
            }
        }
        catch(Exception e)
        {
            // The move failed. 
            logger.error(methodLogName + " - error [" + e.getMessage() 
                         + "] whilst attempting to move [" 
                         + sourcePath + "] to [" + destPath + "]."); 
        } 

        return(false); 
    }

    /** Writes the specified string to the specified file, appending to or 
     * overwriting any existing file, as appropriate. This method is best 
     * suited to one-off writes. When making repeated writes to the same 
     * file, applications can gain efficiency by performing the writes 
     * directly in a loop. 
     * 
     * @param string the string to write. 
     * @param append true if the string should be appended to any existing 
     *               file, false if the file should be overwritten. 
     * @param outputFile the file to write to. 
     *
     * @throws NullPointerException if string or outputFile is null. 
     * 
     * @return true if the string was successfully written to the file, 
     *         false otherwise. 
     */ 
    public static boolean writeStringToFile(final String string, 
                                            final boolean append, 
                                            final File outputFile) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::writeStringToFile()"; 

        logger.trace(methodLogName); 

        if((string == null) || (outputFile == null)) 
        {
            final String message = 
                methodLogName + " - string [" + string 
                + "] and outputFile [" + outputFile + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        try
        {
            // The string could be large; use buffered writes for efficiency. 
            final BufferedWriter writer = 
                new BufferedWriter(new FileWriter(outputFile, append)); 

            // Write the entire string to the file, then close it again.  
            writer.write(string); 
            writer.close(); 

            logger.debug(methodLogName 
                         + " - string successfully written to file [" 
                         + outputFile + "]."); 

            return(true); 
        }
        catch(Exception e)
        {
            logger.error(methodLogName + " - error [" + e.getMessage() 
                         + "] whilst attempting to write to file [" 
                         + outputFile + "]."); 
        }

        return(false); 
    }

    /** Checks whether a file or directory exists at the specified path. 
     *
     * @param path the file or directory whose existence should be checked. 
     *
     * @throws NullPointerException if path is null. 
     * 
     * @return true if the file or directory exists, false otherwise. 
     */
    public static boolean exists(final File path) throws NullPointerException 
    {
        final String methodLogName = classLogName + "::exists()"; 

        logger.trace(methodLogName); 

        if(path == null)
        {
            final String message = 
                methodLogName + " - path must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        try
        {
            // Check whether the file or directory exists. 
            if(path.exists()) 
            {
                logger.debug(methodLogName + " - [" + path + "] exists."); 

                return(true); 
            }
            else
            {
                logger.debug(methodLogName 
                             + " - [" + path + "] does not exist."); 

                return(false); 
            }
        }
        catch(Exception e)
        {
            logger.error(methodLogName + " - error [" + e.getMessage() 
                         + "] whilst attempting to check whether [" 
                         + path + "] exists."); 
        }

        return(false); 
    }

    /** Checks whether the specified path points to a normal file which 
     * exists on the disk. 
     *
     * @param path the path to check for the presence of a normal file. 
     *
     * @throws NullPointerException if path is null. 
     * 
     * @return true if the path points to a normal file which exists on the 
     *         disk, false otherwise. 
     */
    public static boolean isFile(final File path) throws NullPointerException 
    {
        final String methodLogName = classLogName + "::isFile()"; 

        logger.trace(methodLogName); 

        if(path == null)
        {
            final String message = 
                methodLogName + " - path must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        try
        {
            // Check whether the path points to a normal file which 
            // exists on the disk. 
            if(path.isFile()) 
            {
                logger.debug(methodLogName + " - [" + path 
                             + "] is a normal file and exists on the disk."); 

                return(true); 
            }
            else
            {
                logger.debug(methodLogName + " - [" + path 
                             + "] is not a normal file, or does not exist."); 

                return(false); 
            }
        }
        catch(Exception e)
        {
            logger.error(methodLogName + " - error [" + e.getMessage() 
                         + "] whilst attempting to check whether [" + path 
                         + "] is a normal file and exists on the disk."); 
        }

        return(false); 
    }

    /** Checks whether the specified path points to a directory which 
     * exists on the disk. 
     *
     * @param path the path to check for the presence of a directory. 
     *
     * @throws NullPointerException if path is null. 
     * 
     * @return true if the path points to a directory which exists on the 
     *         disk, false otherwise. 
     */
    public static boolean isDirectory(final File path) 
        throws NullPointerException 
    {
        final String methodLogName = classLogName + "::isDirectory()"; 

        logger.trace(methodLogName); 

        if(path == null)
        {
            final String message = 
                methodLogName + " - path must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        try
        {
            // Check whether the path points to a directory which 
            // exists on the disk. 
            if(path.isDirectory()) 
            {
                logger.debug(methodLogName + " - [" + path 
                             + "] is a directory and exists on the disk."); 

                return(true); 
            }
            else
            {
                logger.debug(methodLogName + " - [" + path 
                             + "] is not a directory, or does not exist."); 

                return(false); 
            }
        }
        catch(Exception e)
        {
            logger.error(methodLogName + " - error [" + e.getMessage() 
                         + "] whilst attempting to check whether [" + path 
                         + "] is a directory and exists on the disk."); 
        }

        return(false); 
    }

    /** Adds the names of the files in the specified directory to the 
     * given {@link java.util.Set} of strings, filtering with a 
     * {@link java.io.FilenameFilter} if desired. It is perfectly acceptable 
     * for a particular filename to already exist within the set; in that 
     * case, this method will return true. 
     * 
     * @param path the directory path. 
     * @param filter the filter to apply to the filenames, or null if no 
     *               filter is required. 
     * @param set the set of strings to store the filenames in. 
     *
     * @throws NullPointerException if path or set is null. 
     * @throws IllegalArgumentException if path is not a directory, 
     *                                  or could not be accessed. 
     * 
     * @return true if the directory was successfully listed, false otherwise.
     */ 
    public static boolean 
    listFilenamesInDirectory(final File path, 
                             final FilenameFilter filter, 
                             final Set<String> set) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::listFilenamesInDirectory()"; 

        logger.trace(methodLogName); 

        if((path == null) || (set == null)) 
        {
            final String message = 
                methodLogName + " - path [" + path 
                + "] and set [" + set + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        try
        {
            // The path must be a directory. 
            if(!path.isDirectory()) 
            {
                final String message = 
                    methodLogName 
                    + " - path [" + path + "] is not a directory."; 

                logger.fatal(message); 
                throw new IllegalArgumentException(message); 
            }
        }
        catch(Exception e) 
        {
            final String message = 
                methodLogName + " - error [" + e.getMessage() 
                + "] whilst attempting to access directory [" + path + "]."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // An array of filename strings. 
        String[] filenames = null; 

        try
        {
            // List the names of the files within the directory. 
            if(filter == null) 
            {
                filenames = path.list(); 
            }
            else
            {
                filenames = path.list(filter); 
            }
        }
        catch(Exception e) 
        {
            logger.error(
                methodLogName + " - error [" + e.getMessage() 
                + "] whilst attempting to list contents of directory [" 
                + path + "]."); 

            return(false); 
        } 

        if(filenames == null) 
        {
            logger.error(
                methodLogName 
                + " - unable to list contents of directory [" + path + "]."); 

            return(false); 
        }

        logger.debug(methodLogName + " - " + filenames.length 
                     + (filenames.length == 1 
                        ? " file found in directory [" 
                        : " files found in directory [") + path + "]."); 

        // Store the filenames into the set of strings. 
        for(int i = 0; i < filenames.length; ++i) 
        {
            set.add(filenames[i]); 
        }

        return(true); 
    }

    /** Deletes the file at the specified path. 
     *
     * @param path the file to delete. 
     *
     * @throws NullPointerException if path is null. 
     * 
     * @return true if the file was successfully deleted, false otherwise. 
     */
    public static boolean delete(final File path) throws NullPointerException 
    {
        final String methodLogName = classLogName + "::delete()"; 

        logger.trace(methodLogName); 

        if(path == null)
        {
            final String message = 
                methodLogName + " - path must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        try
        {
            // Attempt to delete the file. 
            if(path.delete()) 
            {
                logger.debug(methodLogName + " - deleted [" + path + "]."); 

                return(true); 
            }
            else
            {
                logger.error(methodLogName 
                             + " - could not delete [" + path + "]."); 

                return(false); 
            }
        }
        catch(Exception e)
        {
            logger.error(methodLogName + " - error [" + e.getMessage() 
                         + "] whilst attempting to delete [" + path + "]."); 
        }

        return(false); 
    }
}
