package org.eso.util.misc;

import java.util.HashMap; 
import java.util.HashSet; 
import java.util.Iterator; 
import java.util.Map; 
import java.util.Set; 

import java.util.concurrent.ScheduledThreadPoolExecutor; 
import java.util.concurrent.TimeUnit; 

import org.apache.log4j.Logger;

/** <p>This class provides a means for threads to schedule delayed-action 
 * interrupts on themselves via the {@link #register()} method. This method 
 * returns an {@link org.eso.util.misc.ThreadWatchdog.Interrupt Interrupt} 
 * instance which can be used to: 
 * <ul> 
 *   <li>cancel the interrupt before it is executed.</li>
 *   <li>explicitly execute the interrupt early.</li>
 *   <li>query the interrupt's status.</li>
 * </ul>
 * Interrupts can also be cancelled en-masse for a thread using the 
 * {@link #unregister()} method.</p> 
 *
 * <p>The {@link org.eso.util.misc.ThreadWatchdog.Interrupt Interrupt} 
 * instances provided by this class represent one-shot interrupts, i.e. they 
 * can only be executed once, whether explicitly or automatically. If an 
 * interrupt is explicitly executed before the automatic execution takes 
 * place, the automatic execution will be cancelled.</li> 
 * 
 * <p>This type of watchdog facility is often used to wake a thread up 
 * when it is blocked on a task which is taking too long.</p> 
 * 
 * <p><b>WARNING:</b> see {@link #unregister()} and 
 * {@link org.eso.util.misc.ThreadWatchdog.Interrupt#cancel() Interrupt.cancel()}. 
 * </p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */ 
public class ThreadWatchdog 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(ThreadWatchdog.class);

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "ThreadWatchdog"; 

    /** This class is a singleton; this is the single instance. */ 
    private static final ThreadWatchdog instance = new ThreadWatchdog(); 

    /** A scheduled execution pool containing one thread to execute 
     * Interrupts as they become due. 
     */ 
    private static final ScheduledThreadPoolExecutor executor = 
        new ScheduledThreadPoolExecutor(1, new DefaultDaemonThreadFactory()); 

    /** A map of threads to sets of Interrupts scheduled for them. */ 
    private static final Map<Thread, Set<Interrupt>> 
        interruptSetMap = new HashMap<Thread, Set<Interrupt>>(); 

    /** Constructs a new ThreadWatchdog. */ 
    private ThreadWatchdog() 
    {
        final String methodLogName = classLogName + "::ThreadWatchdog()"; 

        logger.trace(methodLogName); 

        // This constructor does not need to do anything. 
    }

    /** Returns a new Interrupter for the specified thread. */ 
    private Interrupter getInterrupter(final Thread thread) 
    {
        final String methodLogName = classLogName + "::getInterrupter()"; 

        logger.trace(methodLogName); 

        return(new Interrupter(thread)); 
    }

    /** Registers the specified Thread with the ThreadWatchdog, but without 
     * scheduling an automatic interrupt. Threads can be registered for as 
     * many interrupts as necessary; once all pending interrupts have been 
     * executed or cancelled, the thread will be unregistered automatically. 
     * 
     * @param thread the thread to register an interrupt for. 
     * 
     * @throws NullPointerException if thread is null. 
     * 
     * @return a one-shot 
     *         {@link org.eso.util.misc.ThreadWatchdog.Interrupt Interrupt} 
     *         which can only be executed explicitly. 
     */ 
    public static Interrupt register(final Thread thread) 
    {
        final String methodLogName = classLogName + "::register(Thread)"; 

        logger.trace(methodLogName); 

        // Return a delayed-action Interrupt with an unfeasibly long delay 
        // - 292 billion years! 
        return(register(thread, Long.MAX_VALUE, TimeUnit.SECONDS)); 
    }

    /** Registers the specified Thread with the ThreadWatchdog and schedules 
     * an automatic interrupt, so that it will be interrupted after the 
     * specified number of time units have passed. Threads can be registered 
     * for as many interrupts as necessary; once all pending interrupts have 
     * been executed or cancelled, the thread will be unregistered 
     * automatically. 
     * 
     * @param thread the thread to register an interrupt for. 
     * @param delay the length of time that the ThreadWatchdog should wait 
     *              before interrupting the thread. 
     * @param unit the unit of time that the delay is specified in. 
     * 
     * @throws IllegalArgumentException if delay is less than one. 
     * @throws NullPointerException if thread or unit is null. 
     *
     * @return a one-shot 
     *         {@link org.eso.util.misc.ThreadWatchdog.Interrupt Interrupt} 
     *         which may be executed explicitly or automatically. 
     */ 
    public static Interrupt register(final Thread thread, 
                                     final long delay, 
                                     final TimeUnit unit) 
        throws IllegalArgumentException, NullPointerException 
    {
        final String methodLogName = 
            classLogName + "::register(Thread, long, TimeUnit)"; 

        logger.trace(methodLogName); 

        if((thread == null) || (unit == null)) 
        {
            final String message = 
                methodLogName + " - thread [" + thread 
                + "] and unit [" + unit + "] must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        if(delay < 1)
        {
            final String message = 
                methodLogName 
                + " - delay [" + delay + "] must be at least one."; 

            logger.fatal(message); 
            throw new IllegalArgumentException(message); 
        }

        // Get an Interrupter for the thread. 
        final Interrupter interrupter = instance.getInterrupter(thread); 

        synchronized(interruptSetMap)
        {
            // Get this thread's interrupt set. 
            Set<Interrupt> interruptSet = interruptSetMap.get(thread); 

            if(interruptSet == null)
            {
                // This thread does not have an interrupt set yet; create one 
                // and map the thread to it. 
                interruptSet = new HashSet<Interrupt>(); 
                interruptSetMap.put(thread, interruptSet); 
            }

            // Add the Interrupter to the thread's interrupt set... 
            interruptSet.add(interrupter); 

            // ...and schedule it. 
            executor.schedule(interrupter, delay, unit); 
        }

        logger.debug(methodLogName 
                     + " - scheduled interrupt in [" + delay + "][" + unit 
                     + "] for thread [" + thread.getName() + "]."); 

        return(interrupter); 
    }

    /** <p>Unregisters the specified thread from the ThreadWatchdog. It 
     * will no longer have {@link java.lang.Thread#interrupt()} called on 
     * it by the watchdog. All interrupts registered for that thread will 
     * be cancelled.</p> 
     * 
     * <p><b>WARNING:</b> interrupts may be issued even as this method is 
     * running; if that could be a problem, the caller must guard against 
     * it. Interrupt cancellation is not guaranteed until this method has 
     * returned.</p>
     * 
     * @param thread the thread to unregister. 
     * 
     * @throws NullPointerException if thread is null. 
     */ 
    public static void unregister(final Thread thread) 
        throws NullPointerException
    {
        final String methodLogName = classLogName + "::unregister()"; 

        logger.trace(methodLogName); 

        if(thread == null) 
        {
            final String message = 
                methodLogName + " - thread must not be null."; 

            logger.fatal(message); 
            throw new NullPointerException(message); 
        }

        synchronized(interruptSetMap)
        {
            // Get the thread's interrupt set. 
            final Set<Interrupt> interruptSet = interruptSetMap.get(thread); 

            if(interruptSet == null)
            {
                logger.debug(methodLogName 
                             + " - thread [" + thread.getName() 
                             + "] not registered with watchdog."); 
                return; 
            }

            // Iterate through the interrupts, cancelling each one. 
            // 
            // To avoid a concurrent modification exception, the iterator 
            // must use a temporary copy of the set. 
            Iterator<Interrupt> interruptIterator = 
                (new HashSet<Interrupt>(interruptSet)).iterator(); 

            while(interruptIterator.hasNext())
            {
                interruptIterator.next().cancel(); 
            }
        }

        // The thread is automatically unregistered from the watchdog when 
        // the last interrupt is cancelled. 
        logger.debug(methodLogName + " - all interrupts cancelled; thread [" 
                     + thread.getName() + "] unregistered from watchdog."); 
    }

    /** This interface describes the public interface to {@link Interrupt}s
     * returned by {@link ThreadWatchdog#register()}, allowing its callers 
     * to cancel interrupts, execute them explicitly and query their status. 
     * 
     * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
     */ 
    public interface Interrupt 
    {
        /** <p>Cancels the interrupt.</p> 
         * 
         * <p><b>WARNING:</b> the interrupt may be issued even as this 
         * method is running; if that could be a problem, the caller 
         * must guard against it. Cancellation is not guaranteed until 
         * this method has returned.</p> 
         */ 
        public void cancel(); 

        /** Returns true if this interrupt has been cancelled. */ 
        public boolean cancelled(); 

        /** Executes the interrupt immediately. */ 
        public void execute(); 

        /** Returns true if this interrupt has been executed. */ 
        public boolean executed(); 
    }

    /** Instances of this class represent single {@link Interrupt}s 
     * which have been registered with the {@link ThreadWatchdog}. 
     * 
     * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
     */ 
    private class Interrupter implements Interrupt, Runnable 
    {
        /** The name of this class for logging purposes. */ 
        private final String classLogName = "Interrupter"; 

        /** The thread this Interrupter should interrupt when executed. */ 
        private final Thread thread; 

        /** True if this Interrupt has been cancelled. */ 
        private volatile boolean cancelled = false; 

        /** True if this Interrupt has been executed. */ 
        private volatile boolean executed = false; 

        /** Constructs a new Interrupter object to interrupt the specified 
         * thread. 
         *
         * @param thread the thread to interrupt when {@link #run()} is 
         *               called. 
         * 
         * @throws NullPointerException if thread is null. 
         */ 
        Interrupter(Thread thread) throws NullPointerException 
        {
            final String methodLogName = classLogName + "::Interrupter()"; 

            logger.trace(methodLogName); 

            if(thread == null) 
            {
                final String message = 
                    methodLogName + " - thread must not be null."; 

                logger.fatal(message); 
                throw new NullPointerException(message); 
            }

            // Set the thread to interrupt. 
            this.thread = thread; 
        }

        // Documentation is inherited from the Interrupt interface. 
        public void cancel()
        {
            final String methodLogName = classLogName + "::cancel()"; 

            logger.trace(methodLogName); 

            synchronized(interruptSetMap)
            {
                if(cancelled)
                {
                    logger.debug(methodLogName 
                                 + " - interrupt already cancelled."); 

                    return; 
                }

                // Unregister this interrupt and mark it cancelled. 
                unregister(); 
                cancelled = true; 

                logger.debug(methodLogName 
                             + " - interrupt cancelled for thread [" 
                             + thread.getName() + "]."); 
            }
        }

        // Documentation is inherited from the Interrupt interface. 
        public boolean cancelled() 
        {
            final String methodLogName = classLogName + "::cancelled()"; 

            logger.trace(methodLogName); 

            return(cancelled); 
        }

        // Documentation is inherited from the Interrupt interface. 
        public void execute() 
        {
            final String methodLogName = classLogName + "::executed()"; 

            logger.trace(methodLogName); 

            // Delegate the actual execution. 
            run(); 
        }

        // Documentation is inherited from the Interrupt interface. 
        public boolean executed() 
        {
            final String methodLogName = classLogName + "::executed()"; 

            logger.trace(methodLogName); 

            return(executed); 
        }

        /** Interrupts the thread held by this Interrupter. */ 
        public void run() 
        {
            final String methodLogName = classLogName + "::run()"; 

            logger.trace(methodLogName); 

            synchronized(interruptSetMap)
            {
                if(cancelled)
                {
                    logger.debug(methodLogName 
                                 + " - this interrupt has been cancelled."); 

                    return; 
                }

                if(executed)
                {
                    logger.debug(
                        methodLogName 
                        + " - this interrupt has already been executed."); 

                    return; 
                }

                // Unregister this interrupt and mark it executed. 
                // 
                // This must be done before executing the interrupt, so that 
                // the interrupted thread will get the correct result if it 
                // calls executed(). 
                unregister(); 
                executed = true; 

                // Execute the interrupt. 
                thread.interrupt(); 

                logger.debug(methodLogName 
                             + " - interrupt executed on thread [" 
                             + thread.getName() + "]."); 
            }
        }

        /** Unregisters this Interrupt, and only this Interrupt, from the 
         * {@link ThreadWatchdog}. 
         */ 
        private void unregister()
        {
            final String methodLogName = classLogName + "::unregister()"; 

            logger.trace(methodLogName); 

            synchronized(interruptSetMap) 
            {
                // Get the thread's interrupt set. 
                final Set<Interrupt> 
                    interruptSet = interruptSetMap.get(thread); 

                // Unschedule the interrupt, then remove it from the thread's 
                // interrupt set. 
                executor.remove(this); 
                interruptSet.remove(this); 

                logger.debug(methodLogName 
                             + " - interrupt unregistered for thread [" 
                             + thread.getName() + "]."); 

                // If the thread's interrupt set is empty, unmap the thread. 
                if(interruptSet.isEmpty()) 
                {
                    interruptSetMap.remove(thread); 

                    logger.debug(methodLogName 
                                 + " - no further interrupts scheduled;" 
                                 + " unregistering thread [" 
                                 + thread.getName() + "] from watchdog."); 
                }
            }
        }
    }
}
