package org.eso.util.stream; 

import org.apache.log4j.Logger;

/** <p>A CrossOverTask represents a direction-changer within a stream - 
 * if placed within a downstream flow, it will forward all messages to 
 * the upstream side, and vice versa. CrossOverTasks never hold or modify 
 * any messages - all messages received are immediately forwarded to the 
 * sibling task in the containing module.</p> 
 * 
 * <p>The containing module must be set before using an instance of this 
 * class.</p> 
 * 
 * @author Neil Ferguson ESO/SDD/DFI/BET <nferguso@eso.org> 
 */
public class CrossOverTask extends RoutingTask 
{
    /** Apache Log4J logger. */ 
    static final Logger logger = Logger.getLogger(CrossOverTask.class); 

    /** The name of this class for logging purposes. */ 
    private static final String classLogName = "CrossOverTask"; 

    /** Constructs a new CrossOverTask with the name "CrossOverTask". */ 
    public CrossOverTask()
    {
        super("CrossOverTask"); 

        final String methodLogName = 
            classLogName + "::CrossOverTask() [for task CrossOverTask]"; 

        logger.trace(methodLogName); 
    }

    /** Constructs a new CrossOverTask with the specified name. 
     *
     * @param taskName the name of this task. 
     */
    public CrossOverTask(final String taskName)
    {
        super(taskName); 

        final String methodLogName = 
            classLogName 
            + "::CrossOverTask(String) [for task " + taskName + "]"; 

        logger.trace(methodLogName); 
    }
    
    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get() 
    {
        final String methodLogName = 
            classLogName + "::get() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** This task never has any messages available. 
     *
     * @return null. 
     */ 
    public final Message get(final long timeout) 
    {
        final String methodLogName = 
            classLogName + "::get(long) [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        return(null); 
    }

    /** {@inheritDoc} All messages are handed off to a 
     * {@link #putNext(Message message)} call. 
     */
    public final void put(final Message message)
    {
        final String methodLogName = 
            classLogName + "::put() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 

        // Delegate to putNext(). 
        putNext(message); 
    }

    /** Sends a message to the sibling task of this task, as determined by 
     * its containing module. 
     * 
     * @param message the message for the sibling task. 
     */
    public final void putNext(final Message message) 
    {
        final String methodLogName = 
            classLogName + "::putNext() [for task " + getTaskName() + "]"; 

        logger.trace(methodLogName); 
        logger.debug(methodLogName 
                     + " - forwarding message to sibling task."); 

        getContainingModule().getSiblingTask(this).put(message); 
    }
}
