package org.eso.dfs.gui;

import javax.swing.ComboBoxModel;
import javax.swing.JComboBox;

/**
 * @author Nick Kornweibel
 * 
 * The RecallingComboBox, if set as editable and used with the a RecallingComboBoxModel, will
 * record all new entries in the text area in the RecallingComboBoxModel.
 * The RecallingComboBoxModel typically stores all new entries, making them available
 * for quick selection in the ComboBox list.
 * 
 * This class can be used without a RecallingComboBoxModel, but it provides no features
 * over the standard JComboBox.
 * 
 * The class sets editable to true, as the RecallingComboBox is of most use when editable.
 * 
 * The finalize() method is included as the RecallingComboBox is aware of the
 * SharedListComboBoxModel and will calls its detach() method automatically when
 * finalizing(). 
 */
public class RecallingComboBox extends JComboBox {

    /**
     * Constructor for RecallingComboBox.
     */
    public RecallingComboBox() {
        setEditable(true);
        setModel(new RecallingComboBoxModel());
    }
    
    /**
     * Constructor for RecallingComboBox with a sorted
     * model.
     * @param sorted
     */
    public RecallingComboBox(boolean sorted, int maxCapacity) {
        setEditable(true);
        RecallingComboBoxModel mdl = new RecallingComboBoxModel(sorted);
        mdl.setMaxCapacity(maxCapacity);
        setModel(mdl);
    }
    
    public RecallingComboBox(RecallingComboBoxModel mdl) {
        setEditable(true);
        setModel(mdl);
    }

    /**
     * The setModel() takes special action if the current or new model is a
     * RecallingComboBoxModel.
     * The RecallingComboBox model is updated via a listener, and this listener
     * must be added or removed when adding/removing a RecallingComboBoxModel.
     * 
     * @see javax.swing.JComboBox#setModel(javax.swing.ComboBoxModel)
     */
    public void setModel(ComboBoxModel aModel) {
        // remove the old model from the action listener list if is it the right type.    
        if (getModel() != null && getModel() instanceof RecallingComboBoxModel) {
            RecallingComboBoxModel mdl = (RecallingComboBoxModel) getModel();
            removeActionListener(mdl);
            if (mdl instanceof SharedListComboBoxModel) {
                ((SharedListComboBoxModel)mdl).detach();
            }
        }
        super.setModel(aModel);
        // add the new model as an action listener if is it the right type.
        if (aModel != null && aModel instanceof RecallingComboBoxModel) {
            RecallingComboBoxModel mdl = (RecallingComboBoxModel)aModel;
            addActionListener(mdl);
        }
    }
    
    /**
     * This method returns the String value of the current selection.
     * It is included to permit easy migration of the JTextField to
     * RecallingComboBox.
     * @return String
     */
    public String getText() {
        Object val = getSelectedItem();
        return val == null ? "" : val.toString();
    }
    
    /**
     * This method sets the String value of the current selection.
     * It is included to permit easy migration of the JTextField to
     * RecallingComboBox.
     * @return String
     */
    public void setText(String val) {
        setSelectedItem(val);
    }

    /**
     * This method is overriden to ensure that the detach() method
     * of the RecallingComboBoxModel class must be called when the
     * model is no longer needed.
     * It is TBD if we need to override addNotify() to "reattach()" to
     * the model.
     * @see java.awt.Component#removeNotify()
     */
    public void removeNotify() {
        if (getModel() != null && getModel() instanceof SharedListComboBoxModel) {
            ((SharedListComboBoxModel)getModel()).detach();
        }
        super.removeNotify();
    }
}