package org.eso.dfs.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

import org.eso.dfs.io.MemMappedFileInputStream;
import org.eso.dfs.io.MemMappedFileReader;
import org.eso.dfs.io.ZCatReader;

/**
 * @author nkornwei
 *
 * Read a Fits format stream from a reader or file address
 * and process it as tokens. Currently only KeywordTokens
 * are extracted.
 * 
 * If a reader is not provided, the tokenizer will select
 * the most appropriate type of reader based on the system and
 * the fits file extension. .zip and .gz extensions are supported.
 * 
 * Have a look at main() for an example on how to use this class suite.
 */

public class FitsTokenizer {

    /**
     * If true, attempt memory mapping, else do not even
     * attempt memory mapping.
     */
    private static boolean memoryMappingEnabled = true;

    /**
     * Defines a comment token in a keyword record.
     */
    public static String COMMENT_TOKEN = " / ";

    /**
     * Name of the first keyword in an extended keyword block
     */
    public static final String EXTENDSTART = "XTENSION";

    /**
     * Name of the first keywork in the standard keyword block
     * NB: may also be used instead of EXTENDSTART....
     */
    public static final String SIMPLESTART = "SIMPLE";
    
    /**
     * Name of the keyword used to identify the number of axis in the image
     */
    public static final String NAXIS = "NAXIS";
    
    /**
     * Name of the prefix of the keywords idenitying the pixel size of each
     * axis in the image (e.g. if NAXIS is 2, there should be a NAXIS1 and NAXIS2
     * keyword.
     */
    public static final String AXIS = NAXIS;
    
    /**
     * Name of the keyword that identifies the size of the pixel speicified in
     * the NAXIS keywords.
     */
    public static final String BITPIX = "BITPIX";

    /**
     * Try reading files using memory mapping before reading with "standard"
     * Java I/O. This is a performance boost, but if resources do not allow
     * it, the "standard" Java I/O is used.
     */
    private static boolean useMemoryMapping = true;

    /**
     * The length of a single keyword in the header.
     */
    public static final int FITS_REC_LENGTH = 80;

    /**
     * The length of a single record in a fits file. Keyword blocks
     * always start on these record boundaries. This member could also
     * have been named "BLOCK" or "BLOCKSIZE"...
     */
    public static final int BOUNDARY = 2880;

    /**
     * Used to mark the end of keyword block
     */
    public static final String endToken = "END";

    /**
     * The set of KeywordListers registered with this
     * tokenizer.
     */
    private List keywordListeners = new Vector();

    /**
     * Indicates if extensions are to be read or not.
     */
    private boolean readExtensions = true;

    /**
     * The reader used for all data input.
     */
    private Reader otherReader;

    /**
     * the address of the file (null if a reader was
     * provided by the caller.
     */
    private String fileAddress;

    /**
     * Constructor for FitsTokenizer. Reads from the file
     * specified, use the most appropriate reader, based on
     * the file extension and system.
     *  
     * @param fileAddress the file path to the fits file.
     * @param readExt indicates if the
     * FITS extensions are to be read as well as the
     * primary header.
     */
    public FitsTokenizer(String fileAddr, boolean readExt) {
        readExtensions = readExt;
        fileAddress = fileAddr;
    }

    /**
     * Constructor for FitsTokenizer. Reads from the
     * provided reader.
     *  
     * @param rdr the reader to read from.
     * @param readExt indicates if the
     * FITS extensions are to be read as well as the
     * primary header.
     */
    public FitsTokenizer(Reader rdr, boolean readExt) {
        readExtensions = readExt;
        otherReader = rdr;
    }

    /**
     * Add a keyword listner to the set of listeners.
     * 
     * @param kwl
     */
    public void addKeywordListener(KeywordListener kwl) {
        if (!keywordListeners.contains(kwl)) {
            keywordListeners.add(kwl);
        }
    }

    /**
     * Remove the given KeywordListener.
     * @param kwl to be removed.
     * @return boolean true if found/removed, else false.
     */
    public boolean removeKeywordListener(KeywordListener kwl) {
        return keywordListeners.remove(kwl);
    }


    /**
     * Disable or enable memory mapping of files.
     * If enabled, memory mapping is not gaurenteed, but attempted.
     * If disabled, memory mapping is not attempted.
     * @param b
     */
    public static void setMemoryMappingEnabled(boolean b) {
        memoryMappingEnabled = b;
    }
    
    /**
     * Returns the readExtensions
     * @return boolean true if this tokenizer
     * is to read extensions (and primary header)
     * false if it is to read only the primary header.
     */
    public boolean isReadExtensions() {
        return readExtensions;
    }

    /**
     * Read the Fits file, calling the token listeners as found
     * @throws IOException on error.
     * @throws FileNotFoundException if fileAddress specified was not found.
     */
    public void read() throws IOException, FileNotFoundException {
        boolean endParsed = false;
        File fitsFile = new File(fileAddress);
        boolean isCompressed = false;
        FileReader dataSource = null;
        Reader input = null;

        if (otherReader != null) {
            input = otherReader;
        } else {
            dataSource = new FileReader(fileAddress);
            if (dataSource == null) {
                throw new IOException(
                    "Failed to create reader on " + fileAddress);
            }

            if (fileAddress.endsWith(".fits.Z")) {
                input = zcatFile(fileAddress);
                isCompressed = true;
                // cannot read extensions for a compressed file.
                readExtensions = false;
            } else if (fileAddress.endsWith(".fits.gz")) {
                input = gzipFile(fileAddress);
                isCompressed = true;
                // cannot read extensions for a compressed file.
                readExtensions = false;
            } else {
                if (memoryMappingEnabled && useMemoryMapping) {
                    input = memMapFile(fileAddress);
                }
                // If memory mapping the file fails/is not set, we try a normal read.
                if (input == null) {
                    input = new BufferedReader(dataSource);
                }
            }
        }
        char cbuf[] = new char[FITS_REC_LENGTH];
        int bBuffered = 0;
        int totalRead = 0;

        int currentExtension = 0;
        int currentKeywordIndex = 0;
        
        // set to true if the keyword indicating extension is found.
        boolean hasExtensions = false;
        long requiredToSkip = 0L;
        int requiredToRead = FITS_REC_LENGTH;
        
        Dictionary naxisKeys = new Hashtable();
        while (true) {
            int bRead = input.read(cbuf, bBuffered, requiredToRead - bBuffered);
            if (bRead >= 0) {
                bBuffered += bRead;
                totalRead += bRead;
            } else {
                // bRead == -1 - end of stream.
                break;
            }
            if (bBuffered == requiredToRead) {
                String rec = new String(cbuf);
                //System.out.println("Reading..."+totalRead+" "+rec.substring(0, 5));.
                if (endParsed
                    && (rec.startsWith(EXTENDSTART)
                        || rec.startsWith(SIMPLESTART))) {
                    //
                    // we have found the start of an extension.
                    //
                    /*
                    System.out.println(
                        "START/EXTSTART found at: "
                            + (totalRead - requiredToRead)
                            + ".");
                    */
                    endParsed = false;
                    currentExtension++;
                    currentKeywordIndex = 0;
                } else if (rec.startsWith(endToken)) {
                    /*
                    System.out.println(
                        "END found at: "
                            + (totalRead - requiredToRead)
                            + ".");
                    */
                    //
                    // we have read the END token at the end of a
                    // keyword primary header or extension.
                    //
                    endParsed = true;
                    if (!readExtensions) {
                        // 
                        // we were only required to read the header,
                        // so we are done.
                        //
                        break;
                    }
                    if (!hasExtensions) {
                        /*
                         * Although many files mark the presence of extensions
                         * with the EXTEND keyword, not all do - so viewing
                         * the value of this keyword is not used to determine
                         * if the Fits file has extensions or not - instead
                         * is is assumed that is does.
                         */
                        // we have just finished reading the header
                        // lets see if this file has extensions:
                        // Keyword k = getKeyword(EXTENSION);
                        // if (k != null
                        //     && k.getValue().equalsIgnoreCase("T")) {
                        // this file is expected to have extensions
                        //     hasExtensions = true;
                        // }
                        hasExtensions = true;
                    }
                    //
                    // see if we have enough info to find the next 
                    // extension, otherwise, mark file as having no
                    // extensions.
                    //
                    long bytesToNextHeader =
                        calculateByteOffsetToNextHeader(naxisKeys);
                        
                    // clear the dictionary now we have finished with it.
                    naxisKeys = new Hashtable();

                    if (!hasExtensions || bytesToNextHeader < 0L) {
                        //
                        // we are done - extensions were not found.
                        //
                        break;
                    }
                    // 
                    // the offset to the next header is calculated from the end of this block,
                    // so we need to calculate how far to skip to the end of this block.
                    //
                    long bytesToEndOfBlock = totalRead % BOUNDARY;
                    bytesToEndOfBlock =
                        bytesToEndOfBlock > 0L
                            ? BOUNDARY - bytesToEndOfBlock
                            : 0L;

                    bytesToNextHeader += bytesToEndOfBlock;

                    seek(input, bytesToNextHeader);
                    totalRead += bytesToNextHeader;
                }
                //
                // we have a keyword to parse
                //
                KeywordToken k = parseKeyword(rec, currentExtension, currentKeywordIndex);
                // System.out.println("Keyword: "+(k == null ? "null" : k.toString())+".");
                if (k != null) {
                    // see if it is a keyword in which we are interested:
                    if (k.getName().equalsIgnoreCase(BITPIX) || k.getName().startsWith(NAXIS)) {
                        // put all BITPIX and NAXIS* keys in a hashtable for use in calculating
                        // distance to next extension
                        naxisKeys.put(k.getName(), k);
                    }
                    
                    // notify all listeners.
                    Iterator i = keywordListeners.iterator();
                    boolean stopReading = false;
                    while (i.hasNext()) {
                        stopReading = stopReading || ((KeywordListener)i.next()).tokenRead(k);
                    }
                    
                    if (stopReading) {
                        // a listener requested to stop reading.
                        break;
                    }
                    // increment keyword index 
                    currentKeywordIndex++;
                }

                bBuffered = 0;
            }
        } // end of while loop.
        // close the reader
        if (otherReader == null) {
            // only clode the input reader if it was not provided
            // by the caller.
            input.close();
        }
        // close the data source.
        dataSource.close();
    }

    /**
     * @param dataAddress address of file to unzip.
     * @return Reader (buffered) to a GZip stream.
     * @throws IOException on error.
     * @throws FileNotFoundException on error.
     */
    private Reader gzipFile(String dataAddress)
        throws IOException, FileNotFoundException {
        BufferedReader buffRed = null;
        File f = new File(dataAddress);
        InputStream is = null;
        try {
            is = new MemMappedFileInputStream(f);
        } catch (IOException e) {
            System.out.println("Error memory mapping file: " + e.getMessage());
            //System.out.println(
	    //"Memory mapping of files is not possible and will be disabled.");
            //FitsTokenizer.setUseMemoryMapping(false);
        }
        if (is == null) {
            is = new FileInputStream(f);
        }
        buffRed =
            new BufferedReader(new InputStreamReader(new GZIPInputStream(is)));
        return buffRed;
    }

    /**
     * @param dataAddress address of file to zcat.
     * @return Reader to the zcat stream (buffered).
     * @throws IOException on error.
     */
    private Reader zcatFile(String dataAddress) throws IOException {
        return new BufferedReader(ZCatReader.createZCatReader(dataAddress));
    }

    /**
     * @param dataAddress address of file to zcat.
     * @return Reader to the memory mapped byte stream.
     */
    private Reader memMapFile(String dataAddress) {
        Reader rc = null;
        try {
            File f = new File(dataAddress);
            rc = new MemMappedFileReader(f);
        } catch (IOException e) {
            System.out.println("Error memory mapping file: " + e.getMessage());
            //System.out.println(
	    //"Memory mapping of files is not possible and will be disabled.");
            //setUseMemoryMapping(false);
        }
        return rc;
    }

    /**
     * Returns the useMemoryMapping.
     * @return boolean
     */
    public static boolean isUseMemoryMapping() {
        return useMemoryMapping;
    }

    /**
     * Sets the useMemoryMapping.
     * @param useMemoryMapping The useMemoryMapping to set
     */
    public static void setUseMemoryMapping(boolean useMemoryMapping) {
        FitsTokenizer.useMemoryMapping = useMemoryMapping;
    }

    /**
     * @param s The text data read from the Fits file header
     * @return The Keyword object representing the text data, or null
     * if no Keyword could be parsed from the data.
     */
    private KeywordToken parseKeyword(
        String s,
        int currentExtension,
        int currentIndex) {
            
        String name = null;
        String value = null;
        KeywordToken rc = null;
        int eqPos = s.indexOf((int) '=');
        if (eqPos <= 0) {
            return rc;
        }
        try {
            name = s.substring(0, eqPos);
            name = name.trim();
        } catch (StringIndexOutOfBoundsException e) {
            // System.out.println("Name extraction failed for:<"+s+">");
            return rc;
        }
        int commPos = s.indexOf(COMMENT_TOKEN, eqPos);
        if (commPos < 0) {
            commPos = s.length();
        }
        try {
            value = s.substring(eqPos + 1, commPos);
            value = value.trim();
        } catch (StringIndexOutOfBoundsException e) {
            // System.out.println("value extraction failed for:<"+s+">");
            return rc;
        }
        rc = new KeywordToken(name, value, currentExtension, currentIndex);
        // System.out.println(rc.toString());
        return rc;
    } // end of method parseKeyword().

    /**
     * Move the rdr along until it has skipped the required bytes (or thrown an exception).
     * @param rdr
     * @param skip
     * @throws IOException
     */
    protected static void seek(Reader rdr, long skip) throws IOException {
	long skipped = 0;
        //System.out.println("SEEK: " + skip);
        try {
            while (skip > 0L) {
                skipped = rdr.skip(skip);
		if (skipped <= 0) {
		    System.out.println("Skipped <= 0: exit");
		    break;
		}
		skip -= skipped;
            }
        } catch (IllegalArgumentException e) {
            // seek error - possibly invalid/truncated file.
            throw new IOException(e.getMessage());
        }
    }

    /**
     * Calculate the number of bytes to start of the next extension.
     * @param Dictionary the keyword set to use to calculate the offset - assumes not-null.
     * @return long the byte offset to the start of the next header, or -1L
     * if calculation not possible.
     */
    public static long calculateByteOffsetToNextHeader(Dictionary naxisKeys) {
                
        long pixelSize = 0L;
        long axis[] = null;
        
        // go through the dictionary of keyword tokens and find the values
        // we need, if present.
        KeywordToken k;
        k = (KeywordToken)naxisKeys.get(BITPIX);
        if (k != null) {
            pixelSize = stringToLong(k.getValue(), 0L);
            // System.out.println("BITPIX = "+pixelSize);
        }
        k = (KeywordToken)naxisKeys.get(NAXIS);
        if (k != null) {
            int naxis = (int)stringToLong(k.getValue(), -1L);
            // System.out.println("NAXIS = "+naxis);
            if (naxis > 0L) {
                axis = new long[naxis];
            }
        }
        if (axis != null) {
            for (int i = 0; i < axis.length; i++) {
                long axisVal = -1L;
                k = (KeywordToken)naxisKeys.get(NAXIS+(i+1));
                if (k != null) {
                    axisVal = stringToLong(k.getValue(), -1L);
                }
                if (axisVal < 0) {
                    // set pixelSize to 0 to indicate calculation has failed.
                    pixelSize = 0L;
                    break;
                } else {
                    axis[i] = axisVal;
                    // System.out.println("NAXIS"+(i+1)+" = "+axisVal);
                }
            }
        }
        
        long rc = -1L;

        // if axis == null or pixelSize == 0L then the calculation can not be performed.
        if (axis != null || pixelSize != 0) {
            // interested in the absolute value of the pixel size
            pixelSize = Math.abs(pixelSize);

            long imageSize = 0L;
            for (int i = 0; axis != null && i < axis.length; i++) {
                if (imageSize == 0L) {
                    imageSize = 1L;
                }
                imageSize *= axis[i];
            }
            if (imageSize >= 0L) {
                //
                // pixelSize (in bytes, not bits)*imageSize 
                // = imageSize in bytes.
                //
                imageSize = pixelSize / 8 * imageSize;

                //
                // Calculate the amount of space left empty at the end of the
                // last block - extensions only start on block boundaries, so
                // we need to add the unused space in the last block to the imageSize
                // to get the byte offset to the next extension.
                //
                long byteUsedInLastBlock = imageSize % BOUNDARY;
                long unusedBytesInBlock =
                    byteUsedInLastBlock > 0L
                        ? BOUNDARY - byteUsedInLastBlock
                        : 0L;

                //
                // Offset to next extension, in bytes, is the imageSize + unused bytes
                // in the last block, if any.
                //
                rc = imageSize + unusedBytesInBlock;

            } // else no image size calculated.
        }
        return rc;
    }
    
    /**
     * Utility function to convert a string to a long.
     * @param in
     * @return long value of string, or errorValue on failure
     */
    private static long stringToLong(String in, long errorValue) {
        long rc;
        try {
            rc = Long.parseLong(in);
        } catch (NumberFormatException e) {
            rc = errorValue;
        }
        return rc;
    }

    /**
     * Example program and test for this class suite.
     * 
     * run this class with following parameters:
     * 1. the path to a fits file,
     * either normal (.fits), zipped (.fits.Z) or gzipped (.fits.gz).
     * 2. "all": optional parameter, will print out all extensions if given.
     * 
     * Program will print out the
     * 
     * @param args
     */
    public static void main(String[] args) {
        
        if (args.length < 1) {
            System.out.println("usage: <file path to fits file> [all]");
            return;
        }
        boolean doAll = false;
        String filePath = args[0];
        if (args.length >= 2 && args[1].equalsIgnoreCase("all")) {
            doAll = true;
        }

        try {
            FitsTokenizer ft = new FitsTokenizer(filePath, doAll);
            ft.addKeywordListener(new KeywordListener() {
                public boolean tokenRead(KeywordToken kwt) {
                    System.out.println("Ext: "+kwt.getExtension()+"\tidx:"+kwt.getIndexInExtension()
                        +"\t"+kwt.getName()+"\t"+kwt.getValue());
                    return false;
                }
                
            });
            ft.read();
        } catch (IOException e) {
            e.printStackTrace();
        }
        
    }

}
