package org.eso.gasgano;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;
import javax.swing.border.EmptyBorder;

import org.eso.gasgano.gui.DBDisplayPanel;
import org.eso.gasgano.gui.DBPanel;
import org.eso.gasgano.gui.DataPanel;
import org.eso.gasgano.gui.DisplayPanel;
import org.eso.gasgano.gui.GUIUtils;
import org.eso.gasgano.gui.PropertyEditorSubPanel;
import org.eso.gasgano.gui.RecipePropertyPanel;
import org.eso.gasgano.properties.GasProp;

/**
 * GasPropEditor - a class to edit a GasProp object
 * main constructor to be called as GasPropEditor(GasProp e).
 * 
 * The properties are updated and saved (all done by this class)
 * when the OK button is pressed.
 * 
 * This class is strongly coupled  to the class GasProp
 * and any changes in the latter should be reflected here
 *
 * There is lots of code repetition which could be probably avoided
 * The repaint of the Font Test string does not work
 * 
 * main method to be used after calling the constructor is 
 * startEditing(), which will make the frame visible and return 
 * a boolean value according to the changes being accepted or discarded
 * @author Michele Zamparelli, Nick Kornweibel
 * @see GasProp
 */

public class GasPropEditor extends JFrame implements ActionListener {

    private static final int UserCancelled = -1;
    private JButton okButton = new JButton("OK");
    private JButton cancelButton = new JButton("Cancel");

    private boolean debug = false;

    // all the panels for the JTabbedPane
    private PropertyEditorSubPanel panels[];
    
    // just a convenience reference to one of the panels
    private DataPanel dataPanel;
    
    private ReclassificationListener reclassificationListener;

    /**
     * 
     * @author nkornwei
     * the "copy" of the actuall properties - this becomes the actual
     * property DB if the user saves the changes made.
     */
    public GasProp gpCopy;

    /**
     * Constructor.
     */
    public GasPropEditor() {
        super("Preferences Editor");

        JPanel buttonPanel = new JPanel();
        JTabbedPane jtp = new JTabbedPane(SwingConstants.TOP);

        gpCopy = new GasProp(GasProp.getProperties());

        panels = new PropertyEditorSubPanel[5];
        dataPanel = new DataPanel(gpCopy);
        panels[0] = dataPanel;
        panels[1] = new DisplayPanel(gpCopy);
        panels[2] = new RecipePropertyPanel(gpCopy);
        panels[3] = new DBDisplayPanel(gpCopy);
        panels[4] = new DBPanel(gpCopy);
        

        for (int i = 0; i < panels.length; i++) {
            panels[i].setDebug(debug);
        }
        // 
        jtp.setBorder(new EmptyBorder(3, 3, 3, 3));

        jtp.addTab(
            "Data",
            null,
            panels[0],
            "File set, Filesystem settings, Verbose level");
        jtp.addTab(
            "File Display & Grouping",
            null,
            panels[1],
            "Display options, Grouping, External Viewers, Fonts, Printing");
        jtp.addTab(
            "Recipe Configuration",
            null,
            panels[2],
            "Configure the set of recipes available.");
        if (GasProp.isEnvironmentSafe()) {
            jtp.addTab(
                "Database Display",
                null,
                panels[3],
                "Database display options, Custom pulldown menus");
            jtp.addTab(
                "Databases",
                null,
                panels[4],
                "Database access parameters for Observations and OB Status");
        }

        jtp.setSelectedIndex(0);

        buttonPanel.add(okButton);
        buttonPanel.add(cancelButton);
        okButton.addActionListener(this);
        cancelButton.addActionListener(this);

        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(jtp, BorderLayout.CENTER);
        getContentPane().add(buttonPanel, BorderLayout.SOUTH);
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                // equivalent to pressing cancel button
                setVisible(false);
            }
        });
        pack();
        setResizable(false);
        setLocation(100, 300);
    }

    /**
     * Load the GUI frame with the current preferences
     * taken from the GasProp instance.
     */
    public void loadProperties() {
        gpCopy = new GasProp(GasProp.getProperties());
        for (int i = 0; i < panels.length; i++) {
            panels[i].loadProperties(gpCopy);
        }
    }

    /**
     * TODO: the logic to decide on a shutdown does not belong here - it
     * should be a decision of the GasPropListener or a decission of the
     * GasProp instance its self. For now OK will save all properties
     * back to the main properties database.
     * 
     * GUI Acton handler.
     * @see java.awt.event.ActionListener#actionPerformed(ActionEvent)
     */
    public void actionPerformed(ActionEvent e) {
        if (e.getSource() == okButton) {
            // user is happy with the changes and the modifications should
            // be applied to the GasProp instance.
            int response = 0;
            boolean shutdown = false;
            boolean reload = false;
            
            for (int i = 0; i < panels.length; i++) {
                int action = panels[i].restartRequired();
                if (action == PropertyEditorSubPanel.RELOAD_ON_SAVE) {
                    reload = true; 
                } 
                if (action == PropertyEditorSubPanel.RESTART_ON_SAVE) {
                    shutdown = true; 
                } 
                // want to find out the most drastic action, where restart is more
                // drastic than reload. 
                if (response < action) {
                    response = action;
                }
            }
            
            if (response == PropertyEditorSubPanel.RELOAD_ON_SAVE) {
                if (JOptionPane
                    .showConfirmDialog(
                        this,
                        "These changes require reloading the data set.\n"
                            + "OK to continue?",
                        "Data Reload Required",
                        JOptionPane.OK_CANCEL_OPTION)
                    != JOptionPane.OK_OPTION) {
                        response = UserCancelled;
                        shutdown = false;
                        reload = false;
                }
            } else if (response == PropertyEditorSubPanel.RESTART_ON_SAVE) {
                if (JOptionPane
                    .showConfirmDialog(
                        this,
                        "These changes will take effect after Gasgano is restarted.\n"
                            + "OK to continue?",
                        "Restart Required",
                        JOptionPane.OK_CANCEL_OPTION)
                    != JOptionPane.OK_OPTION) {
                        response = UserCancelled;
                        shutdown = false;
                        reload = false;
                }
            }
            
            if (response == UserCancelled) {
                // cancel changes by user
                return;
            }
            // force all Panels to write down what's in their JTextFields
            for (int i = 0; i < panels.length; i++) {
                String error = panels[i].saveProperties();
                if (error != null) {
                    GUIUtils.userError(this, error);
                    panels[i].loadProperties(gpCopy);
                    // error - return to the property editor
                    return;
                }
            }

            // make sure that all vectors modified with the panels
            // are put back into string properties
            gpCopy.dumpVectors();

            if (debug) {
                System.out.println("will reset the preferences");
            }

            setVisible(false);
            // set the new properties, reloading if required.
            // the shutdown does not get enforced - it is up to the user to do that
            GasProp.setProperties(gpCopy, reload);

            // now that the properties have been all saved, check if it is
            // necessary to reclassify the files
	    	if( dataPanel.isClassificationRequired() ) {
	        	reclassificationListener.reclassifyNow();
	    	}
            
        } else if (e.getSource() == cancelButton) {
            setVisible(false);
        }
    }

	public void startEditing() {
        setVisible(true);
    }

    /*
     * Small test harness....
     */
    public static void main(String args[]) {
        test1(args);
        test2(args);
    }

    private static void test1(String args[]) {
        // example of file constructor
        GasProp gp = new GasProp();
        if (args.length == 1) {

            if (!gp.loadProperties(args[0])) {
                System.out.println(
                    "failed to load " + args[0] + " properties file");
            }
        }
        // example of copy constructor
        GasProp gp2 = new GasProp(gp);

        GasPropEditor pe = new GasPropEditor();
        pe.loadProperties();
        pe.startEditing();
    }

    private static void test2(String args[]) {
        // testing part
        // example of file constructor
        GasProp gp = new GasProp();

        if (args.length == 1) {

            if (!gp.loadProperties(args[0])) {
                System.out.println(
                    "failed to load " + args[0] + " properties file");
            }
        }
        // example of copy constructor
        GasProp gp2 = new GasProp(gp);
        gp2.instruments.setElementAt("CAZZO", 0);
        System.out.println(gp2.instruments.elementAt(0));
        System.out.println(gp.instruments.elementAt(0));

        if (gp2.getProperty("INSTRUMENT_GROUPING")
            == gp.getProperty("INSTRUMENT_GROUPING")) {
            System.out.println("the two properties are the same object");
        }

        if (gp2.instruments.elementAt(0) == gp.instruments.elementAt(0)) {
            System.out.println(
                "I did something very, very wrong (instruments)");
        }
        if (gp2.instrumentGrouping.elementAt(0)
            == gp.instrumentGrouping.elementAt(0)) {
            System.out.println("I did something very very wrong");
        }
        System.exit(0);
        // example of setting the property
        System.out.println(gp.getProperty("OB_STATUS_FROM_DB_REQUIRED"));
        GasPropEditor pe = new GasPropEditor();
        while (true) {
            pe.loadProperties();
            pe.startEditing();
        }
    }

    /**
     * Sets the reclassification listener.
     * @param reclassificationListener
     */
    public void setReclassificationListener(ReclassificationListener reclassificationListener) {
    	this.reclassificationListener = reclassificationListener;
    }
    
    /**
     * Simple listener used to inform the caller that a reclassification
     * of the files is required.
     */
    public interface ReclassificationListener {
    	public void reclassifyNow();
    }
}
