/**
 * @author N. Kornweibel
 * @version 1.0
 * An abstract base class for defining a component of an Observation.
 */

package org.eso.gasgano.datamodel.gui;

import gnu.regexp.RE;
import gnu.regexp.REException;
import gnu.regexp.REMatch;
import gnu.regexp.RESyntax;

import java.awt.Font;
import java.io.File;
import java.util.Vector;

import javax.swing.JComponent;

import org.eso.gasgano.datamodel.DataModelChangeNotifier;
import org.eso.gasgano.datamodel.DataModelEvent;
import org.eso.gasgano.datamodel.FileDescription;
import org.eso.gasgano.datamodel.Observation;
import org.eso.gasgano.datamodel.ReductionBlockCommon;
import org.eso.gasgano.datamodel.filesystem.DFSDataModel;
import org.eso.gasgano.datamodel.filesystem.DFSFile;
import org.eso.gasgano.properties.GasProp;
import org.eso.gasgano.properties.PropertyDB;

public abstract class OBComponent extends DataModelChangeNotifier {

    private Observation[] obs = new Observation[3];
    protected String storeProgram = null;
    protected String storeObservation = null;
    protected String storeTelescope = null;

    // set of ReductionBlockOBComponents using this OBComponent.
    private Vector associatedReductionBlocks = null;

    private static FileDescription descFactory = null;

    private static RESyntax theSyntax = new RESyntax(RESyntax.RE_SYNTAX_PERL5);
    private static String baseNamePure = "[^" + File.separator + "]*$";
    private static String baseNameWithDatum =
        "\\d{4}-\\d{2}-\\d{2}" + File.separator + "[^" + File.separator + "]*$";
    private static REMatch theMatch = null;
    public static RE baseNameRe = null;
    private static boolean shortenName = true;

    private boolean isPipelineProduct = false;

    // font for use in the Label.
    private Font labelFont = null;

    public OBComponent() {
        if (baseNameRe == null) {
            try {
                baseNameRe = new RE(baseNamePure);
            } catch (REException ree) {
                // fall back to false is no RE can be instantiated
                shortenName = false;
            }
        }
    }

    static public void setShortName(boolean t) {
        shortenName = t;
    }

    /**
     * @param p True if this is a data product from the DFS Pipeline.
     */
    public void setPipeProduct(boolean p) {
        isPipelineProduct = p;
    }

    /**
     * @param rb The ReductionBlockOBComponent to be added to the list.
     */
    public void addAssociatedReductionBlock(ReductionBlockCommon rb) {
        if (associatedReductionBlocks == null) {
            associatedReductionBlocks = new Vector();
        }
        boolean alreadyStored = false;
        for (int i = 0; i < associatedReductionBlocks.size(); i++) {
            if (associatedReductionBlocks.elementAt(i) == rb) {
                alreadyStored = true;
                break;
            }
        }
        if (!alreadyStored) {
            associatedReductionBlocks.addElement(rb);
        }
    }

    /**
     * @param rb The ReductionBlockOBComponent to be deleted from the list.
     */
    public void removeAssociatedReductionBlock(ReductionBlockCommon rb) {
        if (associatedReductionBlocks != null) {
            associatedReductionBlocks.removeElement(rb);
        }
    }

    /**
     * @return The Vector of associated ReductionBlockOBComponents
     * or null if there are none.
     */
    public Vector getAssociatedReductionBlocks() {
        return associatedReductionBlocks;
    }

    /**
     * @return The status of this component.
     * @see org.eso.gasgano.DFSFile
     */
    public abstract int getStatus();

    /**
     * @param o The observation under which this component belongs.
     */
    public void setObservation(Observation o, int mode) {
        obs[mode] = o;
    }

    /**
     * @return The observation under which this component belongs.
     */
    public Observation getObservation(int mode) {
        return obs[mode];
    }

    /**
     * @return True if this component is a data product from the DFS
     * Pipeline.
     */
    public boolean isPipeProduct() {
        return isPipelineProduct;
    }

    public abstract DFSFile getSourceFile();

    /**
     * @return The current status after executing the update.
     * Check if the file requires updating, update the data
     * and return the new DFSFile status.
     * @see org.eso.gasgano.DFSFile
     */
    public abstract int update();

    /**
     * @return The JComponent representing the component data.
     */
    public abstract JComponent getDisplayComponent();

    public abstract JComponent getTreeCellComponent();

    /**
     * @return The unique id for this component 
     * - typically the file name, without the path.
     */
    public abstract String getId();

    public String toString() {
        String rc = null;
        // set up all needed expressions (should be done in <init>)
        if (shortenName) {
            // Paola doesn't like it
            // theMatch = baseNameRe.getMatch(getId());
            // rc = theMatch.toString();
            rc = getId().substring(getId().lastIndexOf(File.separator) + 1);
        } else {
            rc = getId();
        }
        return rc;
    }

    /**
     * @param d The FileDescription object used to set the derscriptions of
     * all OBComponent objects.
     */
    public static void setDefaultFileDescription(FileDescription d) {
        descFactory = d;
    }

    /**
     * @return The FileDescription object used to set the derscriptions of
     * all OBComponent objects.
     */
    public static FileDescription getDefaultFileDescription() {
        return descFactory;
    }

    /**
     * Take the necessary actions to detach this object from the
     * DataModel and it components - typically called before an object is
     * deleted.
     * NOTE: This method _must_ be called even if over-ridden.
     */
    public void cleanup() {
        //System.out.println("OBComponent cleanup for "+getId());
        // inform listeners that the component is being removed.
        setObservation(null, GasProp.TREE_TELESCOPE);
        setObservation(null, GasProp.TREE_DIRECTORY);
        setObservation(null, GasProp.TREE_DEFAULT);
        fireDataModelChangeEvent(
            new DataModelEvent(DataModelEvent.FILE_DELETED, this));
    }

    /**
     * The method will remove this from the DataModel.
     */
    public void removeFromDataModel() {
        DFSDataModel dm = DFSDataModel.getDataModel();
        // the following call will result in the DataModel calling
        // removeFromModel() on this.
        dm.removeObComponent(this);
    }

    public void setLabelFont(Font f) {
        labelFont = f;
    }

    public Font getLabelFont() {
        if (labelFont == null) {
            DFSDataModel dm = DFSDataModel.getDataModel();

            String fontString =
                (String) (PropertyDB
                    .getInstance()
                    .getProperty("COMPONENT_FONT"));
            String fontName =
                fontString.substring(0, fontString.indexOf(":", 0));
            String fontSize =
                fontString.substring(
                    fontString.indexOf(":", 0) + 1,
                    fontString.length());

            labelFont =
                new Font(fontName, Font.PLAIN, Integer.parseInt(fontSize));
            // new Font("Dialog", Font.PLAIN, 12);
        }
        return labelFont;
    }

    /**
     * @todo WHAT IS THIS????? The functionality required for
     * HasHeaderOBComponent should be implemented in HasHeaderOBComponent.
     * I do not understand what it is doing or why it is there.
     * @return String
     */
    public String getIntermediateGrouping() {
        String rc = null;
        if (this instanceof KeywordContainerOBComponent) {
            rc = ((KeywordContainerOBComponent) this).getIntermediateGrouping();
        }
        return rc;
    }

} // end of class OBComponent.java

// end of file OBComponent.java
