package org.eso.cpl;

import java.util.List;
import java.io.File;

import org.eso.dfs.services.ServiceException;
import org.eso.dfs.services.data.AssociationService;
import org.eso.dfs.services.data.FileAssociationNode;

/**
 * @author Nick Kornweibel.
 *
 * Utility static methods for accessing the AssociationService singleton.
 */
public class AssociationUtils {

    /**
     * Using the Data Association service, classify all files and return an array
     * of FileAssociationNode objects.
     * Files that can not be classified are ignored. 
     * @param files
     * @return FrameList
     */
    public static FileAssociationNode[] classifyFiles(List files) {
        AssociationService service = AssociationService.getInstance();
        FileAssociationNode[] nodes = new FileAssociationNode[files.size()];
        for (int i = 0; i < nodes.length; i++) {
            nodes[i] = new FileAssociationNode((File)files.get(i));
        }
        try {
            service.processNodes(nodes);
        } catch (ServiceException e) {
            e.printStackTrace();
        }
        return nodes;
    }
    
    /**
     * Using the Data Association service, classify all files and place them
     * as Frames in the FrameList. Files that can not be classified are ignored. 
     * @param files
     * @return FrameList
     */
    public static FrameList classifyFrames(List files) {
        return getFramesFromNodes(classifyFiles(files));
    }
    
    /**
     * Take a set of FileAssociationNode objects and return a FrameList of Frame 
     * objects, built from the FileAssociationNodes.
     * @param nodes
     * @return FrameList
     */
    public static FrameList getFramesFromNodes(FileAssociationNode[] nodes) {
        FrameList rc = new FrameList();
        for (int i = 0; i < nodes.length; i++) {
            Frame frm = getFrameFromNode(nodes[i]);
            if (frm.getTag() != null) {
                // frame was processed and classified..add it to the list.
                rc.add(frm);
            }
        }
        return rc;
    }
    
    /**
     * Convert a FileAssociationNode to a Frame.
     * @param node
     * @return Frame
     */
    public static Frame getFrameFromNode(FileAssociationNode node) {
        Frame frm = new Frame(node.getAddress());
        String classification = null;
        if (node.getClassification() != null && 
            node.getClassification().length > 0) {
                classification = node.getClassification()[0];
        }
        if (classification != null) {
            frm.setTag(classification); 
        }
        // default to RAW category.
        frm.setGroup((node.getCategory() != null 
                && node.getCategory().equals("CALIB")) ?
            FrameGroup.CALIB : FrameGroup.RAW);
        frm.setLevel(FrameLevel.INTERMEDIATE);
        frm.setType(FrameType.IMAGE);
        return frm;
    }
}
