package org.eso.cpl;

import java.io.File;
import java.io.IOException;

/**
 * A frame which forms the input or output data for a recipe invocation.
 * This corresponds to a FITS file.
 *
 * @author   Mark Taylor
 */
public class Frame {

    private File file_;
    private String tag_;
    private FrameType type_;
    private FrameGroup group_;
    private FrameLevel level_;
    private String canonicalPath_;

    /**
     * Constructs a new blank Frame.
     */
    public Frame() {
    }

    /**
     * Constructs a new Frame with a given filename.
     *
     * @param  filename  the frame's filename
     */
    public Frame(File file) {
        setFile(file);
    }

    /**
     * Clone constructor.  This constructs a new Frame with exactly the
     * same characteristics as an existing one.  The new one will match
     * the old one for the purposes of the <tt>equals</tt> method, but
     * modifying the fields of the one will not affect the other.
     *
     * @param  frame  frame to copy
     */
    public Frame(Frame frame) {
        this();
        setFile(frame.getFile());
        setTag(frame.getTag());
        setType(frame.getType());
        setGroup(frame.getGroup());
        setLevel(frame.getLevel());
    }

    public File getFile() {
        return file_;
    }
    
    public String getCanonicalPath() {
        return canonicalPath_;
    }

    /**
     * Set the file and canonical path members. The canonical path
     * is set if possible (if there is no IOException).
     * This is perhaps not ideal - there is no way to know of
     * an error if the canonical path is not correctly set.
     * @param file
     */
    public void setFile(File file) {
        file_ = file;
        try {
            canonicalPath_ = file_.getCanonicalPath();
        }
        catch (IOException e) {
            canonicalPath_ = file_.getAbsolutePath();
        }
    }

    public String getTag() {
        return tag_;
    }

    public void setTag(String tag) {
        tag_ = tag;
    }

    public FrameType getType() {
        return type_;
    }

    public void setType(FrameType type) {
        type_ = type;
    }

    public FrameGroup getGroup() {
        return group_;
    }

    public void setGroup(FrameGroup group) {
        group_ = group;
    }

    public FrameLevel getLevel() {
        return level_;
    }

    public void setLevel(FrameLevel level) {
        level_ = level;
    }

    /**
     * Indicates whether two Frames are the same.
     *
     * @return <tt>true</tt> iff this frame and <tt>obj</tt> have matching
     *         values for all fields
     */
    public boolean equals(Object obj) {
        if (!getClass().equals(obj.getClass())) {
            return false;
        }
        Frame other = (Frame) obj;
        return sameValue(canonicalPath_, other.canonicalPath_) &&
               sameValue(tag_, other.tag_) &&
               sameValue(group_, other.group_) &&
               sameValue(level_, other.level_) &&
               sameValue(type_, other.type_);
    }

    /**
     * Overridden for consistency with the overridden <tt>equals</tt> method.
     *
     * @return  hash code
     */
    public int hashCode() {
        int result = 23;
        result = 17 * result + hashValue(canonicalPath_);
        result = 17 * result + hashValue(tag_);
        result = 17 * result + hashValue(group_);
        result = 17 * result + hashValue(level_);
        result = 17 * result + hashValue(type_);
        return result;
    }

    /**
     * Utility function to check equality of values.  Does the same job
     * as <tt>equals</tt> but works for <tt>null</tt> values too.
     *
     * @param  o1  first value
     * @param  o2  second value
     * @return  <tt>true</tt> iff <tt>o1</tt> and <tt>o2</tt> match
     */
    private static boolean sameValue(Object o1, Object o2) {
        return (o1 == null && o2 == null) ||
               (o1 != null && o1.equals(o2));
    }

    /**
     * Utility function to return a hash value for an object.  Does the
     * same job as <tt>hashCode</tt>, but works for <tt>null</tt> values too.
     *
     * @param  o  object
     * @return  hash code
     */
    private static int hashValue(Object o) {
        return o == null ? -1 : o.hashCode();
    }

    /**
     * Returns a string representation of this frame, containing all the
     * information about it.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(getFile().getPath());
        if (group_ != null) {
            buf.append(" group=").append(group_);
        }
        if (level_ != null) {
            buf.append(" level=").append(level_);
        }
        if (type_ != null) {
            buf.append(" type=").append(type_);
        }
        if (tag_ != null) {
            buf.append(" tag=\"").append(tag_).append('"');
        }
        return buf.toString();
    }
}
