package org.eso.cpl;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameLevel;
import org.eso.cpl.FrameType;

/**
 * Groups a list of {@link Frame} objects together.  This provides the 
 * following features additional to that of the {@link java.util.List}
 * interface which it implements:
 * <ul>
 * <li>All elements of the list are guaranteed to be <tt>Frame</tt>s 
 *     (or <tt>null</tt>). 
 *     Any attempt to insert a non-<tt>Frame</tt> object will fail
 *     with a <tt>ClassCastException</tt>.
 * <li>The {@link #getFrame} and {@link #toFrameArray} methods are provided
 *     which are typed versions of corresponding <tt>List</tt> methods.
 * <li>Some <tt>selectBy*</tt> methods are provided
 *     to select elements of the list by values of their fields.
 * </ul>
 *
 * @author   Mark Taylor
 */
public class FrameList extends AbstractList {

    private final List list_;

    /**
     * Constructs a new empty FrameList.
     */
    public FrameList() {
        list_ = new ArrayList();
    }

    /**
     * Constructs a new FrameList from a collection of Frames.
     *
     * @param   frames  a collection of Frame objects whose contents form
     *          the initial contents of this list
     * @throws  ClassCastException  if any element of <tt>frames</tt> is 
     *          a non-Frame object
     */
    public FrameList(Collection frames) {
        this();
        addAll(0, frames);
    }

    /**
     * Constructs a new FrameList from an array of Frames.
     *
     * @param  frames  an array of Frame objects whose elements form
     *         the initial contents of this list
     */
    public FrameList(Frame[] frames) {
        this(Arrays.asList(frames));
    }

    /**
     * Returns the Frame at a given index in the list.
     * This is just a typed version of {@link #get}.
     * 
     * @param  i  index to interrogate
     * @return  the Frame at index <tt>i</tt> in this list
     */
    public Frame getFrame(int i) {
        return (Frame) list_.get(i);
    }

    /**
     * Returns the elements of the list in an array.
     * This is just a typed version of {@link #toArray}.
     *
     * @return  a Frame array with the same contents as this list
     */
    public Frame[] toFrameArray() {
        return (Frame[]) list_.toArray(new Frame[list_.size()]);
    }

    /**
     * Returns a new FrameList containing all the elements of this one 
     * with a given tag.
     *
     * @param  tag  the tag to match
     * @return  all those frames in this list with a tag equal to <tt>tag</tt>
     */
    public FrameList selectByTag(String tag) {
        FrameList selected = new FrameList();
        for (Iterator it = iterator(); it.hasNext();) {
            Frame frame = (Frame) it.next();
            if (sameValue(frame.getTag(), tag)) {
                selected.add(frame);
            }
        }
        return selected;
    }

    /**
     * Returns a new FrameList containing all the elements of this one
     * with a given FrameGroup.
     *
     * @param  group the group to match
     * @return  all those frames in this list with a group equal to 
     *          <tt>group</tt>
     */
    public FrameList selectByGroup(FrameGroup group) {
        FrameList selected = new FrameList();
        for (Iterator it = iterator(); it.hasNext();) {
            Frame frame = (Frame) it.next();
            if (sameValue(frame.getGroup(), group)) {
                selected.add(frame);
            }
        }
        return selected;
    }

    /**
     * Returns a new FrameList containing all the elements of this one
     * with a given FrameLevel.
     *
     * @param  level  the level to match
     * @return  all those frames in this list with a level equal to 
     *          <tt>level</tt>
     */
    public FrameList selectByLevel(FrameLevel level) {
        FrameList selected = new FrameList();
        for (Iterator it = iterator(); it.hasNext();) {
            Frame frame = (Frame) it.next();
            if (sameValue(frame.getLevel(), level)) {
                selected.add(frame);
            }
        }
        return selected;
    }

    /**
     * Returns a new FrameList containing all the elements of this one
     * with a given FrameType.
     *
     * @param   type  the type to match
     * @return  all those frames in this list with a type equal to 
     *          <tt>type</tt>
     */
    public FrameList selectByType(FrameType type) {
        FrameList selected = new FrameList();
        for (Iterator it = iterator(); it.hasNext();) {
            Frame frame = (Frame) it.next();
            if (sameValue(frame.getType(), type)) {
                selected.add(frame);
            }
        }
        return selected;
    }

    /**
     * Checks equality; like <tt>equals</tt>, but doesn't choke on 
     * <tt>null</tt>s.
     */
    private boolean sameValue(Object o1, Object o2) {
        return (o1 == null && o2 == null) ||
               (o1 != null && o1.equals(o2));
    }

    /*
     * The remaining methods are those required to implement a variable
     * length modifiable list.
     */

    public int size() {
        return list_.size();
    }

    public Object get(int i) {
        return list_.get(i);
    }

    /**
     * @throws  ClassCastException  if <tt>element</tt> is not a Frame
     */
    public Object set(int i, Object element) {
        return list_.set(i, (Frame) element);
    }

    /**
     * @throws  ClassCastException  if <tt>element</tt> is not a Frame
     */
    public void add(int i, Object element) {
        list_.add(i, (Frame) element);
    }

    public Object remove(int i) {
        return list_.remove(i);
    }

}
