package org.eso.cpl;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * Represents a data type which a parameter may assume.
 * All the permissible types are provided as static final members of this
 * class.
 * <p>
 * Exemplifies the Typesafe Enum pattern.
 *
 * @author   Mark Taylor
 */
public abstract class ParameterType {

    private String name_;
    private Class clazz_;

    /**
     * Private constructor, only used to instantiate final static members
     * of this class.
     *
     * @param   cplType  the CPL type to which this type corresponds
     * @param   clazz    the Java class used to represent this type
     */
    private ParameterType(String name, Class clazz) {
        name_ = name;
        clazz_ = clazz;
    }

    /**
     * Returns the name of this type.
     *
     * @return   type name
     */
    public String getName() {
        return name_;
    }

    /**
     * Returns the Java class used to represent values that have this type.
     *
     * @param  representation class
     */
    public Class getValueClass() {
        return clazz_;
    }

    /**
     * Provides a String representation of a value of this type.
     * Behaviour is undefined if <tt>value</tt> does not have the 
     * class corresponding to this type.
     *
     * @param   value  the value to format
     */
    public String toString(Object value) {
        return value.toString();
    }

    public String toString() {
        return getName();
    }

    /**
     * Interprets a String as a value of this type, and returns the
     * resulting value.  The returned value will be an instance of
     * {@link #getValueClass}.
     *
     * @param  str  a string representation of an object of this type
     * @return  a value of this type corresponding to <tt>str</tt>
     * @throws  ParameterValueException  if <tt>str</tt> cannot be interpreted
     *          as a value of this type
     */
    public abstract Object fromString(String str) throws ParameterValueException;

    /** Boolean type. */
    public static ParameterType BOOL = new ParameterType("bool",
                                                         Boolean.class) {
        String[] trueStrings_ = {"true", "yes", "1", "t"};
        String[] falseStrings_ = {"false", "no", "0", "f"};
        Set trueValues_ = new HashSet( Arrays.asList(trueStrings_));
        Set falseValues_ = new HashSet( Arrays.asList(falseStrings_));

        public Object fromString(String str) throws ParameterValueException {
            str = str.toLowerCase();
            if (trueValues_.contains(str)) {
                return Boolean.TRUE;
            }
            else if (falseValues_.contains(str)) {
                return Boolean.FALSE;
            }
            else {
                throw new ParameterValueException(str + 
                                                  " not a recognized boolean");
            }
        }
    };

    /** Integer type. */
    public static ParameterType INT = new ParameterType("int",
                                                        Integer.class) {
        public Object fromString(String str) throws ParameterValueException {
            try {
                return Integer.valueOf(str);
            }
            catch (NumberFormatException e) {
                throw new ParameterValueException(str + " not a recognised " +
                                                  "integer", e);
            }
        }
    };

    /** Double type. */
    public static ParameterType DOUBLE = new ParameterType("double",
                                                           Double.class) {
        public Object fromString(String str) throws ParameterValueException {
            try {
                return Double.valueOf(str);
            }
            catch (NumberFormatException e) {
                throw new ParameterValueException(str + " not a recognised " +
                                                  "double", e);
            }
        }
    };

    /** String type. */
    public static ParameterType STRING = new ParameterType("string",
                                                           String.class) {
        public Object fromString(String str) {
            return str;
        }
    };

    /** Unmodifiable list containing all the instances of this class.  */
    public static final List ALL_TYPES = 
        Collections.unmodifiableList(Arrays.asList(new ParameterType[] {
            BOOL, INT, DOUBLE, STRING,
        }));
   
}
