package org.eso.cpl;

/**
 * Defines a scripting syntax.  Objects implementing this interface
 * are used to generate the actual script commands by a 
 * {@link CommandSequence} object.
 * The methods return a string which is to form part of an
 * eventual script.  Most return full lines, but as noted in the method 
 * documentation, some return line fragments.  Those which return full
 * lines should include any required line-end characters.
 * <p>
 * For the sake of tractability, this interface does not attempt to be
 * completely general, but makes some assumptions that are reasonable
 * for scripting-type languages.  It would probably not be possible to
 * write an implementation for FORTRAN 77, for instance.
 * <p>
 * The interaction between this class and CommandSequence is an example
 * of the Bridge pattern.
 *
 * @author   Mark Taylor
 */
public interface ScriptSyntax {

    /**
     * Returns the name of the syntax defined by this object.
     */
    String getName();

    /**
     * Returns a file extension (including a "." if applicable) suitable
     * for storing a script in this syntax.
     */
    String getExtension();

    /**
     * Returns a sequence of (zero or more) lines which should appear at the 
     * start of a script.
     *
     * @return  intro string
     */
    String startScript();

    /**
     * Returns a string representing a visual break in the script
     * (no semantic content).
     *
     * @return blank line string
     */
    String blankLine();

    /**
     * Returns a string containing a given text string, with no semantic
     * content.
     *
     * @param   text  comment text (no line breaks)
     * @return  comment line
     */
    String commentLine(String text);

    /**
     * Returns a string which will exec a given command string.
     *
     * @param  cmd  the command line to execute (words separated by spaces,
     *              no line breaks)
     * @return  command invocation line
     */
    String invokeCommand(String cmd);

    /**
     * Returns a string which will define a short macro.
     * The expansion <tt>value</tt> should be 'short' (one-line); 
     * for potentially long macro values use 
     * {@link #defineMacro(String,String[])}.
     * <p>
     * <tt>name</tt> is assumed to be syntactically acceptable -
     * it's a good idea for it to be the result of a {@link #namify} call.
     * 
     * @param  name  macro name
     * @param  value  macro expansion (as returned by {@link #useMacro}),
     *                no newlines
     * @return  macro definition line
     */
    String defineMacro(String name, String value);

    /**
     * Returns a string which will define a potentially long macro.
     * The macro is set to a space-separated concatenation of the
     * elements of an array <tt>words</tt>.
     * <p>
     * <tt>name</tt> is assumed to be syntactically acceptable -
     * it's a good idea for it to be the result of a {@link #namify} call.
     * 
     * @param  name  macro name
     * @param  words  array of words forming the macro expansion;
     *                no newline characters should appear in the elements
     * @return  macro definition line
     */
    String defineMacro(String name, String[] words);

    /**
     * Dereferences a macro previously defined by one of the 
     * <tt>defineMacro</tt> calls.  Returns a single word which can form
     * part of a line.
     *
     * @param  name  the name of the macro
     * @return  a word which expands to the value defined for 
     *          macro <tt>name</tt>
     */
    String useMacro(String name);

    /**
     * Quotes a word which may contain special characters.
     * Returns a string which within the script can function as a literal
     * meaning the string <tt>text</tt>.
     *
     * @param  text  text to quote
     * @return  a word which is the quoted version of <tt>text</tt>
     */
    String quoteWord(String text);

    /**
     * Returns a version of a string which can be used as a macro name
     * (does not contain any prohibited or special characters).
     * This will typically be done by turning any character which might
     * cause trouble into an underscore.  Nothing in this interface 
     * definition prevents the possibility that two names might map to 
     * the same namified string, but as long as users and implementors 
     * don't act deliberately stupid this is unlikely to happen.
     *
     * @param  name  any string representing a macro name
     * @return   a string related to <tt>name</tt> which can be used in
     *           this syntax as a macro name
     */
    String namify(String name);
}
