package org.eso.cpl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.stream.StreamSource;
import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameLevel;
import org.eso.cpl.FrameList;
import org.eso.cpl.FrameType;
import org.eso.cpl.Parameter;
import org.eso.cpl.ParameterValueException;
import org.eso.cpl.jni.LTDLException;
import org.eso.cpl.jni.LibraryLoader;
import org.eso.cpl.jni.PluginLibrary;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Class which deals with reading execution state from an XML document.
 * The expected format matches that written by {@link StateSaver}.
 * An instance of this class is initialized from an input stream
 * (which provides XML input) and that instance can then be interrogated
 * to find out what the XML in the input said about recipe state.
 *
 * @author   Mark Taylor
 * @see   StateSaver
 * @see   "cpl-state.dtd"
 */
class StateLoader {

    private static Map groupMap_;
    private static Map levelMap_;
    private static Map typeMap_;
    private static Map namerMap_;
    private LibraryLoader libLoader_;
    private Element recipeEl_;
    private Element libraryEl_;
    private Element parametersEl_;
    private Element inFramesEl_;
    private Element productsEl_;

    /** Initialize maps for translation of strings to objects. */
    static {
        groupMap_ = new HashMap();
        for (Iterator it = FrameGroup.ALL_GROUPS.iterator(); it.hasNext();) {
            FrameGroup group = (FrameGroup) it.next();
            groupMap_.put(group.getName(), group);
        }

        levelMap_ = new HashMap();
        for (Iterator it = FrameLevel.ALL_LEVELS.iterator(); it.hasNext();) {
            FrameLevel level = (FrameLevel) it.next();
            levelMap_.put(level.getName(), level);
        }

        typeMap_ = new HashMap();
        for (Iterator it = FrameType.ALL_TYPES.iterator(); it.hasNext();) {
            FrameType type = (FrameType) it.next();
            typeMap_.put(type.getName(), type);
        }

        namerMap_ = new HashMap();
        for (Iterator it = NamingScheme.STANDARD_SCHEMES.iterator(); 
             it.hasNext();) { 
            NamingScheme namer = (NamingScheme) it.next();
            namerMap_.put(namer.getName(), namer);
        }
    }

    /**
     * Initializes a new StateLoader object from an input stream.
     *
     * @param  istrm  a stream containing XML text serializing state
     * @param  libLoader  a LibraryLoader object which can be used to
     *         locate PluginLibraries which may contain recipes referenced
     *         in the serialized state
     */
    public StateLoader(InputStream istrm, LibraryLoader libLoader)
            throws IOException {
        libLoader_ = libLoader;

        // Read the input stream into a DOM.
        Source xsrc = new StreamSource(istrm);
        DOMResult xres = new DOMResult();
        try {
            Transformer trans = TransformerFactory.newInstance()
                                                  .newTransformer();
            trans.transform(xsrc, xres);
            Document doc = (Document) xres.getNode();
            Element docel = doc.getDocumentElement();

            // See if there are any <recipe> elements.  If so, use the
            // first one.  More complicated behaviours (searching for one 
            // with the right recipe name or something) are possible but
            // not currently implemented.
            NodeList requests = doc.getElementsByTagName("request");
            if (requests.getLength() >= 1) {
                readRequestElement((Element) requests.item(0));
            }
            else {
                throw new IOException("No <request> element found in stream");
            }
        }
        catch (TransformerException e) {
            throw (IOException) new IOException("XML Error: " + e.getMessage())
                 .initCause( e );
        }
    }

    /**
     * Examines a &lt;recipe&gt; element and configures the state of
     * this object according to its contents.
     *
     * @param  requestEl  &lt;request&gt; element
     */
    private void readRequestElement(Element requestEl) {
        checkElementTag(requestEl, "request");
        for (Node node = requestEl.getFirstChild(); node != null;
             node = node.getNextSibling()) {
            if (node instanceof Element) {
                Element childEl = (Element) node;
                String name = childEl.getTagName();
                if (name.equals("recipe")) {
                    recipeEl_ = childEl;
                }
                else if (name.equals("frames")) {
                    inFramesEl_ = childEl;
                }
                else if (name.equals("products")) {
                    productsEl_ = childEl;
                }
            }
        }
        if (recipeEl_ != null) {
            for (Node node = recipeEl_.getFirstChild(); node != null;
                 node = node.getNextSibling()) {
                if (node instanceof Element) {
                    Element childEl = (Element) node;
                    String name = childEl.getTagName(); 
                    if (name.equals("library")) {
                        libraryEl_ = childEl;
                    }
                    else if (name.equals("parameters")) {
                        parametersEl_ = childEl;
                    }
                }
            }
        }
    }

    /**
     * Tries to locate and return the recipe which is referenced in the 
     * serialized state.  This may fail for a number of reasons, for
     * instance the library is not found or the recipe of the given
     * name in the library has the wrong version number.
     *
     * @return  the recipe corresponding to the state loaded by this object
     * @throws  CPLException  if it could not be located
     */
    public Recipe getRecipe() throws CPLException {
        if (recipeEl_ == null) {
            return null;
        }
        if (libraryEl_ == null) {
            return null;
        }
        String recName = getAttributeValue(recipeEl_, "name");
        String recVers = getAttributeValue(recipeEl_, "version");
        String libName = getAttributeValue(libraryEl_, "name");
        String libLoc = getAttributeValue(libraryEl_, "location");
        String libClass = getAttributeValue(libraryEl_, "class");
        if (libClass.equals(PluginLibrary.class.getName())) {
            if (libLoader_ == null) {
                throw new CPLException("No LibraryLoader specified");
            }

            // Try to get the library.
            RecipeLibrary pluglib = null;
            if (pluglib == null && libLoc != null) {
                try {
                    pluglib = libLoader_.getLibrary(libLoc);
                }
                catch (LTDLException e) {
                }
            }
            if (pluglib == null) {
                try { 
                    pluglib = libLoader_.getLibrary(libName);
                }
                catch (LTDLException e) {
                    throw new CPLException(e.getMessage(), e);
                }
            }
            if (pluglib == null) {
                throw new CPLException("PluginLibrary " + pluglib + 
                                       " not found");
            }

            // We've got a library, interrogate it for the required recipe.
            Recipe[] recipes = pluglib.getRecipes();
            for (int i = 0; i < recipes.length; i++) {
                Recipe recipe = recipes[i];
                if (recipe.getName().equals(recName)) {
                    if (Long.toString(recipe.getVersion())
                            .equals(recVers)) {
                        return recipe;
                    }
                    else {
                        throw new CPLException("Recipe version mismatch " +
                                               "for " + recName + ": " +
                                               recipe.getVersion() + 
                                               " != " + recVers);
                    }
                }
            }
            throw new CPLException("No recipe " + recName + " found in " +
                                   pluglib);
        }
        else {
            throw new CPLException("Unknown recipe library class " + libClass);
        }
    }

    /**
     * Sets the values of a given list of parameters according to the
     * state of this object.
     *
     * @param  params  the array of parameters whose values are to be altered
     */
    public void setParameterValues(Parameter[] params)
            throws ParameterValueException {
        if (parametersEl_ != null) {
            setParameterValues(params, parametersEl_);
        }
    }

    /**
     * Returns the list of input frames stored by this object.
     *
     * @return   input frames
     */
    public FrameList getInputFrames() {
        return inFramesEl_ == null ? null
                                   : readFrameList(inFramesEl_);
    }

    /**
     * Returns the product root location stored by this object.
     *
     * @return   product root string
     */
    public String getProductRoot() {
        return productsEl_ == null 
             ? null 
             : getAttributeValue(productsEl_, "directory");
    }

    /**
     * Returns the naming scheme stored by this object.
     *
     * @return  naming scheme object
     */
    public NamingScheme getNamingScheme() {
        String schemeName = productsEl_ == null 
                               ? null
                               : getAttributeValue(productsEl_, "naming");
        return (NamingScheme) namerMap_.get(schemeName);
    }

    /**
     * Sets the value of a Parameter from a DOM element.
     *
     * @param  paramMap  a (name, Parameter) map of Parameters whose
     *         values may be set from those stored in <tt>paramEl</tt>
     * @param  paramEl  a &lt;parameter&gt; element such as that stored by
     *         {@link #storeParameter} containing parameter values
     * @throws  ParameterValueException  if an attempt is made to
     *          set a parameter to an illegal value
     */
    private void setParameterValue(Map paramMap, Element paramEl)
            throws ParameterValueException {
        checkElementTag(paramEl, "parameter");
        String name = getAttributeValue(paramEl, "name");
        if (paramMap.containsKey(name)) {
            Parameter param = (Parameter) paramMap.get(name);
            String value = getAttributeValue(paramEl, "value");
            param.setValueFromString(value);
        }
        else {
            throw new ParameterValueException("Ignoring parameter \"" + name +
                                              "\" not in recipe");
        }
    }

    /**
     * Sets the values of an array of parameters according to
     * the &lt;parameter&gt; elements in a &lt;parameters&gt; element.
     * Value settings are attempted for elements with names which match
     * the names of parameters in the supplied list.
     *
     * @param  params  an array of parameters
     * @param  paramsEl  a &lt;parameters&gt; element
     */
    private void setParameterValues(Parameter[] params, Element paramsEl)
            throws ParameterValueException {
        checkElementTag(paramsEl, "parameters");
        Map paramMap = new HashMap();
        for (int i = 0; i < params.length; i++) {
            Parameter param = params[i];
            paramMap.put(param.getName(), param);
        }
        for (Node child = paramsEl.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            if (child instanceof Element) {
                Element childEl = (Element) child;
                if (childEl.getTagName().equals("parameter")) {
                    setParameterValue(paramMap, childEl);
                }
            }
        }
    }

    /**
     * Constructs a new Frame object from a suitable DOM &lt;frame&gt; element.
     *
     * @param  frameEl  a &lt;frame&gt; element
     * @return  a new Frame as per <tt>frameEl</tt>
     */
    private Frame readFrame(Element frameEl) {
        checkElementTag(frameEl, "frame");
        Frame frame = new Frame();
        frame.setFile(new File(getAttributeValue(frameEl, "filename")));
        frame.setTag(getAttributeValue(frameEl, "tag"));
        frame.setGroup(getFrameGroup(getAttributeValue(frameEl, "group")));
        frame.setLevel(getFrameLevel(getAttributeValue(frameEl, "level")));
        frame.setType(getFrameType(getAttributeValue(frameEl, "type")));
        return frame;
    }

    /**
     * Constructs a new FrameList object from a suitable DOM &lt;frames&gt;
     * element.
     *
     * @param  framesEl a &lt;frames&gt; element
     * @return   a new FrameList populated from <tt>frameEl</tt>
     */
    private FrameList readFrameList(Element framesEl) {
        checkElementTag(framesEl, "frames");
        FrameList framelist = new FrameList();
        for (Node child = framesEl.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            if (child instanceof Element) {
                Element childEl = (Element) child;
                if (childEl.getTagName().equals("frame")) {
                    framelist.add(readFrame(childEl));
                }
            }
        }
        return framelist;
    }

    /**
     * Returns the value of an element's attribute.
     * This does two jobs additional to <tt>el.getAttribute(attname)</tt>:
     * <ol>
     * <li>If <tt>el</tt> has no attribute <tt>attname</tt> it returns
     *     <tt>null</tt> (not the empty string)
     * <li>It decodes any escaped characters in the attribute value
     * </ol>
     *
     * @param  el  the element to be interrogated
     * @param  attname  the attribute whose value is sought
     * @return  the value of <tt>el</tt>'s <tt>attname</tt> attribute
     */
    private static String getAttributeValue(Element el, String attname) {
        return el.hasAttribute(attname)
                   ? unescapeAttributeValue(el.getAttribute(attname))
                   : null;
    }

    /**
     * Unescapes a string from an XML attribute value so that any XML
     * escaped characters (character entity references) are turned back into
     * normal characters.
     *
     * @param  text  the escaped string
     * @return  the unescaped version
     */
    private static String unescapeAttributeValue(String text) {
        return text.replaceAll("&lt;", "<")
                   .replaceAll("&gt;", ">")
                   .replaceAll("&apos;", "'")
                   .replaceAll("&quot;", "\"")
                   .replaceAll("&amp;", "&");
    }

    /**
     * Ensures that an element has the expected tag name.  If so, nothing
     * happens.  If not, an exception is thrown.
     *
     * @param  el  element
     * @param  tagname  the tagname that <tt>el</tt> has to have
     * @throws   IllegalArgumentException  if <tt>el</tt> isn't a
     *           &lt;tagname&gt;
     */
    private static void checkElementTag(Element el, String tagname) {
        if (!el.getTagName().equals(tagname)) {
            throw new IllegalArgumentException("Wrong element type: <" +
                                               el.getTagName() + "> not <" +
                                               tagname + ">");
        }
    }

    /**
     * Translates a string into a FrameGroup object.
     *
     * @param  groupName  name of a FrameGroup static final member
     * @return  a FrameGroup object corresponding to <tt>groupName</tt>,
     *          or <tt>null</tt>
     */
    private static FrameGroup getFrameGroup(String groupName) {
        return (FrameGroup) groupMap_.get(groupName);
    }

    /**
     * Translates a string into a FrameLevel object.
     *
     * @param  levelName  name of a FrameLevel static final member
     * @return  a FrameLevel object corresponding to <tt>levelName</tt>,
     *          or <tt>null</tt>
     */
    private static FrameLevel getFrameLevel(String levelName) {
        return (FrameLevel) levelMap_.get(levelName);
    }

    /**
     * Translates a string into a FrameType object.
     *
     * @param  typeName  name of a FrameType static final member
     * @return  a FrameType object corresponding to <tt>typeName</tt>,
     *          or <tt>null</tt>
     */
    private static FrameType getFrameType(String typeName) {
        return (FrameType) typeMap_.get(typeName);
    }
}
