package org.eso.cpl.gui;

import javax.swing.DefaultCellEditor;
import javax.swing.JTextField;
import javax.swing.DefaultCellEditor;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

/**
 * This class describes columns in a table.  Such columns
 * have names and supply their own data cells as well as information about
 * how to render and edit their cells.
 *
 * @see  DataColumnTableModel
 * @author   Mark Taylor
 */
public abstract class DataColumn {

    private String name_;
    private boolean isEditable_;
    private TableCellEditor editor_;
    private TableCellRenderer renderer_;
    
    // default preferred width - based on TableColumn width.
    private int preferredWidth_ = 75;

    /**
     * Constructs a new DataColumn with a given name and content class
     * which may or may not be editable.  The supplied <tt>clazz</tt>
     * is used to determine the cell renderer and editor which will be
     * provided by default.
     *
     * @param  name  the name of the column
     * @param  isEditable  whether this column can be edited
     * @param  clazz  the class expected to be held in this column
     * @param  prefWidth  the preferred display width for the column 
     */
    public DataColumn(String name, boolean isEditable, Class clazz, int prefWidth) {
        this(name, isEditable, clazz);
        preferredWidth_ = prefWidth;
    }

    /**
     * Constructs a new DataColumn with a given name and content class
     * which may or may not be editable.  The supplied <tt>clazz</tt>
     * is used to determine the cell renderer and editor which will be
     * provided by default.
     * 
     * @param name  the name of the column
     * @param isEditable  whether this column can be edited
     * @param clazz  the class expected to be held in this column
     */
    public DataColumn(String name, boolean isEditable, Class clazz) {
        name_ = name;
        isEditable_ = isEditable;
        clazz = (clazz == null) ? Object.class : clazz;
        renderer_ = Renderers.getDefaultRenderer(clazz);
        editor_ = Renderers.getDefaultEditor(clazz); 
    }

    /**
     * Constructs a new DataColumn which may or may not be editable.
     *
     * @param  name  the name of the column
     * @param  isEditable  whether this column can be edited
     */
    public DataColumn(String name, boolean isEditable) {
        this(name, isEditable, null);
    }

    /**
     * Constructs a new DataColumn which may or may not be editable.
     *
     * @param  name  the name of the column
     * @param  isEditable  whether this column can be edited
     * @param  prefWidth  the preferred display width for the column
     */
    public DataColumn(String name, boolean isEditable, int prefWidth) {
        this(name, isEditable, null);
        preferredWidth_ = prefWidth;
    }

    /**
     * Returns the entry at a given row in this column.
     *
     * @param  irow  the row for which this column is being queried
     */
    abstract public Object getValue(int irow);

    /**
     * Indicates whether the entry at a given row in this column can
     * be edited (whether {@link #setValue} may be called on <tt>irow</tt>).
     * This implementation returns the value of the <tt>isEditable</tt>
     * flag set up at construction.
     *
     * @param  irow  the row for which this column is being queried
     * @return  whether the entry at <tt>irow</tt> may be written to
     */
    public boolean isEditable(int irow) {
        return isEditable_;
    }

    /**
     * Sets the value of the entry in this column at a given row.
     * This should only be called if {@link #isEditable}(irow) returns true.
     * This implementation throws an <tt>UnsupportedOperationException</tt>;
     * subclasses which may return <tt>true</tt> from <tt>isEditable</tt>
     * should override it.
     *
     * @param   irow  the row whose value in this column is to be set
     * @param   value  the new value of the entry at <tt>irow</tt>
     */
    public void setValue(int irow, Object value) {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns the name of this column.
     *
     * @return  name
     */
    public String getName() {
        return name_;
    }

    /**
     * Returns a cell renderer suitable for rendering the cell at a given
     * row in this column.  This implementation returns a default renderer
     * (same one for all rows).
     *
     * @param  irow  the row in this column to be rendered
     * @return   a renderer for the cell at row <tt>irow</tt>
     */
    public TableCellRenderer getCellRenderer(int irow) {
        return renderer_;
    }

    /**
     * Returns tooltip text to be displayed for the renderer in a given row.
     * By default, this returns <tt>null</tt> so no tool tip is displayed.
     *
     * @param  irow  the row 
     * @return  the tooltip text for row <tt>irow</tt>
     */
    public String getRendererTip(int irow) {
        return null;
    }

    /**
     * Returns a cell editor suitable for editing the cell at a given row
     * in this column.  This implementation returns a default editor
     * (same one for all rows).  This should only be called on rows for
     * which {@link #isEditable} returns true.
     *
     * @param  irow  the row in this column to be edited
     * @return  an editor for the cell at row <tt>irow</tt>
     */
    public TableCellEditor getCellEditor(int irow) {
        return editor_;
    }
    
    /**
     * Returns the preferredWidth.
     * @return int
     */
    public int getPreferredWidth() {
        return preferredWidth_;
    }

    /**
     * Sets the preferredWidth.
     * @param preferredWidth The preferredWidth to set
     */
    public void setPreferredWidth(int preferredWidth) {
        this.preferredWidth_ = preferredWidth;
    }

}
