package org.eso.cpl.gui;

import java.awt.Color;
import java.awt.Component;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;

import org.eso.dfs.gui.DnDJTable;

/**
 * A JTable designed for use with a {@link DataColumnTableModel}.
 * Some JTable methods are overridden to ensure that rendering is done
 * as specified by {@link DataColumn} objects.
 * The model isn't required to be a DataColumnTableModel, but there's not
 * much point using this class if it's not.
 * 
 * @author   Mark Taylor
 */
public class DataColumnTable extends DnDJTable {

    private Color editableColor_;
    private Color fixedColor_;

    /**
     * Constructs a new DataColumnTable with a default table model.
     */
    public DataColumnTable() {
        super();
        fixedColor_ = getBackground();
        editableColor_ = mixColors(Color.yellow, fixedColor_);
    }

    /**
     * Constructs a new DataColumnTable with a given DataColumnTableModel.
     *
     * @param  model  the model
     */
    public DataColumnTable(DataColumnTableModel model) {
        super(model);
    }
    
    

    /**
     * Ensure that rendering is done by the JTable using a renderer
     * suitable for the cell in question.
     */
    public TableCellRenderer getCellRenderer(int irow, int icol) {
        DataColumn dcol = getDataColumn(icol);
        return dcol == null ? super.getCellRenderer(irow, icol)
                            : dcol.getCellRenderer(irow);
    }

    /**
     * Ensure that editing is done by the JTable using an editor
     * suitable for the cell in question.
     */
    public TableCellEditor getCellEditor(int irow, int icol) {
        DataColumn dcol = getDataColumn(icol);
        return dcol == null ? super.getCellEditor(irow, icol)
                            : dcol.getCellEditor(irow);
    }

    /**
     * Ensures that cells will be displayed with some visual indication
     * that they are or are not editable.
     */
    public Component prepareRenderer(TableCellRenderer renderer, 
                                     int irow, int icol) {
        Component cell = super.prepareRenderer(renderer, irow, icol);;
        Object val = getValueAt(irow, icol);
        
        // HACK for Double type values.
        if (val instanceof Double && cell instanceof DefaultTableCellRenderer) {
            // the NumberFormat class in javax.swing.JTable$DoubleRenderer does not
            // permit precission greater than 3 dicimal places, and is therefor a problem
            // on occation.
            // This HACK replaces the cell value good olde toString().
            ((DefaultTableCellRenderer)cell).setText(val.toString());
        } else {
            cell = super.prepareRenderer(renderer, irow, icol);   
        }
        // Now modify the properties as we want them to appear.
        if (cell instanceof JComponent) {
            DataColumn dcol = getDataColumn(icol);
            String tip = dcol == null ? null
                                      : dcol.getRendererTip(irow);
            ((JComponent) cell).setToolTipText(tip);
        }
        if (cell instanceof JLabel ||
            cell instanceof JCheckBox) {
            cell.setBackground(isCellEditable(irow, icol) ? editableColor_ 
                                                          : fixedColor_);
        }
        return cell;
    }

    public Component prepareEditor(TableCellEditor editor,
                                   int irow, int icol ) {
        Component cell = super.prepareEditor(editor, irow, icol);
        if (!(cell instanceof JButton)) {
            // do not change the background colour for buttons.
            cell.setBackground(isCellEditable(irow, icol) ? editableColor_
                                                      : fixedColor_);
        }
        return cell;
    }

    /**
     * Returns the DataColumn object corresponding to a given column of
     * this JTable.  May be null in the event that this table's model is
     * not a DataColumnTableModel.
     *
     * @param  icol  the column in the table (not necessarily the 
     *               TableColumnModel)
     * @return  the corresponding DataColumn
     */
    private DataColumn getDataColumn(int icol) {
        TableModel model = getModel();
        if (model instanceof DataColumnTableModel) {
            int jcol = getColumnModel().getColumn(icol).getModelIndex();
            return ((DataColumnTableModel) model).getDataColumn(jcol);
        }
        else {
            return null;
        }
    }

    /**
     * Returns a mixture of two colours.
     *
     * @param  c1  colour 1
     * @param  c2  colour 2
     * @return  mixture
     */
    private static Color mixColors(Color c1, Color c2) {
        return new Color((c1.getRed() + c2.getRed()) / 2,
                         (c1.getGreen() + c2.getGreen()) / 2,
                         (c1.getBlue() + c2.getBlue()) / 2);
    }
    /**
     * @see javax.swing.JTable#setModel(javax.swing.table.TableModel)
     */
    public void setModel(TableModel dataModel) {
        super.setModel(dataModel);
        TableColumnModel cm = getColumnModel();
        for (int i = 0; i < cm.getColumnCount(); i++) {
            TableColumn tc = cm.getColumn(i);
            DataColumn dc = getDataColumn(i);
            if (dc != null) {
                tc.setPreferredWidth(dc.getPreferredWidth());
            }
        }
    }
}
