package org.eso.cpl.gui;

import java.util.ArrayList;
import java.util.List;
import javax.swing.table.AbstractTableModel;

/**
 * Makes a {@link javax.swing.table.TableModel} 
 * out of a list of {@link DataColumn} objects.
 * The behaviour of the model is supplied entirely on a per-column basis
 * by the list of supplied DataColumns.
 * <p>
 * Concrete subclasses must implement {@link #getRowCount} to indicate
 * the number of rows in the table.
 * <p>
 * Note this model should be displayed in a {@link DataColumnTable}
 * (rather than a plain <tt>JTable</tt>) for the rendering instructions 
 * supplied by the columns to be heeded.
 *
 * @author    Mark Taylor
 */
public abstract class DataColumnTableModel extends AbstractTableModel {

    private List columns_ = new ArrayList();

    /**
     * Constructs a new empty DataColumnTableModel.
     */
    public DataColumnTableModel() {
    }

    /**
     * Sets the list of columns which provide the data for this table model.
     *
     * @param  columns  a list of {@link DataColumn} objects
     */
    public void setDataColumns(List columns) {
        columns_ = columns;
        fireTableStructureChanged();
    }

    /**
     * Returns the list of columns which provide the data for this table model.
     * Modifications to the list will affect the model; if you make any
     * you should take care to fire appropriate events.
     *
     * @return  the list of {@link DataColumn} objects
     */
    public List getDataColumns() {
        return columns_;
    }

    public int getColumnCount() {
        return columns_.size();
    }

    abstract public int getRowCount();

    public Object getValueAt(int irow, int icol) {
        return getDataColumn(icol).getValue(irow);
    }

    public void setValueAt(Object value, int irow, int icol) {
        getDataColumn(icol).setValue(irow, value);
    }

    public String getColumnName(int icol) {
        return getDataColumn(icol).getName();
    }

    public boolean isCellEditable( int irow, int icol ) {
        return getDataColumn(icol).isEditable(irow);
    }

    /**
     * Returns the DataColumn associated with a given column of the table.
     *
     * @param  icol  column index
     * @return  the DataColumn object which supplies the data for 
     *          column <tt>icol</tt>
     */
    public DataColumn getDataColumn(int icol) {
        return (DataColumn) columns_.get(icol);
    }
}
