package org.eso.cpl.gui;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameLevel;
import org.eso.cpl.FrameList;
import org.eso.cpl.FrameType;
import org.eso.dfs.services.ServiceException;
import org.eso.dfs.services.gui.FileImageDisplayService;

/**
 * A TableModel which represents all known information about a set of
 * {@link org.eso.cpl.Frame} objects.  You can modify the FrameList
 * on which the model is based during display, just make sure to 
 * fire appropriate events.
 *
 * @author   Mark Taylor
 */
public class FrameTableModel extends DataColumnTableModel {

    private final FrameList frames_;
    private Map displayButtons_ = new HashMap();
    private Map locateButtons_ = new HashMap();

    /**
     * Constructs a new FrameTableModel from a set of frames.
     *
     * @param  frames  the framelist on which the table will be based
     * @param  isEditable  whether it will be possible to edit the 
     *         contents of the table
     */
    public FrameTableModel(FrameList frames, boolean isEditable) {
        frames_ = frames;

        // Set up custom editors for some of the columns.
        final TableCellEditor groupEditor = 
            Renderers.getOptionEditor(FrameGroup.ALL_GROUPS);
        final TableCellEditor levelEditor =
            Renderers.getOptionEditor(FrameLevel.ALL_LEVELS);
        final TableCellEditor typeEditor =
            Renderers.getOptionEditor(FrameType.ALL_TYPES);
        final TableCellEditor filenameEditor =
            Renderers.getTrailingStringEditor();
        final TableCellRenderer filenameRenderer =
            Renderers.getTrailingStringRenderer();
        final TableCellRenderer buttRenderer =
            Renderers.getComponentRenderer();

        // Set up DataColumn objects for each of the various fields for
        // Frames.
        // filenames are not editable.
        DataColumn filenameColumn = new DataColumn("Filename", false, 200) {
            public Object getValue(int irow) {
                // return just the file name, not the path.
                return getFrame(irow).getFile().getName();
            }
            public void setValue(int irow, Object value) {
                getFrame(irow).setFile(new File((String) value));
            }
            public TableCellRenderer getCellRenderer(int irow) {
                return filenameRenderer;
            }
            public TableCellEditor getCellEditor(int irow) {
                return filenameEditor;
            }
            public String getRendererTip(int irow) {
                return getFrame(irow).getCanonicalPath();
            }
        };
        DataColumn groupColumn = new DataColumn("Group", isEditable, 15) {
            public Object getValue(int irow) {
                return getFrame(irow).getGroup();
            }
            public void setValue(int irow, Object value) {
                getFrame(irow).setGroup((FrameGroup) value);
            }
            public TableCellEditor getCellEditor(int irow) {
                return groupEditor;
            }
            public String getRendererTip(int irow) {
                return getFrame(irow).getCanonicalPath();
            }
        };
        DataColumn levelColumn = new DataColumn("Level", isEditable) {
            public Object getValue(int irow) {
                return getFrame(irow).getLevel();
            }
            public void setValue(int irow, Object value) {
                getFrame(irow).setLevel((FrameLevel) value);
            }
            public TableCellEditor getCellEditor(int irow) {
                return levelEditor;
            }
            public String getRendererTip(int irow) {
                return getFrame(irow).getCanonicalPath();
            }
        };
        DataColumn tagColumn = new DataColumn("Classification", isEditable,
                                              String.class) {
            public Object getValue(int irow) {
                return getFrame(irow).getTag();
            }
            public void setValue(int irow, Object value) {
                getFrame(irow).setTag((String) value);
            }
            public String getRendererTip(int irow) {
                return getFrame(irow).getCanonicalPath();
            }
        };
        DataColumn typeColumn = new DataColumn("Type", isEditable) {
            public Object getValue(int irow) {
                return getFrame(irow).getType();
            }
            public void setValue(int irow, Object value) {
                getFrame(irow).setType((FrameType) value);
            }
            public TableCellEditor getCellEditor(int irow) {
                return typeEditor;
            }
            public String getRendererTip(int irow) {
                return getFrame(irow).getCanonicalPath();
            }
        };
        DataColumn optionsColumn1 = new DataColumn("", true) {
            TableCellEditor editor = Renderers.getComponentEditor();
            public Object getValue(int irow) {
                return getDisplayButton(getFrame(irow), false);
            }
            public TableCellRenderer getCellRenderer(int irow) {
                return buttRenderer;
            }
            public TableCellEditor getCellEditor(int irow) {
                return editor;
            }
            public String getRendererTip(int irow) {
                return "Locate file in Gasgano";
            }
        };
        DataColumn optionsColumn2 = new DataColumn("", true) {
            TableCellEditor editor = Renderers.getComponentEditor();
            public Object getValue(int irow) {
                return getDisplayButton(getFrame(irow), true);
            }
            public TableCellRenderer getCellRenderer(int irow) {
                return buttRenderer;
            }
            public TableCellEditor getCellEditor(int irow) {
                return editor;
            }
            public String getRendererTip(int irow) {
                return "Send file to external viewer";
            }
        };
 
        // Set up a list of DataColumns in the order in which they should
        // appear in the table.
        List columns = new ArrayList();
        columns.add(filenameColumn);
        // dont show group column
        // columns.add(groupColumn);
        // dont show level and type.
        // columns.add(levelColumn);
        //columns.add(typeColumn);
        columns.add(tagColumn);
        columns.add(optionsColumn1);
        columns.add(optionsColumn2);

        // Use the list as the data for this table model.
        setDataColumns(columns);
    }

    /**
     * @see javax.swing.table.TableModel#getRowCount()
     */
    public int getRowCount() {
        return frames_.size();
    }

    /**
     * Returns the frame at a given row in this TableModel.
     *
     * @param  irow  table row
     * @return  the frame at <tt>irow</tt>
     */
    public Frame getFrame(int irow) {
        return frames_.getFrame(irow);
    }

    /**
     * Returns the FrameList which this model represents.
     * @return  the frames
     */
    public FrameList getFrames() {
        return frames_;
    }

    /**
     * Get a cell button for the given frame (representing a row in the table).
     * The button will have action and label based on the showImage value.
     * @param frm
     * @param showImage if true, button will send the frame to the image display service, 
     * requesting that the image be displayed (Gasgano External Viewer), if false,
     * only the frame primary header or summary need be shown (in the Gasgano Main Window).
     * @return Component the button component to be used in the JTable cell
     */
    private Component createOptionButton(final Frame frm, final boolean showImage) {   
        final String tag = showImage ? "Display" : "Locate";
        Action action = new AbstractAction(tag) {
            public void actionPerformed(ActionEvent evt) {
                //java.awt.Toolkit.getDefaultToolkit().beep();
                System.out.println(tag+" frame " + frm.getCanonicalPath());
                try {
                    FileImageDisplayService.getInstance()
                        .displayFile(new String[] {frm.getCanonicalPath()}, showImage);
                } catch (ServiceException e) {
                    e.printStackTrace();
                }
            }
        };
        return new JButton(action);
    }
    
    /**
     * Returns a button giving additional options for a given frame. The button will either be
     * for displaying the frame image or simply locating it in the Gasgano Main Window, depending
     * on the showImage parameter.
     */
    private Component getDisplayButton(Frame frm, boolean showImage) {
        Map buttonMap = showImage ? displayButtons_ : locateButtons_;
        if (!buttonMap.containsKey(frm)) {
            if (frm.getType() == FrameType.IMAGE || frm.getType() == FrameType.TABLE) {
                // can display this type of frames - create a display button for the cell.
                Component butt = createOptionButton(frm, showImage);
                buttonMap.put(frm, butt);
            } else {
                // can not display this frame type - create an empty panel for the cell.
                final JPanel panel = new JPanel();
                buttonMap.put(frm, panel);
            }
        }
        return (Component)buttonMap.get(frm);
    }
}
