package org.eso.cpl.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import org.eso.cpl.Recipe;
import org.eso.cpl.RecipeLibrary;

/**
 * Displays a recipe selector and information about whichever recipe
 * is currently selected. 
 * Buttons permit the currently selected recipe to be chosen.
 * Concrete subclasses should implement the {@link #approveSelection} and
 * {@link #cancelSelection} methods to do something useful when the
 * user indicates a decision.
 * <p>
 * Recipe display is currently done using a <tt>JTree</tt> 
 * ({@link RecipeTree}), but this could be replaced by a different GUI
 * if required, for instance a flatter one.
 *
 * @author   Mark Taylor
 */
public abstract class RecipeSelector extends JPanel {

    private final RecipeTree tree_;
    private final RecipeDisplay display_;
    private Recipe selectedRecipe_ = null;
    private boolean hasCancelSelection = true;

    /**
     * Constructs a new empty RecipeSelector, with a cancel selection button.
     */
    public RecipeSelector() {
        this(true);
    }
 
    /**
     * Constructs a new empty RecipeSelector, if cancelSelection is true,
     * with a cancel selection button, otherwise without. The cancelSelection
     * method still needs to be implemented.
     */   
    public RecipeSelector(boolean cancelSelection) {
        super(new BorderLayout());
        hasCancelSelection = cancelSelection;
        setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

        // Put a split pane in the middle, which will hold the main components.
        JSplitPane splitter = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        splitter.setResizeWeight(0.5);
        add(splitter, BorderLayout.CENTER);

        // Put the recipe tree in the top part of the split pane.
        tree_ = new RecipeTree();
        JScrollPane tscroller = new JScrollPane(tree_);
        tscroller.setPreferredSize(new Dimension(Short.MAX_VALUE, 200));
        splitter.setTopComponent(tscroller);

        // Put the recipe detail display in the bottom part.
        display_ = new RecipeDisplay();
        splitter.setBottomComponent(display_);
        setPreferredSize(new Dimension(500, 400));
        
        final Action okAction = new AbstractAction("OK") {
            public void actionPerformed(ActionEvent evt) {
                approveSelection();
            }
        };
        okAction.setEnabled(false);
        JButton cancelButton = null;
        if (hasCancelSelection) {       
            // Set up actions for cancel and accept.
            Action cancelAction = new AbstractAction("Cancel") {
                public void actionPerformed(ActionEvent evt) {
                    cancelSelection();
                }
            };
            cancelButton = new JButton(cancelAction);
        }

        // Put selection buttons at the bottom of the window.
        Box bottomBox = Box.createVerticalBox();
        bottomBox.add(Box.createVerticalStrut(10));
        add(bottomBox, BorderLayout.SOUTH);
        Box controlBox = Box.createHorizontalBox();
        bottomBox.add(controlBox);
        JButton okButton = new JButton(okAction);
        
        controlBox.add(Box.createHorizontalGlue());
        controlBox.add(okButton);
        bottomBox.add(Box.createVerticalStrut(5));
        if (hasCancelSelection) {
            controlBox.add(cancelButton);
        }
        controlBox.add(Box.createHorizontalStrut(5));

        // Set up a listener so the display responds to tree selections.
        tree_.addTreeSelectionListener(new TreeSelectionListener() {
            public void valueChanged(TreeSelectionEvent evt) {
                Object uo = tree_.getUserObject(evt.getPath()
                                                   .getLastPathComponent());
                selectedRecipe_ = (uo instanceof Recipe) ? (Recipe) uo : null;
                display_.setRecipe(selectedRecipe_);
                okAction.setEnabled(selectedRecipe_ != null);
            }
        } );
    }

    /**
     * Constructs a new RecipeSelector with an initial set of recipe libraries.
     *
     * @param  libs  an array of RecipeLibrary objects with which to
     *         populate the recipe selection tree
     */
    public RecipeSelector(RecipeLibrary[] libs) {
        this(libs, true);
    }
    
    public RecipeSelector(RecipeLibrary[] libs, boolean hasCancelSelection) {
        this(hasCancelSelection);
        if (libs != null) {
            for (int i = 0; i < libs.length; i++) {
                tree_.addLibrary(libs[i]);
            }
        }
    }

    /**
     * Returns the currently selected recipe.
     * If no recipe is selected, <tt>null</tt> is returned.
     *
     * @return  the currently selected recipe
     */
    public Recipe getSelectedRecipe() {
        return selectedRecipe_;
    }

    /**
     * This method should be implemented by subclasses providing an action
     * to perform when a recipe has been selected.  Calling the
     * {@link #getSelectedRecipe} method from this method will return
     * a non-<tt>null</tt> recipe.
     */
    protected abstract void approveSelection();

    /**
     * This method should be implemented by subclasses providing an action
     * to perform when the user has indicated that no recipe will be 
     * selected.
     */
    protected abstract void cancelSelection();

    /**
     * Returns the RecipeDisplay component used by this object.
     *
     * @return  the recipe display widget
     */
    public RecipeDisplay getDisplay() {
        return display_;
    }

    /**
     * Returns the RecipeTree component used by this object.
     * This may be manipulated directly to add new recipes if necessary.
     *
     * @return  the recipe tree widget
     */
    public RecipeTree getTree() {
        return tree_;
    }
}
