package org.eso.cpl.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.KeyStroke;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import org.eso.cpl.FrameList;
import org.eso.cpl.MarkedFrameList;
import org.eso.cpl.NamingScheme;
import org.eso.cpl.Request;

/**
 * Panel which displays the characteristics of a {@link org.eso.cpl.Request}
 * object.
 * <p>
 * Currently no mechanism (listener/event framework) exists by which 
 * a Request can notify a RequestPanel that its contents have changed,
 * so users of this object should not directly modify 
 * a Request's data while it is contained in a RequestPanel (since the
 * panel would not reflect such changes).  Note however that the 
 * displayed Request may change as a result of user interaction with
 * this panel.  Methods such as {@link #addInputFrames} which operate 
 * on this display are also OK of course.
 *
 * @author   Mark Taylor
 */
public class RequestPanel extends JPanel {

    private final LockableTable paramTable_ = new LockableTable();
    private final LockableTable inFrameTable_ = new LockableTable();
    private final JComponent headingBox_;
    private final JComponent controlBox_;
    private final NamingPanel namingPanel_;
    private Request request_;
    private boolean isLocked_;

    /**
     * Constructs a new request panel.
     */
    //  DS FEB 2007 - SPR DFS02892
    // Added overloaded constructor to pass the product default directory:
    public RequestPanel() { this(null); }
    public RequestPanel(String productDirPath) {
        super(new BorderLayout());
        JComponent mainBox = Box.createVerticalBox();
        add(mainBox, BorderLayout.CENTER);
        mainBox.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

        // Do some cosmetic configuration on the tables.
        GUIUtils.configureTable(paramTable_);
        GUIUtils.configureTable(inFrameTable_);
        GUIUtils.enableFileListDragDrop(inFrameTable_);

        // Ensure that the tables with buttons in them will have mouse
        // events forwarded appropriately.
        String deleteCurrent = "DELETE_CURRENT_FRAME";
        inFrameTable_.addMouseListener(new TableComponentMouseListener());
        inFrameTable_.getInputMap().put(
            KeyStroke.getKeyStroke(KeyEvent.VK_DELETE, 0),
            deleteCurrent);
        inFrameTable_.getActionMap().put(deleteCurrent, new AbstractAction() {
            public void actionPerformed(ActionEvent e) {
                if (request_ != null) {
                    removeSelectedInputFrame();
                }
            }
        });

        // Place parameter table.
        mainBox.add(
            GUIUtils.wrapInBorder(
                GUIUtils.wrapInScroller(paramTable_),
                "Parameters"));
        mainBox.add(Box.createVerticalStrut(5));

        // Place input frame table.
        JScrollPane scroller = GUIUtils.wrapInScroller(inFrameTable_);
        GUIUtils.enableFileListDragDrop(scroller);

        mainBox.add(GUIUtils.wrapInBorder(scroller, "Input Frames"));
        mainBox.add(Box.createVerticalStrut(5));

        // Place naming panel.
        
        // DS FEB 2007 - SPR DFS02892
        // // namingPanel_ = new NamingPanel();
        namingPanel_ = new NamingPanel(productDirPath);
        
        mainBox.add(namingPanel_);

        // Set up heading and control boxes.
        headingBox_ = new Box(BoxLayout.X_AXIS);
        add(headingBox_, BorderLayout.NORTH);
        controlBox_ = new Box(BoxLayout.X_AXIS);
        add(controlBox_, BorderLayout.SOUTH);
    }

    /**
     * Returns the Request object being displayed by this panel.
     * @return request
     */
    public Request getRequest() {
        stopEditing();
        return request_;
    }
    
    /**
     * Stop the editors (and save the contents) for all table
     * cell editors. This call is made before creating a Request
     * from the RequestPanel.
     */
    public void stopEditing() {
        stopEditing(paramTable_);
        stopEditing(inFrameTable_);
    }
    
    
    /**
     * Stop the editing on the specified table.
     * @param tbl
     */
    private static void stopEditing(JTable table) {
        int row = table.getEditingRow();
        int col = table.getEditingColumn();
        if (row != -1 && col != -1)
        table.getCellEditor(row,col).stopCellEditing(); 
    }

    /**
     * Sets the Request object to be displayed by this panel.
     * This will update the entire panel to display the current contents
     * of the given request.
     *
     * @param  request  the request to display
     */
    public void setRequest(Request request) {
        // stop and save any existing edit sessions
        // before setting the new request.
        stopEditing();
        request_ = request;

        ParameterTableModel paramModel =
            new ParameterTableModel(request_.getParameters());
        paramTable_.setModel(paramModel);

        MarkedFrameTableModel frameModel =
            new MarkedFrameTableModel(request_.getInputFrames());
        inFrameTable_.setModel(frameModel);

        if (request_.getNamingScheme() == null) {
            request_.setNamingScheme(NamingScheme.DEFAULT);
        }
        namingPanel_.setRequest(request_);
    }

    /**
     * Adds a given list of frames to the input frame list.
     * Does nothing if the table is locked.
     *
     * @param  frames  new frames to add
     */
    public void addInputFrames(FrameList addedFrames) {

        if (isLocked() || request_ == null) {
            return;
        }
        // Get this request's frame list.
        MarkedFrameList frames = request_.getInputFrames();

        // Append all the frames in the added list.
        frames.addAll(addedFrames);

        // Give the table a nudge.
        ((AbstractTableModel) inFrameTable_.getModel()).fireTableDataChanged();
    }

    /**
     * Remove the currently selected input frame, if the request is not
     * locked.
     */
    public void removeSelectedInputFrame() {
        removeInputFrame(inFrameTable_.getSelectedRow());
    }

    /**
     * Remove the frame at the given row number from the input frame list.
     * Selection is moved to closest row if out of bounds.
     * Does nothing with an empty input FrameList or one that is locked
     *
     * @param  frames  new frames to add
     */
    public void removeInputFrame(int row) {
        if (isLocked() || request_ == null
            || request_.getInputFrames().size() == 0) {
            return;
        }
        if (row < 0) {
            row = 0;
        }
        if (row >= request_.getInputFrames().size()) {
            row = request_.getInputFrames().size() - 1;
        }
        // remove the selected row.
        request_.getInputFrames().remove(row);
        // Give the table a nudge.
        ((AbstractTableModel) inFrameTable_.getModel()).fireTableDataChanged();
    }

    /**
     * Sets the locked state of this panel.  When locked, user actions 
     * cannot affect the contents of the panel.
     * 
     * @param  locked  true iff user actions to change the contents of
     *         this panel should be blocked
     */
    public void setLocked(boolean locked) {
        paramTable_.setLocked(locked);
        inFrameTable_.setLocked(locked);
        namingPanel_.setLocked(locked);
    }

    /**
     * Returns the component into which heading information should be inserted.
     *
     * @return  heading box
     */
    public JComponent getHeadingBox() {
        return headingBox_;
    }

    /**
     * Returns the component into which controls (like buttons) should
     * be inserted.
     *
     * @return  control box
     */
    public JComponent getControlBox() {
        return controlBox_;
    }

    /**
     * Returns the isLocked_.
     * @return boolean
     */
    public boolean isLocked() {
        return isLocked_;
    }

    /**
     * Subclass of DataColumnTable which allows additional per-table
     * control over whether any of the cells are editable.
     */
    private static class LockableTable extends DataColumnTable {
        boolean isLocked_ = false;
        public void setLocked(boolean locked) {
            boolean change = locked != isLocked_;
            isLocked_ = locked;
            TableModel model = getModel();
            if (change && model instanceof AbstractTableModel) {
                ((AbstractTableModel) model).fireTableDataChanged();
            }
        }
        public boolean isCellEditable(int irow, int icol) {
            return !isLocked_ && super.isCellEditable(irow, icol);
        }

        /**
         * Returns the isLocked_.
         * @return boolean
         */
        public boolean isLocked() {
            return isLocked_;
        }
    }
}
