package org.eso.cpl.gui;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eso.cpl.Request;

/**
 * @author Nick Kornweibel for ESO
 * 
 * This class is a pool or Request objects. If extends RequestQueue only for
 * the RequestQueue's enumeration function - it is otherwise not a queue, rather
 * a pool of Request objects, for which a subset are "selected".
 * 
 * The selected Requests may be returned as a RequestQueue.
 * 
 */
public class RequestPool extends RequestQueue {

    // List of selected Request objects
    private List selectedRequests_ = new ArrayList();
    
    /**
     * Constructor for RequestPool.
     */
    public RequestPool() {
        super();
    }

    /**
     * Remove the specified item from the queue.
     * @param i
     */
    public void removeRequest(int i) {
        Request r = (Request)items_.get(i);
        // make sure it is no longer in the selected list.
        setRequestSelected(i, false);
        items_.remove(i);
    }

    /**
     * Mark the requests at index i as either selected or
     * not selected.
     * @param i the index of the Request
     * @param select true if it is to be selected, else false.
     */
    public void setRequestSelected(int i, boolean select) {
        Request r = (Request)items_.get(i);
        selectedRequests_.remove(r);
        if (select) {
            // if selected, and was already selected, it will be removed
            // from the selected list and added again (to the end).
            selectedRequests_.add(r);
        }
    }
    
    /**
     * Clear and reset the request queue to the requests at the given indices.
     * @param indices
     */
    public void setRequestSelection(int[] indices) {
        clearRequestSelection();
        for (int x = 0; x < indices.length; x++) {
            int i = indices[x];
            Request r = (Request)items_.get(i);
            selectedRequests_.add(r);
        }
    }
    
    public void clearRequestSelection() {
        selectedRequests_.clear();
    }

    /**
     * Determine if request at index i is selected or not.
     * @param i
     * @return boolean
     */
    public boolean isRequestSelected(int i) {
        Request r = (Request)items_.get(i);
        return selectedRequests_.contains(r);
    }
    
    /**
     * Get the size of the request queue (not the size
     * of the pool).
     * @return int the number of selected requests.
     */
    public int getNumberSelectedRequests() {
        return selectedRequests_.size();
    }

    /**
     * Get the current selected requests as a queue.
     * The order of the queue is the order the Requests in 
     * the queue were selected.
     * @return RequestQueue
     */
    public RequestQueue getSelectedRequests() {
        RequestQueue rc = new RequestQueue();
        Iterator i = selectedRequests_.iterator();
        while (i.hasNext()) {
            rc.add((Request)i.next());
        }
        return rc;
    }
}
