package org.eso.cpl.jni;

import java.io.File;
import java.util.Iterator;
import org.eso.cpl.CPLException;
import org.eso.cpl.Frame;
import org.eso.cpl.FrameGroup;
import org.eso.cpl.FrameList;
import org.eso.cpl.Parameter;
import org.eso.cpl.ParameterConstraint;
import org.eso.cpl.ParameterImp;
import org.eso.cpl.ParameterType;
import org.eso.cpl.ParameterValueException;
import org.eso.cpl.Recipe;
import org.eso.cpl.RecipeLibrary;

/**
 * An implementation of the <tt>Recipe</tt> interface which works using
 * native code supplied by the CPL C library.
 * An instance of this class corresponds to a struct of type
 * <tt>cpl_recipe</tt> in the native code.
 *
 * @author   Mark Taylor
 * @see  "CPL library"
 */
public class JNIRecipe implements Recipe {

    private final byte[] nativeState_;
    private final PluginLibrary pluglib_;
    private final Parameter[] defaultParams_;
    private String author_;
    private String copyright_;
    private String description_;
    private String email_;
    private String name_;
    private String synopsis_;
    private long version_;
    private boolean isClosed_;

    /**
     * Private constructor; this class is only instantiated by native code.
     *
     * @param  nativeState  a buffer containing opaque bytes used by
     *         native code
     */
    private JNIRecipe(byte[] nativeState, PluginLibrary pluglib)
            throws CPLException {
        nativeState_ = nativeState;
        pluglib_ = pluglib;
        pluglib.register();

        // Set up a copy of the recipe parameters.
        synchronized (CPLControl.LOCK) {
            initialize();
            Parameter[] nativeParams = nativeGetParameterArray();
            int nparam = nativeParams.length;
            defaultParams_ = new Parameter[nparam];
            for (int i = 0; i < nparam; i++) {
                defaultParams_[i] = copyParameter(nativeParams[i]);
            }
            deinitialize();
        }
    }

    /**
     * Releases any resources associated with this recipe.
     * May be called multiple times harmlessly.
     */
    public void close() throws LTDLException {
        if (!isClosed_) {
            pluglib_.deregister();
            isClosed_ = true;
        }
    }

    /**
     * Calls close.
     */
    protected void finalize() throws Throwable {
        try {
            close();
        }
        finally {
            super.finalize();
        }
    }

    //
    // Implementation of the Recipe interface.
    //

    public Parameter[] getDefaultParameters() {
        int nparam = defaultParams_.length;
        Parameter[] params = new Parameter[nparam];
        for (int i = 0; i < nparam; i++) {
            params[i] = copyParameter(defaultParams_[i]);
        }
        return params;
    }

    public FrameList execute(Parameter[] params, FrameList inFrames,
                             File productRoot) 
            throws CPLException {

        // Validate the product root directory.
        if (productRoot != null) {
            if (!productRoot.exists() || !productRoot.isDirectory()) {
                throw new CPLException("No such directory " + productRoot);
            }
            if (!productRoot.canWrite()) {
                throw new CPLException("Directory " + productRoot +
                                       " is not writable");
            }
        }
        String execDir = productRoot == null ? null : productRoot.getPath();
        String currentDir = System.getProperty("user.dir");

        // Do the execution.
        FrameList outFrames;
        int execStatus;
        synchronized (CPLControl.LOCK) {
            initialize();
            nativeSetFrameArray(inFrames.toFrameArray());
            copyParameterValues(params, nativeGetParameterArray());
            execStatus = nativeExecute(currentDir, execDir);
            outFrames = new FrameList(nativeGetFrameArray());
            deinitialize();
        }
        if (execStatus != 0) {
            throw new CPLException("Execution failed with code " + execStatus);
        }
	
	// return to the caller only the frames having group = PRODUCT
	FrameList retFrames = new FrameList();
        for (Iterator it=outFrames.iterator(); it.hasNext();) {
	    Frame frm = (Frame) it.next();
	    if (frm.getGroup() == FrameGroup.PRODUCT) {
		retFrames.add(frm);
	    }
        }
	
        // Then make sure that they claim to be in the right directory.
        // the nativeExecute method works by chdir'ing to the product
        // root directory and then doing the execution (this is the only
        // way you can control what directory CPL products turn up in)
        // so if there was a non-default product root we'll need to
        // modify the directory that the products think they are in.
        if (productRoot != null) {
            for (Iterator it = outFrames.iterator(); it.hasNext(); ) {
                Frame frm = (Frame) it.next();
                frm.setFile(new File(productRoot, frm.getFile().getPath()));
            }
        }

        // Return the product frames.
        return retFrames;
    }

    public String getAuthor() {
        return author_;
    }

    public String getCopyright() {
        return copyright_;
    }

    public String getDescription() {
        return description_;
    }

    public String getEmail() {
        return email_;
    }

    public String getName() {
        return name_;
    }

    public String getSynopsis() {
        return synopsis_;
    }

    public long getVersion() {
        return version_;
    }

    /**
     * Returns the {@link PluginLibrary} from which this Recipe was read.
     *
     * @return  the source of this recipe
     */
    public RecipeLibrary getLibrary() {
        return pluglib_;
    }

    public String toString() {
        return new StringBuffer()
            .append(getName())
            .append(" v")
            .append(getVersion())
            .append(": ")
            .append(getSynopsis())
            .toString();
    }


    //
    // Various private methods.
    //

    /**
     * Initializes the native cpl_recipe.  Must only be called when
     * synchronized on CPLControl.LOCK, when the recipe is not already
     * in the intialized state.
     */
    private void initialize() throws CPLException {
        int status = nativeInitialize();
        if (status != 0) {
            throw new CPLException("Initialization of " + getThisObjName() +
                                   " failed with code " + status);
        }
    }

    /**
     * Deinitializes the native cpl_recipe.  Must only be called when
     * synchronized on CPLControl.LOCK, when the recipe is in the
     * initialized state.
     */
    private void deinitialize() throws CPLException {
        int status = nativeDeinitialize();
        if (status != 0) {
            throw new CPLException("Deinitialization of " + getThisObjName() +
                                   " failed with code " + status);
        }
    }
    
    /**
     * Returns a meaningful name for this object.
     */
    private String getThisObjName() {
    	if(pluglib_ != null)
    		return pluglib_.getName();
    	else
    		return "?";
    }

    /**
     * Constructs a new parameter object which has the same characteristics
     * as a given one.  The new parameter will have no dependence on
     * the original.
     *
     * @param  param  parameter to base the new one on
     * @return   a copy of <tt>param</tt>
     */
    private Parameter copyParameter(Parameter param) {
        final ParameterConstraint constraint = param.getConstraint();
        final String context = param.getContext();
        final String help = param.getHelp();
        final Object initial = param.getDefault();
        final String name = param.getName();
        final String tag = param.getTag();
        final ParameterType type = param.getType();
        try {
            return new Parameter(new ParameterImp() {
                public ParameterConstraint getConstraint() {
                    return constraint;
                }
                public String getContext() {
                    return context;
                }
                public String getHelp() {
                    return help;
                }
                public Object getInitialValue() {
                    return initial;
                }
                public String getName() {
                    return name;
                }
                public String getTag() {
                    return tag;
                }
                public ParameterType getType() {
                    return type;
                }
                public void setValue(Object value) {
                    // no action
                }
            });
        }
        catch (ParameterValueException e) {
            throw new AssertionError(e);
        }
    }

    /**
     * Copies the values from one set of parameters to another.
     * The two sets must be matching - i.e. corresponding elements
     * must have the same names, types, constraints etc.
     *
     * @param  src  source list of parameters
     * @param  dest destination list of parameters
     */
    private void copyParameterValues(Parameter[] src, Parameter[] dest) {
        int nparam = src.length;
        if (nparam != dest.length) {
            throw new IllegalArgumentException("Parameter count mismatch");
        }
        for (int i = 0; i < nparam; i++) {
            Parameter s = src[i];
            Parameter d = dest[i];
            if (! s.getName().equals(d.getName()) ||
                ! s.getType().equals(d.getType())) {
                throw new IllegalArgumentException("Parameter mismatch" +
                                                   s + " != " + d);
            }
            try {
                d.setValue(s.getValue());
            }

            // This shouldn't happen unless the constraints on these
            // parameters don't match.
            catch (ParameterValueException e) {
                throw (IllegalArgumentException) 
                      new IllegalArgumentException("Constraint mismatch: "
                                                   + s + ", " + d + "?")
                     .initCause(e);
            }
        }
    }
    
    

    // Native methods.

    private synchronized native int nativeInitialize();
    private synchronized native int nativeDeinitialize();
    private synchronized native int nativeExecute(String currentDir,
                                                  String execDir)
        throws CPLException;
    private synchronized native Parameter[] nativeGetParameterArray();
    private synchronized native Frame[] nativeGetFrameArray();
    private synchronized native void nativeSetFrameArray(Frame[] frames);

    /**
     * Two recipes are equal if they are Recipe objects with the same name
     * and version number.
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object obj) {
        boolean rc = false;
        if (obj != null && obj instanceof Recipe) {
            Recipe rec = (Recipe)obj;
            rc = (rec.getName().equals(getName()) && rec.getVersion() == getVersion());
        }
        return rc;
    }
}
