package org.eso.cpl.jni;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eso.cpl.CPLException;

/**
 * Factory for producing CPL plugin libaries.  To obtain a {@link PluginLibrary}
 * object, call the {@link #getLibrary} method.  Each instance of this
 * class has a search path (used in addition to system search paths)
 * which is used to locate such libraries which can be configured. 
 * The default search path is system-dependent (e.g. LD_LIBRARY_PATH)
 * as well as one defined by the environment variable LTDL_LIBRARY_PATH.
 *
 * @author   Mark Taylor
 */
public class LibraryLoader {

    private boolean isRegistered_;
    private List searchPath_ = new ArrayList();

    /** Static initializer ensures native library is ready.  */
    static { CPLControl.ensureSetup(); }

    /**
     * Constructs a new LibraryLoader object.
     */
    public LibraryLoader() throws LTDLException {
        register();
        isRegistered_ = true;
    }

    /**
     * Sets the custom search path for this loader.  It is searched in
     * addition to any default system search paths for unqualified
     * library names.  
     * The <tt>toString</tt> method is called on each object in the list
     * to turn it into a path element, so String or File objects are OK.
     *
     * @param path  the new search path
     */
    public void setPath(List path) {
        searchPath_ = path;
    }

    /**
     * Returns the custom search path for this loader.  It is searched in
     * addition to any default system search paths for unqualified
     * library names.  
     * The <tt>toString</tt> method is called on each object in the list
     * to turn it into a path element, so String or File objects are OK.
     *
     * @return  the search path; this may be modified in place
     */
    public List getPath() {
        return searchPath_;
    }

    /**
     * Returns a recipe library with a given name.  If this name does not
     * include a directory component, this loader's search path 
     * is searched for a library with that name.
     * Each library returned from this method
     * will be a distinct java object, but may reference the same library
     * as one returned by an earlier call.
     *
     * @param  the name of the library to load
     * @return  a PluginLibrary defined by the given name, 
     *          or <tt>null</tt> if it couldn't be located
     * @return  LTDLException if an error in libltdl prevents the library 
     *          from being loaded
     * @throws CPLException if an error in JNIRecipe prevents the library
     * 			from being loaded
     */
    public PluginLibrary getLibrary(String libName) throws LTDLException, CPLException {
        String libLoc = null;

        // If the supplied name is absolute, use that.
        if (new File(libName).isAbsolute()) {
            libLoc = libName;
        }

        // Otherwise try to make it absolute by resolving against this
        // loader's path.
        for (Iterator it = searchPath_.iterator(); 
             libLoc == null && it.hasNext();) {
            File dir = new File(it.next().toString());
            File fqfile = new File(dir, libName);
            if (fqfile.exists()) {
                libLoc = fqfile.getAbsolutePath();
            }
        }

        // Otherwise use the path as given.
        if (libLoc == null) {
            libLoc = libName;
        }
         
        // Try to get a new library using the path.
        synchronized (CPLControl.LOCK) {
            return nativeMakeLibrary(libLoc, "");
        }
    }

    /**
     * Releases any resources associated with this object.
     * May harmlessly be called multiple times.
     */
    public void close() throws LTDLException {
        if (!isRegistered_) {
            deregister();
            isRegistered_ = false;
        }
    }
    
    /**
     * @return true if this library loader is closed.
     */
    public boolean isClosed() throws LTDLException {
        return !isRegistered_;
    }

    /**
     * Calls close.
     */
    protected void finalize() throws Throwable {
        try {
            close();
        }
        finally {
            super.finalize();
        }
    }

    /**
     * Indicates that this object will be using the <tt>ltdl</tt> system.
     */
    private void register() throws LTDLException {
        synchronized (CPLControl.LOCK) {
            nativeLTDLInit();
        }
    }

    /**
     * Indicates that this object will no longer need to use the <tt>ltdl</tt>
     * system.
     */
    private void deregister() throws LTDLException {
        synchronized (CPLControl.LOCK) {
            nativeLTDLExit();
        }
    }

    // Native methods.

    private static native void nativeLTDLInit() throws LTDLException;
    private static native void nativeLTDLExit() throws LTDLException;

    private static native PluginLibrary 
        nativeMakeLibrary(String filename, String path) throws LTDLException, CPLException;
}
